
/* t2_lio.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/2/98
 * Class definition for asynchronous reads.
 */

#ifndef _T2_LIO_H_
#define _T2_LIO_H_

#include	"t2_header.h"
#include	"t2_tset.h"
#include	"t2_aio.h"
#include	"t2_clusterreq.h"

#if (T2_TIME_EXECUTE > 1)
  #include	"t2_timer.h"
#endif

class T2_LIO;

//////////////////////////////////////////////////////////////////////////////
//
// T2_LIOCB
//
// A control block for a non-blocking (ie asynchronous) I/O operation.
//
//////////////////////////////////////////////////////////////////////////////

// method to initialize fields for an array of n T2_LIOCB's
void
initLIOCB(T2_LIOCB* lcb, u_int n);

// method to set the values of the fields in a T2_LIOCB for a 
void
setMiscValues(T2_LIOCB& cb, int whence);

inline void
setValues(T2_LIOCB& cb,
	  int op, int fd, int whence, off_t off, char* buf, size_t size)
{
  cb.t2_liocb_opcode = op;		cb.t2_liocb_fildes = fd;
  cb.t2_liocb_offset = off;		cb.t2_liocb_buf = buf;
  cb.t2_liocb_nbytes = size;
  setMiscValues(cb, whence);		// set machine-dependent information
};

// method to extract the return value (ie number of bytes read/written)
// of completed asynchronous I/O operation
int
getReturnValue(T2_LIOCB& p);

// method to check if a control block completes
//bool
//opCompletes(T2_LIOCB& p);


//////////////////////////////////////////////////////////////////////////////
//
// T2_LIO
//
// A service that allows one to
// 1. register an asynchronous I/O request to be issued later, 
// 2. issue a set of asynchronous I/O requests registered earlier,
// 3. poll for pending asynchronous I/O requests issued earlier
//
// This service is the low-level support for individual aio block requests,
// each of which corresponds to a read or write of a consecutive block
// of memory. Services for cluster requests, each likely consisting
// of several block requests must be provided by some other class.
//
// The access protocol:
// The caller registers an I/O request with T2_LIO. Also, at registration,
// the caller associates a "key" with the I/O request registered with T2_LIO.
// Each registered I/O request is put on a ready list (readyl).
// An invocation of the method issueReadyRequests() issues the actual disk
// requests for all the ready I/O requests registered earlier, and move the
// corresponding T2_LIO from the ready list (readyl) to the pending list
// (pendingl).
// An invocation of the method poll() would check all the pending I/O
// requests, and move the corresponding T2_LIO from the pending list (pendingl)
// to the done list (donel).
// The caller uses method getCompletedOperation() to retrieve the key
// that corresponds to the completed I/O request.
//
// I'm not sure what is the appropriate type for the keys. The data type for
// the keys must have a copy constructor and an assignment operator (ie =).
// Since the class would contains both reads and writes for different purposes
// (eg, reads for T2_PhyBlockRequest, writes of the final output),
// for now, I'm using a class that contains two fields:
//	T2_LIOType type		which could be one of the following
//				t2_lioUnknown		-- unknown type
//				t2_lioBlockRead		-- a block read
//				t2_lioOutputWrite	-- an output write
//	void* ptr		depending on type, the pointer points to
//				T2_PhyBlockRequest  if type is t2_lioBlockRead
//				T2_QEnv		    if type is t2_lioOutputWrite
//
//
// The argument to the constructor of T2_LIO specifies the max number of
// concurrent disk I/O requests that T2_LIO can handle.
//
//////////////////////////////////////////////////////////////////////////////
//typedef int	T2_LIOKey;
//typedef T2_PhyBlockRequest*	T2_LIOKey;

enum T2_LIOType { t2_lioUnknown=0, t2_lioBlockRead, t2_lioOutputWrite };

class T2_LIOKey {
public:
  T2_LIOKey(): type(t2_lioUnknown), ptr(NULL) {};
  T2_LIOKey(T2_LIOType t, void* p): type(t), ptr(p) {};
  T2_LIOKey(const T2_LIOKey& p): type(p.type), ptr(p.ptr) {};
  ~T2_LIOKey() {};

  T2_LIOKey& operator =(const T2_LIOKey& p);

  T2_LIOType getType() const { return type; };
  void* getPtr() const { return ptr; };

protected:
  T2_LIOType type;
  void* ptr;

private:
};

inline T2_LIOKey&
T2_LIOKey::operator =(const T2_LIOKey& p)
{
  type = p.type;
  ptr = p.ptr;
  return *this;
}


class T2_LIO {
public:
  T2_LIO(const u_int n);
  ~T2_LIO();

  // query methods
  u_int getNumberFreeCB() const { return freel.getNumberElements(); };
  u_int getNumberPendingCB() const { return pendingl.getNumberElements(); };
  u_int getNumberReadyCB() const { return readyl.getNumberElements(); };
  u_int getNumberDoneCB() const { return donel.getNumberElements(); };
  bool noMoreFreeCB() const { return (getNumberFreeCB() == 0)? true:false; };
  bool noMorePendingCB() const { return (getNumberPendingCB()==0)? true:false;};
  bool noMoreReadyCB() const { return (getNumberReadyCB() == 0)? true:false; };
  bool noMoreDoneCB() const { return (getNumberDoneCB() == 0)? true:false; };
  bool idle() const { return (getNumberFreeCB() == nlcb)? true : false; };

  // register a T2_PhyBlockRequest read request to be issued later
  bool registerRead(const T2_LIOKey& k, T2_PhyBlockRequest& breq,
		    int whence=SEEK_SET);

  // register a write request to be issued later
  bool registerWrite(const T2_LIOKey& k,
		 int fd, int whence, off_t off, const char* buf, size_t size);

  // issue all the ready requests registered earlier
  T2_AsyncResult issueReadyRequests();

  // poll for completed operations, which will be inserted into donel;
  // return the additional number of completed operations
  u_int poll();

  // retrieve the key of the first completed T2_LIOCB in donel, plus the
  // return value of the corresponding I/O request;
  // return true if there is a completed operation, return false otherwise
  bool getDoneKey(T2_LIOKey& k, int &sz);

#if (T2_TIME_EXECUTE > 1)
  // connect to timer
  void attachTimer(T2_ExecuteTimer* t) { exetimer = t; };
#endif

protected:
  u_int nlcb;			// total number of T2_LIOCB (ie size of lcb[])
  T2_LIOCB  *lcb;		// an array of lio_listio control blocks
  T2_LIOKey *key;		// an array of keys corresponding to lcb[]
  T2_TSet<T2_LIOCB*>		// lists of pointers to T2_LIOCB
	   freel,		// a list of free (unused) T2_LIOCB
	   pendingl,		// a list of pending (issued) T2_LIOCB
	   readyl,		// a list of T2_LIOCB ready to be issued
	   donel;		// a list of completed T2_LIOCB

  T2_LIOAux aux;		// interface-dependent aux data

#if (T2_TIME_EXECUTE > 1)
  T2_ExecuteTimer* exetimer;	// a pointer to a timer for the execution
#endif

private:
  sigevent lio_sigevent;	// signal event for lio_listio

  // to initialize and delete the interface-specific auxiliary data T2_LIOAux
  void initAux();
  void deleteAux();

  // register a request to be issued later
  bool registerReq(const T2_LIOKey& k,
		   int op, int fd, int whence, off_t off,
		   char* buf, size_t size);

  // register a read request to be issued later
  bool registerRead(const T2_LIOKey& k, 
		    int fd, int whence, off_t off, char* buf, size_t size);
};


inline bool
T2_LIO::registerRead(const T2_LIOKey& k, 
		     int fd, int whence, off_t off, char* buf, size_t size)
{
  return registerReq(k, LIO_READ, fd, whence, off, buf, size);
}

inline bool
T2_LIO::registerRead(const T2_LIOKey& k, T2_PhyBlockRequest& breq, int whence)
{
  t2_Assert (!breq.hasNoBuffer());
  t2_Assert (breq.getStatus() == t2_bActivated);
  if (registerRead(k, breq.getFileDescriptor(), whence, breq.getOffset(),
		   breq.getBuffer().getBuffer(), breq.getNumberBytes()) == true)
  {
     // registerRead succeeded
     breq.setStatus(t2_bPending);
     return true;
  }
  return false;
}

inline bool
T2_LIO::registerWrite(const T2_LIOKey& k,
		    int fd, int whence, off_t off, const char* buf, size_t size)
{
  return registerReq(k, LIO_WRITE, fd, whence, off, (char *) buf, size);
}



#endif	/* !_T2_LIO_H_ */
