
/* $Id: pbe.C,v 1.14 2000/03/16 00:44:46 chialin Exp $ */

/* pbe.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/5/98
 * Main program for the parallel T2 back-end server.
 */


#include	<iostream.h>
#include	<fstream.h>
#include	<stdlib.h>
#include        <unistd.h>
#include        <new.h>
#include        <string.h>
#include        <errno.h>
#include        <sys/param.h>
#include	"t2_header.h"
#include	"t2_commfunc.h"
#include	"t2_bfile.h"
#include	"t2_qspec.h"
#include	"t2_rts.h"
#include	"t2_bcache.h"
#include	"t2_packets.h"


extern int _g_mynode;

enum usageError { eUNKNOWN, eBADARG };
static void printusage(const char*);
static void badusage(usageError, const char*, const char*);
void query_processing(T2_RTS& rts, T2_QBatch& qbatch, size_t mem, bool verbose);
void be_qbatch_file(const char* qbatch_file, T2_RTS& rts,
		    size_t usr_mem, const char* ofname, bool verbose);
void be_fe_socket(const char* fe_hname, int fe_port, T2_RTS& rts,
             	    size_t usr_mem, const char* ofname, bool verbose);
void
force_output_to_files(T2_QBatch& qbatch, const char* ofname, u_int& qcnt);


void t2_setup(int& argc, char**& argv, char**& envp, T2_RTS& rts);
void t2_finalize();

#ifndef T2_SYS_SUPPORT_GETHOSTNAME
  extern "C" int gethostname(char* name, int namelen);
#endif

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN  256
#endif


//////////////////////////////////////////////////////////////
//
// new handler
//
//////////////////////////////////////////////////////////////
static void t2_newHandler(void)
{
# if defined(T2_ASSERTION)
     static int wCount = 0;
     if (errno == EINTR && ++wCount < 25) {
       t2_say3(cerr, "EINTR malloc, retry ..." << endl);
       return;
     }
# else /* T2_ASSERTION */
     if (errno == EINTR) return;
# endif /* T2_ASSERTION */

  t2_say_label(cerr);
  cerr << "out of free store." << endl;
  abort();
}


//////////////////////////////////////////////////////////////
//
// some default values
//
//////////////////////////////////////////////////////////////
const size_t default_usrmem = 1024*1024*5;	// usr memory in bytes
char default_dscatfile[] = "dataset-catalog";	// dataset catalog
char default_idxcatfile[] = "index-catalog";	// index catalog
char default_qbatchfile[] = "qbatch-file";	// query batch
char *default_diskassignfile = NULL;		// disk assignment file
//char default_outputfile[] = "output-file";	// output file
char *default_outputfile = NULL;		// output file
char *default_servername = "t2_server";		// server name for meta-chaos
bool default_config_byrank = false;		// assign config by ranks


//////////////////////////////////////////////////////////////
//
// main program
//
//////////////////////////////////////////////////////////////
main(int argc, char** argv, char** envp)
{
  // register new handler, and initialize errno to 0
  set_new_handler(t2_newHandler);
  errno = 0;

  T2_RTS rts;

  // initializing environment
  t2_setup(argc, argv, envp, rts);


  // variables
  size_t usr_mem = default_usrmem;	// memory for accumulator (in bytes)
  char *dscat_file = default_dscatfile,		// dataset catalog
       *idxcat_file = default_idxcatfile,	// index catalog
       *qbatch_file = default_qbatchfile,	// query batch
       *diskassign_file = default_diskassignfile, // disk assignment file
       *output_file = default_outputfile;	// output file
  char *fe_hname = NULL;			// front-end hostname
  int  fe_port = 0;				// front-end port
  bool verbose = false;
  char *mc_servername = default_servername;	// server name for meta-chaos
  bool config_byrank = default_config_byrank;


  // deal with the command-line arguments
  int i;
  for (i=0; i<argc; i++) {
     if (argv[i] == NULL)
	continue;

     if (argv[i][0] == '-') {
	switch(argv[i][1]) {
	case 'h':
		printusage(argv[0]); return 0;
		break;
	case 'd':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1]); return -1; }
		dscat_file = argv[i];
		break;
	case 'i':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1]); return -1; }
		idxcat_file = argv[i];
		break;
	case 'q':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1]); return -1; }
		qbatch_file = argv[i];
		fe_hname = NULL;
		break;
	case 'a':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1]); return -1; }
		diskassign_file = argv[i];
		config_byrank = false;
		break;
	case 'r':
		config_byrank = true;
		diskassign_file = NULL;
		break;
	case 'o':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1]); return -1; }
		output_file = argv[i];
		break;
	case 'm':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1]); return -1; }
		usr_mem = atoi(argv[i]);
		break;
	case 'f':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1]); return -1; }
		fe_hname = argv[i];
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-2]); return -1; }
		fe_port = atoi(argv[i]);
		qbatch_file = NULL;
		break;
	case 's':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1]); return -1; }
		mc_servername = argv[i];
		break;
	case 'v':
		verbose = true;
		break;
	default:
		badusage(eUNKNOWN, argv[0], argv[i]); return -1; break;
	}
     }
  }

  t2_Assert (fe_hname == NULL || qbatch_file == NULL);
  const u_int numnodes = rts.getSystem().getNumberNodes(),
              mynode = rts.getSystem().getMyNode();

  if (numnodes > T2_MAX_NUM_BE_PROCS) {
     t2_err_quit(__FILE__, __LINE__,
		"ERROR: maximum number of processors allowed is %d, abort.",
		T2_MAX_NUM_BE_PROCS);
     return -1;
  }

  // reporting arguments
  if (mynode < 1) {
     cout << "dataset catalog prefix = " << dscat_file << endl;
     cout << "index catalog prefix = " << idxcat_file << endl;
     if (fe_hname == NULL)
        cout << "query batch input file = " << qbatch_file << endl;
     else
        cout << "front-end hostname = " << fe_hname << ", port = " << fe_port
   	  << endl;
     cout << "disk-assignment file = " << t2_stringEmptyWrap(diskassign_file)
	  << endl;
     if (output_file != NULL)
        cout << "forcing to write query output into file with prefix = "
   	  << output_file << endl;
     else
        cout << "sending output by however specified by the query specification"
   	  << endl;
     cout << "user memory space = " << usr_mem << " bytes" << endl;
     cout << endl;
  }


  char helpmsg[128];
  sprintf(helpmsg, "use '%s -h' to see usage", argv[0]);

  const char gethostname_failed[] = "gethostname failed";
  char hostname[MAXHOSTNAMELEN];
  if (gethostname(hostname, MAXHOSTNAMELEN) != 0) {	// gethostname failed
     strcpy(hostname, gethostname_failed);
  }

  // initialize T2 rts with system files
  // there are 3 scenarios:
  // 1) no disk assignment file is given (ie diskassign_file == NULL)
  //    and no '-r' was set (ie config_byrank == false)
  //    --> follow the disk assignment in the header of the dataset catalog
  //	    and assign to local processors all logical disk ids with a
  //	    matching hostname
  // 2) a disk assignment file is given (ie diskassign_file != NULL),
  //    in which case all previous '-r' were ignored (ie config_byrank == false)
  //    --> use the disk assignmentment file along with the header of the 
  //	    dataset catalog to assign logical disk ids
  // 3) no disk assignment file is given (ie diskassign_file == NULL),
  //    and '-r' was set (ie config_byrank == false)
  //    --> simply assign logical processor ids to processors based on ranks,
  //	    ie (logical_processor_id) % numnodes == mynode
  if (config_byrank == false) {
     if (strcmp(hostname, gethostname_failed) == 0) {
	// gethostname failed earlier !
	t2_err_quit(__FILE__, __LINE__, "gethostname error");
	return -1;
     }
     if (diskassign_file == NULL) 		// scenario (1)
	rts.initHostName(dscat_file, idxcat_file, hostname, mc_servername);
     else					// scenario (2)
	rts.initDiskAssignFile(dscat_file, idxcat_file, diskassign_file,
			       hostname, mc_servername);
  }
  else {					// scenario (3)
     t2_Assert (diskassign_file == NULL);
     rts.initRank(dscat_file, idxcat_file, mc_servername);
  }

  if (!rts) {
	t2_err_quit(NULL, 0, "T2_RTS initialization failed, %s", helpmsg);
	return -1;
  }

  t2_say(cout, "after rts.init, numnodes=" << rts.getSystem().getNumberNodes()
	 	<< ", mynode = " << rts.getSystem().getMyNode() << endl);

  if (strcmp(hostname, gethostname_failed) != 0) {
     t2_say_label(cout);
     cout << hostname << " is back-end node "
	  << rts.getSystem().getMyNode() << endl;
  }

  // load user-defined constructor functions
  T2_RTS::initStaticArrays();


  // initialize an empty block cache
  T2_BlockCache bcache(0);
  T2_PhyBlockRequest::setBlockCache(&bcache);


  // now, process the queries
  if (fe_hname == NULL)
     be_qbatch_file(qbatch_file, rts, usr_mem, output_file, verbose);
  else
     be_fe_socket(fe_hname, fe_port, rts, usr_mem, output_file, verbose);


  // cleaning up


  t2_finalize();
  t2_say(cout, "server shutting down gracefully" << endl);
  return 0;
}


//////////////////////////////////////////////////////////////
//
// query processing functions
//
//////////////////////////////////////////////////////////////
void
be_qbatch_file(const char* qbatch_file, T2_RTS& rts, size_t usr_mem,
	       const char* ofname, bool verbose)
{
  static u_int qcnt = 0;

  // reading query batch
  T2_IBFile qibf(qbatch_file);
  if (!qibf) {
     t2_err_quit(__FILE__, __LINE__,
		"Cannot open query batch file %s", qbatch_file);
     return;
  }

  t2_say(cout, "reading query batch from " << qbatch_file << endl);
  T2_QBatch qbatch;
  qibf >> qbatch;
  qibf.close();
  if (qbatch.isValid() == false) {	// qbatch read failed
     t2_err_quit(__FILE__, __LINE__,
		"Failed to read query batch from %s", qbatch_file);
     return;
  }
  if (qbatch.getNumberQueries() == 0)	// got an empty query batch
     return;

  if (ofname != NULL)
     force_output_to_files(qbatch, ofname, qcnt);

  query_processing(rts, qbatch, usr_mem, verbose);
}



void
be_fe_socket(const char* fe_hname, int fe_port, T2_RTS& rts,
             size_t usr_mem, const char* ofname, bool verbose)
{
  static u_int qcnt = 0;

  const int numnodes = rts.getSystem().getNumberNodes(),
            mynode = rts.getSystem().getMyNode();
  const u_char endian = rts.getSystem().getBackEndEndian();

  // connecting to the front-end
  T2_FEConnection fecon;
  if (fecon.connectFrontEnd(fe_hname, fe_port, numnodes, mynode, endian)
     == false)
  {
     t2_err_ret(__FILE__, __LINE__,
                "%d: T2_FEConnection cannot connect to front-end %s port %d\n",
                _g_mynode, fe_hname, fe_port);
     return;
  }

  t2_say(cout, _g_mynode
	        << ": connected to front-end " << fe_hname << " at port "
		<< fe_port << " successfully" << endl);


  // recv packets from front-end
  u_char tag;
  u_int cnt = 0;
  while ((tag = fecon.recvPacketTag()) != T2_ErrorFEBEPacketTag) {
     switch (tag) {
	case T2_ConnectPacketTag:
	case T2_AcceptPacketTag:
	case T2_ReadyPacketTag:
			errno = 0;
			t2_err_ret(__FILE__, __LINE__,
				 "unexpected packet tag %d from the front-end",
				  (int) tag);
			break;

	case T2_QueryPacketTag: 
			{
			   T2_QBatch qbatch;
			   fecon.recvQBatch(qbatch);
			   t2_Assert (qbatch.isValid() == true);
			   t2_say(cout,
				endl
				<< _g_mynode << ": back-end"
				<< " gets qbatch # " << cnt++ << " with "
				<< qbatch.getNumberQueries()
				<< " queries"
				<< endl);
			   
			   if (ofname != NULL)
				force_output_to_files(qbatch, ofname, qcnt);

			   query_processing(rts, qbatch, usr_mem, verbose);

			   t2_be_gsync();
			   bool r = fecon.sendQBatchAck(mynode);
			   t2_Assert (r == true);
			}
			break;

	case T2_ShutdownPacketTag:
			t2_say(cout,
			     endl << _g_mynode
			     << ": back-end "
			     << "recvs shutdown msg tag, shutting down"
			     << endl << endl);
			break;

	default:
			errno = 0;
			t2_err_ret(__FILE__, __LINE__,
				"unrecognized packet tag %d from the front-end",
				(int) tag);
     }

     if (tag == T2_ShutdownPacketTag)		// to shutdown
	break;

  }				// while (tag != T2_ErrorFEBEPacketTag)


  if (tag == T2_ErrorFEBEPacketTag) {
     t2_err_ret(__FILE__, __LINE__, "recvPacketTag error, shutting down");
  }
}


void
force_output_to_files(T2_QBatch& qbatch, const char* ofname, u_int& qcnt)
{
  if (ofname != NULL) {		// force write the output to files if specified
     for (u_int i=0; i<qbatch.getNumberQueries(); i++) {
	if (strlen(ofname) > 220) {
	   t2_err_ret(__FILE__, __LINE__, "output file too long");
	   return;
	}

	char outfile[256];
	sprintf(outfile, "%s-%d", ofname, qcnt++);	// get output filename

	T2_QSpec& qspec = qbatch.getQuerySpec(i);
	T2_QSpecOutput& qspecout = qspec.getOutputSpec();
	qspecout.setOutputHandleType(t2_oBFile);	// force output to
	qspecout.setFilePrefix(outfile);		// a file
     }
  }
}



//////////////////////////////////////////////////////////////
//
// some help functions
//
//////////////////////////////////////////////////////////////
void printusage(const char *name)
{
  if (_g_mynode > 0)
     return;

  cout << "Usage: " << name
    << " [-h] | [-d <dataset catalog file>] [-i <index catalog file>] " << endl
    << " [-c <dataset config-set file>] [-q <query batch file>]"
//    << " [-p <number of processors>]"
    << endl
    << " [-o <output file prefix>] [-m <memory size (bytes) for accumulator>]"
    << endl
    << " [-f <front-end hostname> <front-end port>]" << endl
    << " [-s <backend server name for Meta-Chaos>]"
    << " [-v]"
    << endl;

  cout << "Where:" << endl;
  cout << "-h prints this message" << endl;
  cout << "-d specifies the dataset catalog file prefix"
       << " (default = " << default_dscatfile << ")" << endl;
  cout << "-i specifies the index catalog file prefix"
       << " (default = " << default_idxcatfile << ")" << endl;
  cout << "-a specifies the disk assignment file"
       << " (default = " << t2_stringNullWrap(default_diskassignfile)
       << ")" << endl;
  cout << "-r assigns dataset configuration files by processor ranks"
       << " (default is by hostname)" << endl;
  cout << "-q specifies the query batch file"
       << " (default = " << default_qbatchfile << ")" << endl;
  cout << "-o forces the back-end to write outputs into files with the specified prefix"
       << endl
       << "   the complete output file name will be in the form of:"
       << endl
       << "   <output file prefix>-<query serial no.>-<iteration no.>.<proc no.>"
       << endl
       << "   where <query serial no.> is some assigned query id" << endl
       << "         <iteration no.> is a strip-mined iteration counter" << endl
       << "         <proc no.> is the processor number" << endl
       << "   (default output file prefix = NULL)"
       << endl;
  cout << "-m specifies the memory size (bytes) for accumulators"
       << " (default = " << default_usrmem << " bytes)" << endl;
  cout << "-f specifies hostname and port number for the front-end process"
       << endl
       << "   (by default, program reads query batch from a file)"
       << endl;
  cout << "-s specifies the backend server name for Meta-Chaos"
       << " (default = " << default_servername << ")"
       << endl;
  cout << "-v enables verbose mode" << endl;

#ifdef T2_MPI_COMM
  cout << "-p4pg <host filename> specifies the hostfile to be used by MPI"
       << endl;
#endif

}


void badusage(usageError e, const char* p, const char* a)
{
  if (_g_mynode > 1)
     return ;

  switch (e) {
    case eUNKNOWN:
	cerr << "Unknown parameter " << a << " in command line" << endl;
	break;
    case eBADARG:
	cerr << "Missing argument for parameter -" << a << " in command line"
	     << endl;
	break;
    default:
	cerr << "Some kind of error in the parameter " << a
	     << " in command line" << endl;
  }
  printusage(p);
}


