
/* $Id: pexecute.C,v 1.15 2000/11/22 03:28:54 chialin Exp $ */

/* pexecute.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/15/98
 * parallel query execution service
 */


#include	<iostream.h>
#include	<stdio.h>
#include	"t2_header.h"
#include	"t2_qspec.h"
#include	"t2_lio.h"
#include	"t2_work.h"
#include	"t2_rts.h"
#include	"t2_qenv.h"
#include	"t2_qplan.h"
#include	"t2_outhandle.h"
#include	"t2_outsock.h"
#include	"t2_comm.h"

#ifdef T2_TIME_EXECUTE
   #include	"t2_timer.h"
#endif


T2_Iteration niter = 0;
//const T2_Iteration maxIter = 5000000;		// max number of iterations
const T2_Iteration maxIter = 50000;		// max number of iterations

extern int _g_mynode;


//////////////////////////////////////////////////////////////////
//
// some internal settings
//
//////////////////////////////////////////////////////////////////
const u_int workListLen = 100;		// max length of the work list


//////////////////////////////////////////////////////////////////
//
// some help functions
//
//////////////////////////////////////////////////////////////////
void
issueMoreReads(T2_QPlan& qplan, T2_LIO& lio, T2_WorkList& wlist);

u_int
registerDiskReads(T2_ClusterRequest& cluster, u_int& cur_block,
		  T2_LIO& lio, T2_WorkList& wlist);

void
pollDiskReads(T2_LIO& lio, T2_WorkList& wlist, T2_QPlan& qplan);


bool
tooManyIterations(T2_Iteration i, T2_RTS& rts, T2_Array<T2_QEnv>& qenvs,
		  T2_QGlobalStatus& qglobal,
		  T2_QPlan& qplan, T2_LIO& lio, T2_WorkList& wlist,
		  T2_OutputSock& outsocks, T2_CommManager& comm_manager);

void
reportWhileCondition(T2_Array<T2_QEnv>& qenvs, T2_QGlobalStatus& qglobal,
		     T2_QPlan& qplan, T2_LIO& lio, T2_WorkList& wlist,
		     T2_OutputSock& outsocks, T2_CommManager& comm_manager);
void
reportQEnvCondition(u_int i, T2_QEnv& qenv);

void
blockReadCompletes(T2_PhyBlockRequest* blockreq, int sz, T2_WorkList& wlist);

void
outputWriteCompletes(T2_OutputHandle* ohandle, int sz, T2_WorkList& wlist,
		     T2_QPlan& qplan);

/////////////////////////////////////////////////////////////////////
//
// error_output_handling
//
// This function takes the appropriate action for a query that
// was invalidated by the query planning service.
//
// Not sure what is the appropriate action.
// For now, this is what I do.
// If output handle type is t2_oBFile, then create an empty file.
// If output handle type is t2_oSocket, then connect to the client
// and close the connection immediately.
// Do nothing for other types.
//
/////////////////////////////////////////////////////////////////////
inline void
error_output_handling(T2_QEnv& qenv)
{
  t2_Assert (qenv.isValid() == false);
  T2_OutputHandle* outh = qenv.getOutputEnv().getOutputHandle();
  t2_Assert (outh != NULL);
  outh->prematureErrorHandling();
}


//////////////////////////////////////////////////////////////////
//
// handle a work entry in the work list
//
//////////////////////////////////////////////////////////////////
inline void
doWork(T2_WorkList& wlist)
{
  if (!wlist.isEmpty()) {
     T2_WorkEntry* w = wlist.dequeue();
     t2_Assert (w != NULL);
     w->handler();
     delete w;
  }
}



//////////////////////////////////////////////////////////////////
//
// executing a query plan
//
//////////////////////////////////////////////////////////////////
void
execute_qplan(T2_QPlan& qplan, T2_Array<T2_QEnv>& qenv, T2_RTS& rts,
	      bool verbose)
{

  t2_say(cout, endl);
  t2_say(cout, "just entered execute_qplan" << endl);


# ifdef T2_TIME_EXECUTE
     T2_ExecuteTimer executetimer;
     executetimer.startTimer();
# endif


  u_int i;		// loop index
  const u_int nallqry = qenv.getNumberElements();	// # all queries
  u_int nqry = 0;					// # of valid queries
  T2_System& system = rts.getSystem();
  const u_int numnodes = system.getNumberNodes();
  const T2_ProcID mynode = system.getMyNode();

  t2_Assert (qplan.getNumberQQPlans() == nallqry);

  // make sure all queries start at iteration 0
  for (i=0; i<nallqry; i++) {
     if (qenv[i].isValid() == true) {
#	ifdef T2_ASSERTION
           T2_QQPlan& qqplan = 
#	endif
	qplan.getQQPlan(i);
        t2_Assert (qqplan.getCurrentIteration() == 0);
        t2_Assert (qqplan.getCurrentCluster() == 0);
	nqry++;
     }
     else
	// this is a query that will not be executed at all, so call
	// error handling routine for output
	error_output_handling(qenv[i]);
  }
  if (nqry == 0) {
     t2_say(cout, "no valid query from the " << nallqry
		   << "-query batch, so skip execution service completely"
		   << endl);
     return;
  }


  // initialize objects needed by the executor first:

  // initialize lio object with some number of control block entries (nlio_cb);
  // to write the final output, each query will need one control block, so
  // for now, let's make the total number of control blocks the greater of
  // nlio_cb and 5 times the number of queries
  t2_say2(cout, "about to initialize T2_LIO" << endl);
  const u_int nlio_cb = 50;
  T2_LIO lio(t2_max(nlio_cb, 5 * nqry));

  // initialize T2_OutputHandle::lio
  T2_OutputHandle::lio = &lio;


  // initialize communication manager;
  // not sure how many pending requests should T2_Comm handle; for now,
  // let's just make sure enough is given
  t2_say2(cout, "about to initialize T2_Comm and T2_CommManager" << endl);
  T2_Comm comm(numnodes * nqry + numnodes);
  T2_CommManager comm_manager(numnodes, comm.getNumberFreeCB(), & comm);
  comm_manager.initBeforeQueryExecution();

  // initialize T2_OHMetaChaos::comm
//  T2_OHMetaChaos::setCommManager(&comm_manager);


  // set the max number of entries in the work list;
  // we definitely don't want the work list to be shorter than the number of
  // queries (otherwise if all the queries advance to the global combine phase
  // at the same time, those events would fill the work list, which will
  // prevent other things from happening);
  // for now, let's say we set the work list length to be the greater of
  // workListLen and twice the number of queries plus some slacks;
  const u_int wlist_slacks = nqry;
  T2_WorkList wlist(t2_max(workListLen, 2 * nqry + wlist_slacks));
  wlist.init(&rts, &comm_manager);


  // initialize output socket set
  // it needs to be able to contain at least one socket per query
  T2_OutputSock outsocks(nqry);

  // initialize T2_OHSocket::outsock
  T2_OHSocket::outsock = &outsocks;


# if (T2_TIME_EXECUTE > 1)
     lio.attachTimer(&executetimer);		// connect lio to timer
# endif

# ifdef T2_TIME_EXECUTE
     executetimer.setEndQExecSetup();		// end of execution setup
# endif


  t2_say2(cout, "about to call qplan.initBeforeExecution" << endl);


  // initialize qplan before query execution; this must be called before
  // T2_QEnv::initIteration is ever called
  qplan.initBeforeExecution();


# ifdef T2_TIME_EXECUTE
     executetimer.setEndQPlanInit();		// end of qplan initialization
# endif


  // initialize global query status
  T2_QGlobalStatus& qglobal = T2_QStatus::getGlobalStatus();
  qglobal.setNumberQueriesNotDone(nqry);


  // initialize each query env for the entire execution and for iteration 0
  // for each query;
  // this must be called after T2_QPlan::initBeforeExecution is called
  for (i=0; i<nallqry; i++) {
     T2_QEnv& qe = qenv[i];
     if (qe) {
        if (qe.initBeforeQueryExecution(rts) == false) {
	   qe.invalidate();
	   qglobal.decNumberQueriesNotDone();

	   // inform other back-end nodes to kill this query
	   t2_Assert (i == qe.getQEnvID());
	   comm_manager.killRemoteQuery(i, numnodes, mynode);

	   continue;
	}

#	ifdef T2_TIME_QUERY
	   T2_QQPlan* qqp = qe.getQQPlan();
	   t2_Assert (qqp != NULL);
	   t2_Assert (qqp->getNumberSubplans() > 0);
	   t2_Assert (qqp->getCurrentIteration() == 0);
	   T2_QueryIterTimer& qtimer = qqp->getCurrentSubplan();
	   qtimer.setStartInitIteration();	// start of initIteration()
#	endif

        if (qe.initIteration(wlist) == false) {
	   qe.invalidate();
	   qglobal.decNumberQueriesNotDone();

	   // inform other back-end nodes to kill this query
	   t2_Assert (i == qe.getQEnvID());
	   comm_manager.killRemoteQuery(i, numnodes, mynode);

	   continue;
	}

#	ifdef T2_TIME_QUERY
	   qtimer.setEndInitIteration();	// end of initIteration()
#	endif

     }
  }


# ifdef T2_TIME_EXECUTE
     executetimer.setEndQEnvInit();		// end of qenv initialization
# endif


  t2_say(cout, "about to enter main while loop in execution service" << endl);


  // enter the main work loop here
  while (
        qglobal.getNumberQueriesNotDone() > 0
					// not all queries done
//     || comm_manager.getNumberCurrentSends() > 0
					// more sends to complete
//     || comm.getNumberPendingCB() > 0	// not all comm ops complete
//     || comm.getNumberDoneCB() > 0	// not all comm ops polled
     || !lio.idle() 			// no IO activity at all
     || !wlist.isEmpty()		// work list is not empty
  )
  {
     // too many iterations? just a safety valve
//     if (tooManyIterations(++niter, rts, qenv,
//			   qglobal, qplan, lio, wlist, outsocks, comm_manager))
//	break;


#    ifdef T2_TIME_EXECUTE
	double tsnaps[6];
	tsnaps[0] = t2_dclock();
#    endif


     // issue more non-blocking reads when ready
     issueMoreReads(qplan, lio, wlist);

#    ifdef T2_TIME_EXECUTE
	tsnaps[1] = t2_dclock();
#    endif


     // check for pending non-blocking data sends in local-reduction phase
     comm_manager.poll(qenv, wlist);


     // check for pending non-blocking accumulator sends in global-combine phase


     // check for pending non-blocking data recvs in local-reduction phase


     // check for pending non-blocking accumulator recvs in global-combine phase


#    ifdef T2_TIME_EXECUTE
	tsnaps[2] = t2_dclock();
#    endif


     // check for pending disk reads
     pollDiskReads(lio, wlist, qplan);

#    ifdef T2_TIME_EXECUTE
	tsnaps[3] = t2_dclock();
#    endif


     // check for pending output writes
     outsocks.poll(wlist);


#    ifdef T2_TIME_EXECUTE
	tsnaps[4] = t2_dclock();
#    endif


     // grab a work entry from the work list and invoke its handler
     doWork(wlist);

#    ifdef T2_TIME_EXECUTE
	tsnaps[5] = t2_dclock();
#    endif


#    ifdef T2_TIME_EXECUTE
	executetimer.incIssueReadTime		(tsnaps[1] - tsnaps[0]);
	executetimer.incPollCommTime		(tsnaps[2] - tsnaps[1]);
	executetimer.incPollDiskReadTime	(tsnaps[3] - tsnaps[2]);
	executetimer.incPollSockWriteTime	(tsnaps[4] - tsnaps[3]);
	executetimer.incDoWorkTime		(tsnaps[5] - tsnaps[4]);
#    endif


  }						// end of main while loop


  t2_say2(cout, 
         "just left main magic loop after " << niter << " iterations"<< endl);

  // Recall that during the execution service, when a back-end node decides
  // to invalidate (ie kill) a query due to some error, it would send out
  // a kill msg to each of the remote nodes to ask them to kill that query
  // as well. Since these messages are asynchronous, there is a possibility
  // that though the local processor is finished processing all its queries,
  // some other slower processor that is not quite done yet just decides 
  // to kill a query and therefore send a kill message to the local processor.
  // If these late kill messages are not drained from the network, they
  // will be received when the local processor starts to process the next
  // batch of queries, and therefore would mistakenly kill the query in
  // the next batch.
  // On the other hand, there is also the possibility that the local back-end 
  // node kills its query before that query reaches global combine phase but
  // not after that same query on another back-end node reaches the global
  // combine and phase and therfore has sent acc msgs to the local back-end
  // node. The local back-end node might not expect to receive those messages
  // and therefore just exit the main while loop without posting receives
  // for those acc msgs. Thes messages would then be mistakenly received 
  // when processing the next batch of queries.
  // We now therefore need to drain the network.
  //
  // drain the network
  comm_manager.drainMessages(mynode, qenv);


# ifdef T2_TIME_EXECUTE
     executetimer.setEndQExec();		// end of execution

     // report timing info
     cout << executetimer << endl;
# endif

# ifdef T2_TIME_QUERY
  {
     const u_int n = qplan.getNumberQQPlans();
     for (u_int ii=0; ii<n; ii++) {
       if (! (qenv[ii]))
          continue;
       T2_QQPlan& qqplan = qplan.getQQPlan(ii);
       const u_int s = qqplan.getNumberSubplans();
       for (u_int jj=0; jj<s; jj++) {
          T2_QueryIterTimer& qtimer = qqplan.getSubplan(jj);
          cout << qtimer << endl;
       }
     }
  }
# endif


  t2_say(cout, "returning from execute_qplan" << endl);
}



//////////////////////////////////////////////////////////////////
//
// executor
//
//////////////////////////////////////////////////////////////////
//void
//executor(T2_QPlan& qplan, T2_Array<T2_QEnv>& qenvarray, T2_GBuffer& gbuf,
//	 T2_LIO& lio, T2_CommRecv& arecv, T2_CommSend& asend,
//	 T2_CommRecvAcc& accrecv, T2_CommSendAcc& accsend)



//////////////////////////////////////////////////////////////////
//
// report conditions when too many iterations are performed
//
//////////////////////////////////////////////////////////////////
bool
tooManyIterations(T2_Iteration i, T2_RTS& rts, T2_Array<T2_QEnv>& qenvs,
		  T2_QGlobalStatus& qglobal,
		  T2_QPlan& qplan, T2_LIO& lio, T2_WorkList& wlist,
		  T2_OutputSock& outsocks, T2_CommManager& comm_manager)
{
  if (i < maxIter) {
     return false;
  }
  cout << endl;
  cout << rts.getSystem().getMyNode()
       << ": ======= breaking from the main while loop at iteration " 
       << i << endl;
  reportWhileCondition(qenvs, qglobal, qplan, lio, wlist,
		       outsocks, comm_manager);
  return true;
}


void
reportWhileCondition(T2_Array<T2_QEnv>& qenvs, T2_QGlobalStatus& qglobal,
		     T2_QPlan& qplan, T2_LIO& lio, T2_WorkList& wlist,
		     T2_OutputSock& outsocks, T2_CommManager& comm_manager)
{
  cout << endl;

  for (u_int i=0; i<qenvs.getNumberElements(); i++) {
     T2_QEnv& qenv = qenvs[i];
     reportQEnvCondition(i, qenv);
  }
  cout << endl;

  t2_say_label(cout);
  cout << "T2_QGlobalStatus: # queries not completed = "
       << qglobal.getNumberQueriesNotDone() << endl;

//  cout << "T2_QPlan: nqqplanNotCompleted = "
//       << qplan.getNumberQQPlanNotCompleted()
//       << endl;

  t2_say_label(cout);
  cout << "T2_LIO: # ready = " << lio.getNumberReadyCB()
       << ", # pending = " << lio.getNumberPendingCB()
       << ", # done = " << lio.getNumberDoneCB() << endl;

  t2_say_label(cout);
  cout << "T2_WorkList: contains " << wlist.getNumberElements()
       << " entries" << endl;

  t2_say_label(cout);
  cout << "T2_OutputSock: contains " << outsocks.getNumberOutputs()
       << " entries" << endl;

  t2_say_label(cout);
  cout << "T2_CommManager : contains " 
       << comm_manager.getNumberPendingRequests() << " pending entries" 
       << endl;

  cout << endl;
}


void
reportQEnvCondition(u_int i, T2_QEnv& qenv)
{
  if (!qenv) {
     t2_say_label(cout);
     cout << "query " << i << " (" << (int) &qenv
          << ") is invalid" << endl;
     return;
  }

  t2_say_label(cout);
  cout << "query " << i << " (" << (int) &qenv
       << ") at phase " << toString(qenv.getQueryStatus().getState())
       << endl;
  {
    T2_QQPlan* qqplan = qenv.getQQPlan();
    t2_Assert (qqplan != NULL);
    t2_say_label(cout);
    cout << "# iters = " << qqplan->getNumberSubplans()
	 << ", cur_iter = " << qqplan->getCurrentIteration()
	 << ", cur_cluster = " << qqplan->getCurrentCluster()
	 << endl;
  }
  {
    u_int n0, n1;
    qenv.getNumberAccMsgRecvBuffers(n0, n1);
    t2_say_label(cout);
    cout << "query " << i << " (" << (int) &qenv
         << ") # acc msg recv buf = " << n0 << " " << n1 << endl;
  }
  T2_AccMsgRecvBufList** recvbufptr = qenv.getAccMsgRecvBufPointer();
  t2_Assert (recvbufptr != NULL);
  t2_Assert (recvbufptr[0] != NULL);
  t2_Assert (recvbufptr[1] != NULL);
  const T2_BitVector& bv0 = recvbufptr[0]->getBitVector();
  const T2_BitVector& bv1 = recvbufptr[1]->getBitVector();
  t2_say_label(cout);
  cout << "bitvectors = " << bv0 << " , " << bv1 << endl;

  T2_QStatus& status = qenv.getQueryStatus();
  t2_say_label(cout);
  cout << "nclusters_yetprocess = "
       << status.getNumberClustersYetProcess() << endl;
  t2_say_label(cout);
  cout << "naccmsgs_yetprocess = "
       << status.getNumberAccMsgsYetProcess() << endl;
  t2_say_label(cout);
  cout << "naccsends_yetcomplete = "
       << status.getNumberAccSendsYetComplete() << endl;
  t2_say_label(cout);
  cout << "noutputwrites_yetcomplete = "
       << status.getNumberOutputWritesYetComplete() << endl;
}



//////////////////////////////////////////////////////////////////
//
// issue more disk reads when desired
//
//////////////////////////////////////////////////////////////////
void
issueMoreReads(T2_QPlan& qplan, T2_LIO& lio, T2_WorkList& wlist)
{
  // these static variables are used to remember the cluster request
  // whose block requests had not all been issued last time due to
  // insufficient resource (io control block, buffer space, etc);
  // cur_block remembers the first block request of the cluster request
  // that has not been issued (ie where we left off last time)
  static T2_ClusterRequest* cluster = NULL;
  static u_int cur_block = 0;


double tsnaps[6];
tsnaps[0] = t2_dclock();



  // tentative policy: 
  // 1. only issue more reads if number of pending reads and ready reads
  //    in lio is below some low watermark (low_watermark)
  // 2. issue no more than a certain number of reads (max_nreads)
  const u_int low_watermark = 8;
  const u_int max_nreads = 4;

  if (lio.getNumberPendingCB() + lio.getNumberDoneCB() > low_watermark) {

//if (niter > maxIter - 2)
//cout << "iter " << niter
//<< "-- return due to #pending=" << lio.getNumberPendingCB()
//<< "+#done=" << lio.getNumberDoneCB() << " > low_mark=" << low_watermark
//<< endl;

     return;
  }


tsnaps[1] = t2_dclock();


  u_int readcnt = lio.getNumberReadyCB();

  // register reads with lio first
  while (!wlist.isFull() &&	// work list is not full
	 readcnt < max_nreads
  )
  {
     // get a cluster request
     if (cluster == NULL) {	// no unfinished cluster from last time

//if (niter > maxIter - 2)
//cout << "iter " << niter
//<< "-- last cluster = NULL" << endl;

	cluster = qplan.getOneClusterRequest();
	if (cluster == NULL) {	// no cluster request to issue for now

//if (firsttimenull==true) {
//firsttimenull=false;
//cout << "iter " << niter
//<< "-- getOneClusterRequest returns NULL" << endl;
//}

	   break;
	}

	t2_Assert (cluster->getRefCount() == cluster->getNumberBlockRequests());
	t2_Assert (cluster->getRefCount() > 0);
     }


//if (niter > maxIter - 2)
//cout << "iter " << niter
//<< "-- getOneClusterRequest returns a cluster" << endl;


     // register disk reads when possible; registerDiskReads should update
     // cur_block when it steps through the block requests in cluster;
     // registerDiskReads also returns the number of reads registered
     readcnt += registerDiskReads(*cluster, cur_block, lio, wlist);

     if (cur_block == cluster->getNumberBlockRequests()) {
	// all the block requests for the cluster have been inspected
	if (cluster->getRefCount() == 0) {	// are all blocks available?
	   // lucky cluster, all blocks are already available;
	   // add this cluster to the work list
	   bool r = wlist.enqueueCluster(cluster);
	   assert (r == true);
	}

	cluster = NULL;
	cur_block = 0;
     }
  }				// while loop


//if (niter > maxIter - 2)
//cout << "iter " << niter
//<< "-- left while loop: wlist.isFull=" << (int) wlist.isFull()
//<< ", readcnt=" << readcnt << ", max_nreads=" << max_nreads << endl;


  if (lio.getNumberReadyCB() > 0) {

//totnumreadyreads += lio.getNumberReadyCB();
//totnumpendingreads += lio.getNumberPendingCB();
//numlioissue++;

     T2_AsyncResult r = lio.issueReadyRequests();
     if (r != aSUCCEED) {
        t2_err_quit(__FILE__, __LINE__, "lio.issueReadyRequests returns %d",
		   (int) r);
     }
  }

}



u_int
registerDiskReads(T2_ClusterRequest& cluster, u_int& cur_block,
		  T2_LIO& lio, T2_WorkList& wlist)
{
  const u_int nblocks = cluster.getNumberBlockRequests();
  t2_Assert (cur_block < nblocks);

  u_int cnt = 0;	// count how many block requests were registered

  do {
    T2_PhyBlockRequest* blockreq = cluster.getBlockRequest(cur_block);
    t2_Assert (blockreq != NULL);

    // touch the block request
    T2_PhyBlockRequestStatus status = blockreq->touch(cluster.getAddress());

    switch (status) {
	case t2_bIdle:
			// touch failed to allocate buffer space -- maybe due
			// to insufficient buffer space; nothing can be done
			// further
			return cnt;

	case t2_bActivated:
			// a newly activated block, need to register with LIO
			if (lio.registerRead(
					T2_LIOKey(t2_lioBlockRead, blockreq),
					*blockreq) == false)
			{
			   // registerRead failed -- maybe due to lack of free
			   // io control block, so we'll try it again next time
			   return cnt;
			}
			// registerRead succeeded, lio.registerRead() should
			// have set the status to t2_bPending
			t2_Assert (blockreq->getStatus() == t2_bPending);

			cnt++;		// one more block request registered
			break;		// go on to next block request

	case t2_bPending:
			// a block request that had been issued but not
			// yet completed; do nothing for this block request,
			// and move on to the next block request
			break;

	case t2_bAvailable:
			// a block request that had been issued and completed,
			// with data available in its buffer;
			// decrement refcount of cluster, and go on to the
			// next block request
			cluster.decRefCount();
			break;

	case t2_bCached:
	default:	t2_err_quit(__FILE__, __LINE__,
				   "touch() returns unexpected status = %d",
				   (int) status);
			break;
    }

    cur_block++;	// move on to the next block request
  } while (cur_block < nblocks);


//if (niter > maxIter - 2)
//cout << "iter " << niter
//<< "-- registerDiskReads returns " << cnt << endl;


  return cnt;
}



/////////////////////////////////////////////////////////////////////
//
// process a retrieved input data chunk
//
/////////////////////////////////////////////////////////////////////
//void
//process_input_data_chunk(T2_DatasetEnv& dsenv, const T2_Cluster& input,
//			 const T2_Region& oqw, T2_UsrArgReader& meta,
//			 const T2_Box& mbr)
//{
//  T2_QEnv* qenv = dsenv.getQEnv();
//  t2_Assert (qenv != NULL);
//  T2_Dataset* dataset = dsenv.getDatasetPtr();
//  t2_Assert (dataset != NULL);
//  T2_ProjectFuncObj *proj = dsenv.getProjectFuncObj();
//  t2_Assert (proj != NULL);
//  T2_Accumulator *accptr = qenv->getAccumulator();
//  t2_Assert (accptr != NULL);
//  T2_Accumulator& acc = *accptr;
//  T2_AggregateFuncObj* aggr = qenv->getAggregateFuncObject();
//  t2_Assert (aggr != NULL);
//  const T2_Box& iqr = dsenv.getQueryRegion();
//  const T2_DSID dsid = dsenv.getDatasetID();
//
//  T2_Iterator *iter = dataset->genIterator(dsenv.getIteratorID(), input,
//					   mbr, iqr, oqw, *proj, meta);
//  while (!iter->endOfData(input)) {
//     T2_Point pt = iter->getGlobalCoords(input);
//     T2_Region rg;
//     if (proj->project(iqr, oqw, pt, rg)) {
//        T2_AccIterator *acciter = acc.navigate(qenv->getAccNavigatorID(), rg);
//
//        const void *input_elem = iter->getElement(input);
//        while (!acciter->endOfData(acc)) {
//	   void *acc_elem = acciter->getElement(acc);
//	   aggr->dafElem(dsid, input_elem, acc_elem);
//	   acciter->advance(acc);
//        }
//        delete acciter;
//     }
//
//     iter->advance(input);
//  }
//  delete iter;
//}



/////////////////////////////////////////////////////////////////////
//
// poll for disk reads
//
// This is where each completed block request has its status set
// from pending to available, and notifies its current owning
// cluster requests.
//
/////////////////////////////////////////////////////////////////////
void
pollDiskReads(T2_LIO& lio, T2_WorkList& wlist, T2_QPlan& qplan)
{
  lio.poll();

  T2_LIOKey key;
  int s;

//  while (lio.getDoneKey(key, s) == true) {
  while (!wlist.isFull() && lio.getDoneKey(key, s) == true) {
     t2_Assert (key.getPtr() != NULL);
//     t2_Assert (s > 0);

     switch (key.getType()) {
	case t2_lioBlockRead:
			blockReadCompletes((T2_PhyBlockRequest *) key.getPtr(),
					   s, wlist);
			break;
	case t2_lioOutputWrite:
			outputWriteCompletes((T2_OutputHandle *) key.getPtr(),
					     s, wlist, qplan);
			break;
	default:
			t2_err_quit(__FILE__, __LINE__,
				   "unexpected T2_LIOType = %d",
				   (int) key.getType());
     }
  }
}


void
blockReadCompletes(T2_PhyBlockRequest* blockreq, int sz, T2_WorkList& wlist)
{
  t2_Assert (blockreq != NULL);
  blockreq->becomeAvailable();
  blockreq->getBuffer().setBufferSize(sz);

  // need to notify each current owner recorded in blockreq
  const T2_PhyBlockOwnerInfo& ownerinfo = blockreq->getOwnerInfo();
  const u_int nowners = ownerinfo.getNumberCurrentOwners();
  for (u_int i=0; i<nowners; i++) {
	T2_ClusterRequest* cluster = ownerinfo.getCurrentOwner(i);
	t2_Assert (cluster != NULL);
	t2_Assert (cluster->getRefCount() > 0);
	if (cluster->decRefCount() == 0) {	// all blocks are available
	   bool r = wlist.enqueueCluster(cluster);
	   assert (r == true);
	}
  }
}


void
outputWriteCompletes(T2_OutputHandle* ohandle, int sz, T2_WorkList& wlist,
		     T2_QPlan& qplan)
{
  t2_Assert (ohandle != NULL);
  if (ohandle->issueNextOutputWrite() == true) {   // another write is issued
     return;
  }

  // all writes had been issued, so check for end of output handling phase
  T2_QEnv* qenv = ohandle->getQEnv();
  allOutputWritesDone(qenv, wlist);
}

