
/* $Id: t2_bfile.h,v 1.13 2000/01/10 18:02:52 chialin Exp chialin $ */

/* t2_bfile.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 4/3/98
 * Adapted from the code in Titan written by Carter T. Shock
 */

#ifndef _T2_BFILE_H_
#define _T2_BFILE_H_


#include	<unistd.h>
#include	"t2_header.h"
#include	"t2_bstream.h"


/////////////////////////////////////////////////////////////
//
// T2_BFile
//
// A base class for a binary file.
//
/////////////////////////////////////////////////////////////
class T2_BFile {
public:
protected:
  T2_BFile() { fd = errval = -1; };
  T2_BFile(const char *name, int flags, int mode = 0);
  ~T2_BFile() { bclose(); };

  T2_BFile & bopen(const char *name, int flags, int mode = 0);
  T2_BFile & attach(int f);
  int bclose();

  // dealing with various flags
  int getibflags() const { return dflt_ib_flg; };
  int getobflags() const { return dflt_ob_flg; };
  void setibflags(int flags) { dflt_ib_flg = flags; };
  void setobflags(int flags) { dflt_ob_flg = flags; };
  int getfd() const { return fd; };

  int getflags(); // not const because errval can change
  int setflags(int flags);

  int berr() const;

protected:
  int fd;
  int errval;
  bool attached;		// was the file open by attach()?

  static int dflt_ib_flg;
  static int dflt_ob_flg;

  // is any file open ?
  bool noFile() const { return (fd < 0)? true:false; };

private:
};



/////////////////////////////////////////////////////////////
//
// T2_IBFile
//
// A base class for an input binary file.
// This class does not deal with different representations
// across different platforms.
//
/////////////////////////////////////////////////////////////
//const u_int T2_IBFile_buflen = 1024*8;// Chialin Chang: buffer size = 8K
const u_int T2_IBFile_buflen = 4;	// Chialin Chang: buffer size = 8K
class T2_IBFile : protected T2_BFile, public T2_IBStream {
private:
protected:
  char fbuf[T2_IBFile_buflen];		// Chialin Chang: some new members
  u_int cur, len;			// for fast read

  bool saweof;				// a flag to tell if bread has seen
					// eof

  off_t pos;				// current actual file position,
					// measured in bytes from the beginning
					// of the file after filling the read
					// buffer; only modified when the buffer
					// needs to be filled

  off_t curToOffset() const { return pos - (len-cur);};
					// map wherever cur is pointing at to
					// the corresponding file position
					// measured in bytes from the beginning
					// of the file

  // virtual method
  void setErr(int e) { errval = e; };

public:
  T2_IBFile();
  T2_IBFile(const char *name, int mode = 0);

  T2_IBFile & open(const char *name, int mode = 0);
  T2_IBFile & attach(int fd);
  int close();

  off_t bseek(off_t offset, int whence = SEEK_SET);
  off_t btell() { return curToOffset(); };  
  ssize_t bread(void *buf, int count, size_t size);
  operator const void*() const;
  int getfd() const { return T2_BFile::getfd(); };
  int berr() const { return T2_BFile::berr(); };
  bool eof() const;

  T2_IBStream & operator >> (bool &c) ;

  T2_IBStream & operator >> (char &c) ;
  T2_IBStream & operator >> (signed char &c) ;
  T2_IBStream & operator >> (unsigned char &c) ;

  T2_IBStream & operator >> (short &s) ;
  T2_IBStream & operator >> (unsigned short &s) ;

  T2_IBStream & operator >> (int &i);
  T2_IBStream & operator >> (unsigned int &i);

  T2_IBStream & operator >> (long &l);
  T2_IBStream & operator >> (unsigned long &l);

  T2_IBStream & operator >> (float &f);
  T2_IBStream & operator >> (double &d);

  T2_IBStream & operator >> (char* s) { return T2_IBStream::operator >> (s); };
};


inline
T2_IBFile::T2_IBFile():T2_BFile(),T2_IBStream(false)
{
  len = cur = 0; pos = 0; saweof = false;
}

inline
T2_IBFile::T2_IBFile(const char *name, int mode): T2_BFile(),T2_IBStream(false)
{
  bopen(name,dflt_ib_flg,mode);
  len = cur = 0;
  pos = 0;
  saweof = false;
}

inline T2_IBFile&
T2_IBFile::open(const char *name, int mode)
{
  bopen(name,dflt_ib_flg,mode);
  len = cur = 0;
  pos = 0;
  return *this;
}

inline T2_IBFile&
T2_IBFile::attach(int f)
{
  len = 0;
  pos = 0;
  cur = 0;

  if (f >= 0) {
     T2_BFile::attach(f);
     cur = lseek(f, (off_t)0, SEEK_CUR);
  }
  return *this;
}

inline int
T2_IBFile::close()
{
  len = cur = 0;
  pos = 0;
  saweof = false;
  return T2_BFile::bclose();
}

inline T2_IBStream&
T2_IBFile::operator >> (bool &b) 
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&b), 1, sizeof(bool));
  t2_Assert(r == sizeof(bool));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &b;
     t2_swapBytes(bb, sizeof(b));
  }

  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (char &c) 
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&c), 1, sizeof(char));
  t2_Assert(r == sizeof(char));
  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (signed char &c) 
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&c), 1, sizeof(signed char));
  t2_Assert(r == sizeof(signed char));
  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (unsigned char &c) 
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&c), 1, sizeof(unsigned char));
  t2_Assert(r == sizeof(unsigned char));
  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (short &s) 
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&s), 1, sizeof(short));
  t2_Assert(r == sizeof(short));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &s;
     t2_swapBytes(bb, sizeof(s));
  }

  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (unsigned short &s) 
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&s), 1, sizeof(unsigned short));
  t2_Assert(r == sizeof(unsigned short));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &s;
     t2_swapBytes(bb, sizeof(s));
  }

  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (int &i)
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&i), 1, sizeof(int));
  t2_Assert(r == sizeof(int));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &i;
     t2_swapBytes(bb, sizeof(i));
  }

  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (unsigned int &i)
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&i), 1, sizeof(unsigned int));
  t2_Assert(r == sizeof(unsigned int));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &i;
     t2_swapBytes(bb, sizeof(i));
  }

  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (long &l)
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&l), 1, sizeof(long));
  t2_Assert(r == sizeof(long));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &l;
     t2_swapBytes(bb, sizeof(l));
  }

  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (unsigned long &l)
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&l), 1, sizeof(unsigned long));
  t2_Assert(r == sizeof(unsigned long));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &l;
     t2_swapBytes(bb, sizeof(l));
  }

  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (float &f)
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&f), 1, sizeof(float));
  t2_Assert(r == sizeof(float));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &f;
     t2_swapBytes(bb, sizeof(f));
  }

  return *this;
}

inline T2_IBStream&
T2_IBFile::operator >> (double &d)
{
# ifdef T2_ASSERTION
     int r =
# endif
  bread((void *)(&d), 1, sizeof(double));
  t2_Assert(r == sizeof(double));

  if (T2_IBStream::convertEndian() == true) {
     char *bb = (char *) &d;
     t2_swapBytes(bb, sizeof(d));
  }

  return *this;
}

inline
T2_IBFile::operator const void*() const
{
  if (noFile() || ((saweof && cur == len)))
     return 0;
  return this;
}



/////////////////////////////////////////////////////////////
//
// T2_OBFile
//
// A base class for an output binary file.
// This class does not deal with different representations
// across different platforms.
//
/////////////////////////////////////////////////////////////
class T2_OBFile : protected T2_BFile, public T2_OBStream {
private:
protected:
  // virtual method
  void setErr(int e) { errval = e; };

public:
  T2_OBFile(bool cvtEnd=false): T2_BFile(), T2_OBStream(cvtEnd) {};
  T2_OBFile(const char *name, int mode = 0644): T2_BFile(), T2_OBStream(false)
    { bopen(name,dflt_ob_flg,mode); };

  T2_OBFile & open(const char *name, int mode = 0644);
  T2_OBFile & attach(int fd);
  int close() { return T2_BFile::bclose(); };

  off_t bseek(off_t offset, int whence = SEEK_SET);
  off_t btell();
  ssize_t bwrite(const void *buf, int count, size_t size);
  operator const void*() const { return (noFile())? 0:this; };
  int getfd() const { return T2_BFile::getfd(); };
  int berr() const { return T2_BFile::berr(); };

  //  int btrunc(long length);

  T2_OBStream & operator << (bool b);

  T2_OBStream & operator << (char c);
  T2_OBStream & operator << (signed char c);
  T2_OBStream & operator << (unsigned char c) ;

  T2_OBStream & operator << (short s);
  T2_OBStream & operator << (unsigned short s);

  T2_OBStream & operator << (int i);
  T2_OBStream & operator << (unsigned int i);

  T2_OBStream & operator << (long l);
  T2_OBStream & operator << (unsigned long l);

  T2_OBStream & operator << (float f);
  T2_OBStream & operator << (double d);

  T2_OBStream & operator << (char* s) { return T2_OBStream::operator << (s); };
};



inline T2_OBFile&
T2_OBFile::open(const char *name, int mode)
{
  bopen(name,dflt_ob_flg,mode);
  return *this;
}

inline T2_OBFile&
T2_OBFile::attach(int fd)
{
  T2_BFile::attach(fd);
  return *this;
}


inline T2_OBStream&
T2_OBFile::operator << (bool b)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&b), 1, sizeof(bool));
  t2_Assert(r == sizeof(bool));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (char c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&c), 1, sizeof(char));
  t2_Assert(r == sizeof(char));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (signed char c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&c), 1, sizeof(signed char));
  t2_Assert(r == sizeof(signed char));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (unsigned char c) 
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&c), 1, sizeof(unsigned char));
  t2_Assert(r == sizeof(unsigned char));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (short s)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&s), 1, sizeof(short));
  t2_Assert(r == sizeof(short));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (unsigned short s)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&s), 1, sizeof(unsigned short));
  t2_Assert(r == sizeof(unsigned short));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (int i)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&i), 1, sizeof(int));
  t2_Assert(r == sizeof(int));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (unsigned int i)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&i), 1, sizeof(unsigned int));
  t2_Assert(r == sizeof(unsigned int));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (long l)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&l), 1, sizeof(long));
  t2_Assert(r == sizeof(long));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (unsigned long l)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&l), 1, sizeof(unsigned long));
  t2_Assert(r == sizeof(unsigned long));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (float f)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&f), 1, sizeof(float));
  t2_Assert(r == sizeof(float));
  return *this;
}

inline T2_OBStream&
T2_OBFile::operator << (double d)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&d), 1, sizeof(double));
  t2_Assert(r == sizeof(double));
  return *this;
}


#endif
