
/* $Id: t2_bstr.h,v 1.10 2000/04/10 15:55:03 chialin Exp $ */

/* t2_bstr.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 2/14/98
 * strings for holding numbers/char in binary formats
 */

#ifndef _T2_BSTR_H_
#define _T2_BSTR_H_


#include	<strings.h>
#include	"t2_header.h"


/////////////////////////////////////////////////////////////
//
// T2_OBStr
//
// A base class for an output binary string.
//
/////////////////////////////////////////////////////////////
class T2_OBStr {
public:
  T2_OBStr(): buf(NULL), cur(NULL), len(0) {};
  T2_OBStr(char *b, const size_t l);
  ~T2_OBStr();

  void open(char *b, const size_t l);
  void close();

  // write into the buffer
  size_t bwrite(const void* c, u_int count, size_t size);

  // query methods
  size_t getBufferSize() const { return len; };
  const char* getCurrentPosition() const { return cur; };
  bool isFull() const { return (getNumberBytesAhead() == 0)? true:false; };
  size_t getNumberBytesUnwritten() const { return getNumberBytesAhead(); };
  size_t getNumberBytesWritten() const { return getNumberBytesPassed(); };
  operator const void*() const
		{ if (getNumberBytesAhead() == 0) return 0; return this; };

  // print all the bytes that have been written into the binary string
  // (for debugging purpose)
  ostream& writeBytesInBuffer(ostream& os) const;

protected:
  char* buf, *cur;
  size_t len;

  // primitive query methods
  size_t getNumberBytesPassed() const { return cur - buf; };
  size_t getNumberBytesAhead() const { return len - getNumberBytesPassed(); };

private:
  // do not use copy constructor or operator=
  T2_OBStr(const T2_OBStr& a);
  T2_OBStr& operator =(const T2_OBStr& a);

};


inline
T2_OBStr::T2_OBStr(char *b, const size_t l): buf(b), cur(b), len(l)
{
  if (b == NULL) len = 0;
}

inline
T2_OBStr::T2_OBStr(const T2_OBStr& a): buf(NULL), cur(NULL), len(a.len)
{
  t2_err_quit(__FILE__,__LINE__,"do not use copy constructor");
}

inline T2_OBStr&
T2_OBStr::operator =(const T2_OBStr& a)
{
  len = a.len;
  t2_err_quit(__FILE__,__LINE__,"do not use operator =");
  return *this;
}

inline void
T2_OBStr::open(char *b, const size_t l)
{
  t2_Assert (buf == NULL);

  if (b == NULL) {
     cur = NULL;
     len = 0;
     return;
  }
  buf = cur = b;
  len = l;
}

inline void
T2_OBStr::close()
{
  buf = cur = NULL;
  len = 0;
}

inline
T2_OBStr::~T2_OBStr()
{
  close();
}

inline size_t
T2_OBStr::bwrite(const void* c, u_int count, size_t size)
{
  t2_Assert ((buf == NULL && cur == NULL) || (buf != NULL && cur != NULL));
  size_t totsize = count * size;
  if (getNumberBytesAhead() < totsize)
     totsize = getNumberBytesAhead();
  bcopy(c, (void *)cur, totsize);
  cur += totsize;
  return totsize;
}

inline ostream&
T2_OBStr::writeBytesInBuffer(ostream& os) const
{
  const char *p = buf;
  while (p != cur) {
    os << " " << (int) *p;
    p++;
  }
  return os;
}



/////////////////////////////////////////////////////////////
//
// T2_FOBStr
//
// The formatted output binary string that does not deal
// with different representations for numbers. That is, the
// class does not deal with heterogeneity at all, so should
// be more efficient.
//
/////////////////////////////////////////////////////////////
class T2_FOBStr: public T2_OBStr {
public:
  T2_FOBStr(): T2_OBStr() {};
  T2_FOBStr(char *b, const size_t l): T2_OBStr(b, l) {};
  ~T2_FOBStr() {};

  T2_FOBStr& operator << (bool c);

  T2_FOBStr& operator << (char c);
  T2_FOBStr & operator << (signed char c);
  T2_FOBStr & operator << (unsigned char c);

  T2_FOBStr & operator << (short s);
  T2_FOBStr & operator << (unsigned short s);

  T2_FOBStr & operator << (int i);
  T2_FOBStr & operator << (unsigned int i);

  T2_FOBStr & operator << (long l);
  T2_FOBStr & operator << (unsigned long l);

  T2_FOBStr & operator << (float f);
  T2_FOBStr & operator << (double d);

  // print all the bytes that have been written into the binary string
  // (for debugging purpose)
  ostream& writeBytesInBuffer(ostream& os) const
				{ return T2_OBStr::writeBytesInBuffer(os); };

protected:
private:
  // do not use copy constructor and operator =
  T2_FOBStr(const T2_FOBStr& a);
  T2_FOBStr& operator =(const T2_FOBStr& a);

};


inline
T2_FOBStr::T2_FOBStr(const T2_FOBStr&)
{
  t2_err_quit(__FILE__,__LINE__,"do not use copy constructor");
}

inline T2_FOBStr&
T2_FOBStr::operator =(const T2_FOBStr&)
{
  t2_err_quit(__FILE__,__LINE__,"do not use operator =");
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (bool c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&c), 1, sizeof(bool));
  t2_Assert(r == sizeof(bool));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (char c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&c), 1, sizeof(char));
  t2_Assert(r == sizeof(char));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (signed char c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&c), 1, sizeof(signed char));
  t2_Assert(r == sizeof(signed char));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (unsigned char c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&c), 1, sizeof(unsigned char));
  t2_Assert(r == sizeof(unsigned char));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (short s)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&s), 1, sizeof(short));
  t2_Assert(r == sizeof(short));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (unsigned short s)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&s), 1, sizeof(unsigned short));
  t2_Assert(r == sizeof(unsigned short));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (int i)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&i), 1, sizeof(int));
  t2_Assert(r == sizeof(int));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (unsigned int i)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&i), 1, sizeof(unsigned int));
  t2_Assert(r == sizeof(unsigned int));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (long l)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&l), 1, sizeof(long));
  t2_Assert(r == sizeof(long));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (unsigned long l)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&l), 1, sizeof(unsigned long));
  t2_Assert(r == sizeof(unsigned long));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (float f)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&f), 1, sizeof(float));
  t2_Assert(r == sizeof(float));
  return *this;
}

inline T2_FOBStr&
T2_FOBStr::operator << (double d)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bwrite((void *)(&d), 1, sizeof(double));
  t2_Assert(r == sizeof(double));
  return *this;
}



/////////////////////////////////////////////////////////////
//
// T2_IBStr
//
// A base class for both the input and output binary strings.
//
/////////////////////////////////////////////////////////////
class T2_IBStr {
public:
  T2_IBStr(): buf(NULL), cur(NULL), len(0) {};
  T2_IBStr(const char *b, const size_t l);
  ~T2_IBStr();

  void open(const char *b, const size_t l);
  void close();
  void rewind() { cur = buf; };

  // read from the buffer
  size_t bread(void* c, u_int count, size_t size);

  // query methods
  size_t getBufferSize() const { return len; };
  const char* getCurrentPosition() const { return cur; };
  bool isEmpty() const { return (getNumberBytesPassed()==0)? true:false; };
  size_t getNumberBytesUnread() const { return getNumberBytesAhead(); };
  size_t getNumberBytesRead() const { return getNumberBytesPassed(); };
  operator const void*() const
		{ if (getNumberBytesAhead() == 0) return 0; return this; };

  // move cur forward n bytes (or backward is n < 0);
  // return # bytes cur actually moved
  int skip(int n);

  // print all the bytes in the binary string
  // (for debugging purpose)
  ostream& writeBytesInBuffer(ostream& os) const;

protected:
  const char *buf, *cur;
  size_t len;

  // primitive query methods
  size_t getNumberBytesPassed() const { return cur - buf; };
  size_t getNumberBytesAhead() const { return len - getNumberBytesPassed(); };

private:
  // do not use copy constructor or operator=
  T2_IBStr(const T2_IBStr& a);
  T2_IBStr& operator =(const T2_IBStr& a);

};


inline
T2_IBStr::T2_IBStr(const char *b, const size_t l): buf(b), cur(b), len(l)
{
  if (b == NULL) len = 0;
}

inline
T2_IBStr::T2_IBStr(const T2_IBStr&)
{
  t2_err_quit(__FILE__,__LINE__,"do not use copy constructor");
}

inline T2_IBStr&
T2_IBStr::operator =(const T2_IBStr&)
{
  t2_err_quit(__FILE__,__LINE__,"do not use operator =");
  return *this;
}

inline void
T2_IBStr::open(const char *b, const size_t l)
{
  t2_Assert (buf == NULL);

  if (b == NULL) {
     cur = NULL;
     len = 0;
     return;
  }
  buf = cur = b;
  len = l;
}

inline void
T2_IBStr::close()
{
  buf = cur = NULL;
  len = 0;
}

inline
T2_IBStr::~T2_IBStr()
{
  close();
}

inline size_t
T2_IBStr::bread(void* c, u_int count, size_t size)
{
  size_t totsize = count * size;
  if (getNumberBytesAhead() < totsize)
     return 0;
  bcopy((const void *)cur, c, totsize);
  cur += totsize;
  return totsize;
}

inline int
T2_IBStr::skip(int n)
{
  u_int maxn = getNumberBytesAhead();
  if (n >= (int) maxn) {
     cur = buf + len;
     return maxn;
  }
  if (n >= 0) {
     cur += n;
     return n;
  }

  maxn = getNumberBytesPassed();
  u_int pn = 0 - n;				// pn is positive n
  if (pn >= maxn) {
     cur = buf;
     return -maxn;
  }
  cur -= pn;
  return n;
}

inline ostream&
T2_IBStr::writeBytesInBuffer(ostream& os) const
{
  const char *p = buf;
  for (size_t i=0; i<len; i++) {
    os << " " << (int) *p;
    p++;
  }
  return os;
}



/////////////////////////////////////////////////////////////
//
// T2_FIBStr
//
// The input binary string that comes from a machine with
// the same representations for numbers. That is, the
// class does not deal with heterogeneity at all, so should
// be more efficient.
//
/////////////////////////////////////////////////////////////
class T2_FIBStr: public T2_IBStr {
public:
  T2_FIBStr(): T2_IBStr() {};
  T2_FIBStr(const char *b, const size_t l): T2_IBStr(b, l) {};
  ~T2_FIBStr() {};

  T2_FIBStr& operator >> (bool &c);

  T2_FIBStr& operator >> (char &c);
  T2_FIBStr & operator >> (signed char &c);
  T2_FIBStr & operator >> (unsigned char &c);

  T2_FIBStr & operator >> (short &s);
  T2_FIBStr & operator >> (unsigned short &s);

  T2_FIBStr & operator >> (int &i);
  T2_FIBStr & operator >> (unsigned int &i);

  T2_FIBStr & operator >> (long &l);
  T2_FIBStr & operator >> (unsigned long &l);

  T2_FIBStr & operator >> (float &f);
  T2_FIBStr & operator >> (double &d);

  // print all the bytes in the binary string
  // (for debugging purpose)
  ostream& writeBytesInBuffer(ostream& os) const
				{ return T2_IBStr::writeBytesInBuffer(os); };

protected:
private:
  // do not use copy constructor and operator =
  T2_FIBStr(const T2_FIBStr& a);
  T2_FIBStr& operator =(const T2_FIBStr& a);

};


inline
T2_FIBStr::T2_FIBStr(const T2_FIBStr&)
{
  t2_err_quit(__FILE__,__LINE__,"do not use copy constructor");
}

inline T2_FIBStr&
T2_FIBStr::operator =(const T2_FIBStr&)
{
  t2_err_quit(__FILE__,__LINE__,"do not use operator =");
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (bool &c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&c), 1, sizeof(bool));
  t2_Assert(r == sizeof(bool));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (char &c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&c), 1, sizeof(char));
  t2_Assert(r == sizeof(char));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (signed char &c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&c), 1, sizeof(signed char));
  t2_Assert(r == sizeof(signed char));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (unsigned char &c)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&c), 1, sizeof(unsigned char));
  t2_Assert(r == sizeof(unsigned char));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (short &s)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&s), 1, sizeof(short));
  t2_Assert(r == sizeof(short));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (unsigned short &s)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&s), 1, sizeof(unsigned short));
  t2_Assert(r == sizeof(unsigned short));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (int &i)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&i), 1, sizeof(int));
  t2_Assert(r == sizeof(int));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (unsigned int &i)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&i), 1, sizeof(unsigned int));
  t2_Assert (r == sizeof(unsigned int));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (long &l)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&l), 1, sizeof(long));
  t2_Assert(r == sizeof(long));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (unsigned long &l)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&l), 1, sizeof(unsigned long));
  t2_Assert(r == sizeof(unsigned long));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (float &f)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&f), 1, sizeof(float));
  t2_Assert(r == sizeof(float));
  return *this;
}

inline T2_FIBStr&
T2_FIBStr::operator >> (double &d)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  bread((void *)(&d), 1, sizeof(double));
  t2_Assert(r == sizeof(double));
  return *this;
}


#endif	/* _T2_BSTR_H_ */
