
/* t2_packets.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/25/98
 * T2 packets and connections between front-end and back-end
 */


#ifndef _T2_PACKETS_H_
#define _T2_PACKETS_H_


#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_socket.h"
#include	"t2_nbsocket.h"
#include	"t2_qspec.h"


//////////////////////////////////////////////////////////////////////////////
//
// Tags to identify various packets exchanged between front-end and back-end.
//
//////////////////////////////////////////////////////////////////////////////
const u_char
	T2_ErrorFEBEPacketTag		= 30,		// unknown tag
	T2_ConnectPacketTag		= 33,		// be connects to fe
	T2_AcceptPacketTag		= 34,		// fe accepts be
	T2_QueryPacketTag		= 35,		// fe sends query to be
	T2_ReadyPacketTag		= 36,		// be tells be ready
	T2_QueryAckTag			= 37,		// be acks qbatch done
	T2_ShutdownPacketTag		= 40		// fe shuts down be
;


//////////////////////////////////////////////////////////////////////////////
//
// Tags to identify various packets exchanged between the back-end and the
// client.
//
//////////////////////////////////////////////////////////////////////////////
const u_char
	T2_BEClientByeTag		= 51
;


//////////////////////////////////////////////////////////////////////////////
//
// T2_FEBEPacket
//
// A base class for packets exchanged between front-end and back-end.
//
//////////////////////////////////////////////////////////////////////////////
class T2_FEBEPacket {
public:
  T2_FEBEPacket(u_char t): tag(t) {};
  ~T2_FEBEPacket() {};

  // accessing the tag
  u_char getTag() const { return tag; };
//  u_char setTag(int t) { return (tag = t); };

  // send and recv a whole packet (tag + body)
  bool send(T2_Socket& s) const;
  u_char recv(T2_Socket& s);			// returns the tag read

protected:
  const u_char tag;

private:
};



//////////////////////////////////////////////////////////////////////////////
//
// T2_ConnectPacket
//
// A packet that a back-end node uses to connect to the front-end process.
//
//////////////////////////////////////////////////////////////////////////////
class T2_ConnectPacket: public T2_FEBEPacket {
public:
  T2_ConnectPacket(int n=0, int m=0, u_char e=T2_UnknownEndian);
  ~T2_ConnectPacket() {};

  // access methods
  int getNumberNodes() const { return numnodes; };
  int getMyNode() const { return mynode; };
  u_char getMyEndian() const { return my_endian; };

  // send and recv a whole packet (tag + body)
  bool send(T2_Socket& s) const;
  u_char recv(T2_Socket& s, u_char fe_endian);

protected:
  int numnodes, mynode;
  u_char my_endian;

private:
};


inline
T2_ConnectPacket::T2_ConnectPacket(int n, int m, u_char e):
		T2_FEBEPacket(T2_ConnectPacketTag), numnodes(n), mynode(m),
		my_endian(e)
{
}



//////////////////////////////////////////////////////////////////////////////
//
// T2_AcceptPacket
//
// A packet that the front-end process uses to accept the connection of
// a back-end node after the front-end receives a T2_ConnectPacket.
//
//////////////////////////////////////////////////////////////////////////////
class T2_AcceptPacket: public T2_FEBEPacket {
public:
  T2_AcceptPacket(): T2_FEBEPacket(T2_AcceptPacketTag) {};
  ~T2_AcceptPacket() {};

  // send and recv a whole packet (tag + body)
  bool send(T2_Socket& s) const;
  u_char recv(T2_Socket& s);

protected:
private:
};



//////////////////////////////////////////////////////////////////////////////
//
// T2_ReadyPacket
//
// A packet that the back-end node uses to inform the front-end that it
// is ready for another batch of queries.
//
//////////////////////////////////////////////////////////////////////////////
class T2_ReadyPacket: public T2_FEBEPacket {
public:
  T2_ReadyPacket(): T2_FEBEPacket(T2_ReadyPacketTag) {};
  ~T2_ReadyPacket() {};

  // send and recv a whole packet (tag + body)
  bool send(T2_Socket& s) const;
  u_char recv(T2_Socket& s);

protected:
private:
};



//////////////////////////////////////////////////////////////////////////////
//
// T2_ShutdownPacket
//
// A packet that the front-end process uses to inform a back-end node to
// shut down.
//
//////////////////////////////////////////////////////////////////////////////
class T2_ShutdownPacket: public T2_FEBEPacket {
public:
  T2_ShutdownPacket(): T2_FEBEPacket(T2_ShutdownPacketTag) {};
  ~T2_ShutdownPacket() {};

  // send and recv a whole packet (tag + body)
  bool send(T2_Socket& s) const;
//  u_char recv(T2_Socket& s);

protected:
private:
};




//////////////////////////////////////////////////////////////////////////////
//
// T2_QueryAckPacket
//
// A packet that the back-end uses to inform the front-end process that
// a query batch has been processed.
//
//////////////////////////////////////////////////////////////////////////////
class T2_QueryAckPacket: public T2_FEBEPacket {
public:
  T2_QueryAckPacket(): T2_FEBEPacket(T2_QueryAckTag) {};
  ~T2_QueryAckPacket() {};


  // send and recv a whole packet (tag + body)
  bool send(T2_Socket& s) const;
  u_char recv(T2_Socket& s);

protected:
private:
};



//////////////////////////////////////////////////////////////////////////////
//
// T2_BEClientHelloPacket
//
// A packet sent by the back-end to the client at the time a socket is
// connected for the purpose of shipping out the final output. That is,
// this packet is only sent once during the lifetime of the connection
// between a back-end node and the client.
//
// Currently, each back-end node sends the following data to the client:
//	u_int numnodes		the number of back-end nodes that will 
//				connect to the client
//	u_int mynode		my rank in the back-end node processor group
//
//////////////////////////////////////////////////////////////////////////////
class T2_BEClientHelloPacket {
public:
  T2_BEClientHelloPacket(): numnodes(0), mynode(0) {};
  T2_BEClientHelloPacket(u_int n, u_int m): numnodes(n), mynode(m) {};
  ~T2_BEClientHelloPacket() {};

  // query methods
  u_int getNumberNode() const { return numnodes; };
  u_int getMyNode() const { return mynode; };

  // set data members
  u_int setNumberNode(u_int n) { return (numnodes = n); };
  u_int setMyNode(u_int m) { return (mynode = m); };

  // output or input through sockets
  bool send(T2_Socket& s) const;
  bool recv(T2_Socket& s);

protected:
  u_int numnodes, mynode;

private:
};


inline
T2_Socket& operator << (T2_Socket& s, const T2_BEClientHelloPacket& h)
{
  h.send(s);
  return s;
}

inline
T2_Socket& operator >> (T2_Socket& s, T2_BEClientHelloPacket& h)
{
  h.recv(s);
  return s;
}



//////////////////////////////////////////////////////////////////////////////
//
// T2_BEClientByePacket
//
// A packet sent by the back-end to the client at the time when all data has
// been sent the connection is about to be terminated.
//
// Currently, this packet only contains a tag, which should be deteced and
// interpretated by the receiving side of the socket connection (eg the
// client).
//
//////////////////////////////////////////////////////////////////////////////
class T2_BEClientByePacket {
public:
  T2_BEClientByePacket() {};
  ~T2_BEClientByePacket() {};

  // output or input through sockets
  bool send(T2_Socket& s) const;
  bool send(T2_nbSocket& s) const;
  bool recv(T2_Socket& s);

protected:
private:
};



//////////////////////////////////////////////////////////////////////////////
//
// T2_BEConnection
//
// A connection to a back-end node.
//
//////////////////////////////////////////////////////////////////////////////
class T2_BEConnection {
public:
  T2_BEConnection();
  ~T2_BEConnection();

  // do not use copy constructor nor operator =
  T2_BEConnection(const T2_BEConnection& c);
  T2_BEConnection& operator =(const T2_BEConnection& c);

  // access the socket
  const T2_Socket& getSocket() const { return sock; };
  T2_Socket& getSocket() { return sock; };

  // transfer from another T2_BEConnection
  void transfer(T2_BEConnection& c);

  // shut down connection
  void shutdown();

  // recv various packets;
  // returns the appropriate tag when succeeded, the recevied tag if recvs
  // unexpected tag, and T2_ErrorFEBEPacketTag otherwise
  int recvConnectPacket(T2_ConnectPacket& p, u_char my_endian)
					{ return p.recv(sock, my_endian); };

  // send various packets;
  // returns true when succeeds, false otherwise
  bool sendAcceptPacket(const T2_AcceptPacket& p) { return p.send(sock); };
  bool sendShutdownPacket(const T2_ShutdownPacket& p) { return p.send(sock); };

  // send a query batch to each back-end node, a blocking send
  bool sendQBatch(T2_QBatch& q);
  bool sendQBatch(T2_QPtrBatch& q);

protected:
  T2_Socket sock;

private:
};


inline
T2_BEConnection::T2_BEConnection(const T2_BEConnection& c)
{
  t2_err_quit(__FILE__, __LINE__,"T2_BEConnection do not use copy constructor");
}

inline T2_BEConnection&
T2_BEConnection::operator =(const T2_BEConnection& c)
{
  t2_err_quit(__FILE__, __LINE__, "T2_BEConnection do not use operator =");
  return *this;
}



//////////////////////////////////////////////////////////////////////////////
//
// T2_FEConnection
//
// A connection to a front-end process.
//
//////////////////////////////////////////////////////////////////////////////
class T2_FEConnection {
public:
  T2_FEConnection();
  ~T2_FEConnection();

  // do not use copy constructor nor operator =
  T2_FEConnection(const T2_FEConnection& c);
  T2_FEConnection& operator =(const T2_FEConnection& c);

  operator const void*() const { return (const void *) sock; };

  // access the socket
  const T2_Socket& getSocket() const { return sock; };
  T2_Socket& getSocket() { return sock; };

  // transfer from another T2_FEConnection
  void transfer(T2_FEConnection& c);

  // connect to T2 front-end
  bool connectFrontEnd(const char* hname, short port, int nprocs, int myproc,
			u_char endian);

  // shut down connection
  void shutdown();

  // recv various packets;
  // returns the appropriate tag when succeeded, the recevied tag if recvs
  // unexpected tag, and T2_ErrorFEFEPacketTag otherwise
  u_char recvAcceptPacket(T2_AcceptPacket& p) { return p.recv(sock); };
  u_char recvShutdownPacket(T2_ShutdownPacket& p) { return p.recv(sock); };

  // send various packets;
  // returns true when succeeds, false otherwise
  bool sendConnectPacket(const T2_ConnectPacket& p) { return p.send(sock); };
  bool sendQBatchAck(const u_int mynode);

  // recv a tag; return T2_ErrorFEBEPacketTag if error
  u_char recvPacketTag();

  // recv a query batch
  bool recvQBatch(T2_QBatch& q);

protected:
  T2_Socket sock;

private:
};


inline
T2_FEConnection::T2_FEConnection(const T2_FEConnection& c)
{
  t2_err_quit(__FILE__, __LINE__,"T2_FEConnection do not use copy constructor");
}

inline T2_FEConnection&
T2_FEConnection::operator =(const T2_FEConnection& c)
{
  t2_err_quit(__FILE__, __LINE__, "T2_FEConnection do not use operator =");
  return *this;
}



#endif	/* ! _T2_PACKETS_H_ */
