
/* t2_socket.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/10/98
 */


#ifndef _T2_SOCKET_H_
#define _T2_SOCKET_H_


#include	"t2_header.h"
#include	"t2_bstream.h"
#include	"t2_sockdefs.h"


////////
//
// Errors caused by the invocation of various methods will be stored in
// the data member, T2_Sock::errval, which can be retrieved by calling
// T2_Sock::getErrorValue(); errval = 0 means no error
// 
////////


// forward declaration
class T2_nbSocket;
class T2_SocketServer;

//////////////////////////////////////////////////////////////////////////
//
// T2_Sock
//
//////////////////////////////////////////////////////////////////////////
class T2_Sock {
public:
  ~T2_Sock();

  // query methods
  bool connected() const { return isconnect; };
  operator const void*() const { return (connected())? this : NULL; };
  int getFileDescriptor() const { return sockfd; };
  int getErrorValue() const { return errval; };

  // close a socket
  void closeSocket();

protected:
  int sockfd;
  int errval;
  bool isconnect;				// is connected ?

  // constructor 
  T2_Sock(): sockfd(-1), errval(0), isconnect(false) {};
  T2_Sock(const T2_Sock& s):
	sockfd(s.sockfd), errval(s.errval), isconnect(s.isconnect) {};

  // operator =
  T2_Sock& operator = (const T2_Sock& s)
	{ sockfd=s.sockfd; errval=s.errval; isconnect=s.isconnect;
	  return *this; };

  // open a socket
  bool openSocket(int family, int type, int protocol);

  // blocking write and read
  ssize_t bWrite(const void *buf,int count,size_t size,ssize_t& nbytes_wrote);
  ssize_t bRead(void *buf, int count, size_t size, ssize_t& nbytes_read);

private:
};


inline void
T2_Sock::closeSocket()
{
  if (sockfd >= 0) {
     close(sockfd);
     sockfd = -1;
  }
  isconnect = false;
}

inline
T2_Sock::~T2_Sock()
{
  closeSocket();
}

inline bool
T2_Sock::openSocket(int family, int type, int protocol)
{
  closeSocket();			// close old socket
  if ( (sockfd = socket(family, type, protocol)) < 0) {
     errval = errno;
     return false;
  }
  return true;
}




//////////////////////////////////////////////////////////////////////////
//
// T2_Socket
//
// The "client" version of a TCP socket.
//
//////////////////////////////////////////////////////////////////////////
class T2_Socket : public T2_Sock, public T2_IBStream, public T2_OBStream {
friend class T2_nbSocket;
friend class T2_SocketServer;
public:
  T2_Socket():  T2_Sock(), T2_IBStream(false), T2_OBStream(false),
		nbytes_read(0), nbytes_wrote(0) {};
  T2_Socket(bool cvtInpputEndian, bool cvtOutputEndian):
	        T2_Sock(), T2_IBStream(cvtInpputEndian),
		T2_OBStream(cvtOutputEndian),
		nbytes_read(0), nbytes_wrote(0) {};
  ~T2_Socket() {};

  // virtual functions from T2_BStream
  bool eof() const { return connected()? false : true; };
  operator const void*() const { return T2_Sock::operator const void*(); };
  int berr() const { return getErrorValue(); };

  // transfer a socket from a blocking socket s and detach s from that socket;
  void transfer(T2_Socket& s);

  // get/set the convertEndian flag for input or output socket;
  bool inputConvertEndian() const { return T2_IBStream::convertEndian(); }
  bool outputConvertEndian() const { return T2_OBStream::convertEndian(); }
  bool setInputConvertEndian(bool t)
				{ return T2_IBStream::setConvertEndian(t); };
  bool setOutputConvertEndian(bool t)
				{ return T2_OBStream::setConvertEndian(t); };

  //////
  // blocking connect, either by ip address or by hostname
  // Method returns true if connect succeeds, false otherwise with errval
  // set to errno:
  //	errval		reason
  //	-------		---------
  //	EADDRNOTAVAIL	ipaddr or hostname is NULL
  //	others		error occurred in some system call
  //////
  bool connectByAddress(const char* ipaddr, short port);
  bool connectByName(const char* hname, short port);

  //////
  // accessing peer's sockaddr_in, ip address, port
  // errval is set to one of the following possible values:
  //	errval		reason
  //	-------		---------
  //	ENOTCONN	socket is not connected
  //	EADDRNOTAVAIL	buf is NULL in getPeerAddress
  //	others		error occurred in some system call
  // getPeerPort returns -1 if it fails and set errval to ENOTCONN
  //////
  const struct sockaddr_in& getPeerSockAddress() const { return remote; };
  bool getPeerAddress(char* buf, size_t len);	// return false if fails
						// (check errval for reason)
  short getPeerPort();				// return -1 if not connected

  //////
  // blocking read and write;
  // they return the number of bytes written/read when succeeded, -1 if error;
  // however, since a blocking read() on Solaris often returns before the
  // specified number of bytes are read, for now I define two read methods:
  //	bread	a real blocking read that keeps calling read() until:
  //		(1) the specified number of bytes are read, in which case
  //		    bread returns the number of bytes read
  //		(2) some error occurs when invoking read, in which case
  //		    bread returns the number of bytes bread actually read,
  //		    and errval contains the error
  //		    (in particular, if buf to bread is NULL, errval is set
  //		    to EFAULT)
  //		(3) some error is detected during the first read() call, eg
  //		    the socket is not connected, in which case bread returns
  //		    -1, nbytes_read (ie the number of bytes read) is set to 0,
  //		    and errval contains the error number (eg ENOTCONN)
  //	nbread	a read function that only calls read() once and read as many
  // 		bytes as read() can read; it should behave
  //		the same way bread behaves if the operating system truely
  //		implements read() as a blocking read;
  //		NOTE THAT NO ENDIAN CONVERSION IS IMPLEMENTED FOR nbread()
  //////
  ssize_t bwrite(const void *buf, int count, size_t size);
  ssize_t bread(void *buf, int count, size_t size);
  ssize_t nbread(void *buf, int count, size_t size);
  ssize_t getNumberOfBytesLastRead() const { return nbytes_read; };
  ssize_t getNumberOfBytesLastWritten() const { return nbytes_wrote; };

  T2_OBStream & operator << (bool b);

  T2_OBStream & operator << (char c);
  T2_OBStream & operator << (signed char c);
  T2_OBStream & operator << (unsigned char c) ;

  T2_OBStream & operator << (short s);
  T2_OBStream & operator << (unsigned short s);

  T2_OBStream & operator << (int i);
  T2_OBStream & operator << (unsigned int i);

  T2_OBStream & operator << (long l);
  T2_OBStream & operator << (unsigned long l);

  T2_OBStream & operator << (float f);
  T2_OBStream & operator << (double d);

  T2_OBStream & operator << (char* s) { return T2_OBStream::operator << (s); };

  T2_IBStream & operator >> (bool &b) ;

  T2_IBStream & operator >> (char &c) ;
  T2_IBStream & operator >> (signed char &c) ;
  T2_IBStream & operator >> (unsigned char &c) ;

  T2_IBStream & operator >> (short &s) ;
  T2_IBStream & operator >> (unsigned short &s) ;

  T2_IBStream & operator >> (int &i);
  T2_IBStream & operator >> (unsigned int &i);

  T2_IBStream & operator >> (long &l);
  T2_IBStream & operator >> (unsigned long &l);

  T2_IBStream & operator >> (float &f);
  T2_IBStream & operator >> (double &d);

  T2_IBStream & operator >> (char* s) { return T2_IBStream::operator >> (s); };

protected:
  struct sockaddr_in remote;
  ssize_t nbytes_read,		// # of bytes last read
          nbytes_wrote;		// # of bytes last written

  // open a socket
  bool openSocket() { return T2_Sock::openSocket(AF_INET, SOCK_STREAM, 0); };

  // connect by sockaddr_in
  bool connectBySockAddr(struct sockaddr_in& serv_addr);

  // virtual method
   void setErr(int e) { errval = e; };

private:
  // don't use copy constructor and operator =
  T2_Socket(const T2_Socket& s);
  T2_Socket& operator =(const T2_Socket& s);

};


inline ssize_t
T2_Socket::bwrite(const void *buf, int count, size_t size)
{
  if (size > 1 && T2_OBStream::convertEndian() == true) {
     char *bb = (char *) buf;
     t2_swapBytesArray(bb, count, size);
  }

  ssize_t ss = bWrite(buf, count, size, nbytes_wrote);

  if (size > 1 && T2_OBStream::convertEndian() == true) {
     char *bb = (char *) buf;
     t2_swapBytesArray(bb, count, size);
  }
  return ss;
}

inline ssize_t
T2_Socket::bread(void *buf, int count, size_t size)
{
  ssize_t ss = bRead(buf, count, size, nbytes_read);

  if (size > 1 && T2_IBStream::convertEndian() == true) {
     char *bb = (char *) buf;
     t2_swapBytesArray(bb, count, size);
  }
  return ss;
}

inline T2_OBStream&
T2_Socket::operator << (bool b)
{
  // note that sizeof(bool) may be different on different platforms, so
  // I am forcing the code to use one byte to transfer a bool
  const char c[2] = { '\0', '\1' };
  if (b == false)
     bwrite((void *)c, 1, sizeof(char));
  else
     bwrite((void *)(c+sizeof(char)), 1, sizeof(char));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (char c)
{
  bwrite((void *)(&c), 1, sizeof(char));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (signed char c)
{
  bwrite((void *)(&c), 1, sizeof(signed char));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (unsigned char c)
{
  bwrite((void *)(&c), 1, sizeof(unsigned char));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (short s)
{
  bwrite((void *)(&s), 1, sizeof(short));
  return *this;
}
inline T2_OBStream&
T2_Socket::operator << (unsigned short s)
{
  bwrite((void *)(&s), 1, sizeof(unsigned short));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (int i)
{
  bwrite((void *)(&i), 1, sizeof(int));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (unsigned int i)
{
  bwrite((void *)(&i), 1, sizeof(unsigned int));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (long l)
{
  bwrite((void *)(&l), 1, sizeof(long));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (unsigned long l)
{
  bwrite((void *)(&l), 1, sizeof(unsigned long));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (float f)
{
  bwrite((void *)(&f), 1, sizeof(float));
  return *this;
}

inline T2_OBStream&
T2_Socket::operator << (double d)
{
  bwrite((void *)(&d), 1, sizeof(double));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (bool &b)
{
  // note that sizeof(bool) may be different on different platforms, so
  // I am forcing the code to use one byte to transfer a bool
  char c;
  bread((void *)(&c), 1, sizeof(c));
  t2_Assert (c == '\0' || c == '\1');
  if (c == '\0')
     b = false;
  else
     b = true;
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (char &c)
{
  bread((void *)(&c), 1, sizeof(char));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (signed char &c)
{
  bread((void *)(&c), 1, sizeof(signed char));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (unsigned char &c)
{
  bread((void *)(&c), 1, sizeof(unsigned char));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (short &s)
{
  bread((void *)(&s), 1, sizeof(short));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (unsigned short &s)
{
  bread((void *)(&s), 1, sizeof(unsigned short));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (int &i)
{
  bread((void *)(&i), 1, sizeof(int));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (unsigned int &i)
{
  bread((void *)(&i), 1, sizeof(unsigned int));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (long &l)
{
  bread((void *)(&l), 1, sizeof(long));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (unsigned long &l)
{
  bread((void *)(&l), 1, sizeof(unsigned long));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (float &f)
{
  bread((void *)(&f), 1, sizeof(float));
  return *this;
}

inline T2_IBStream&
T2_Socket::operator >> (double &d)
{
  bread((void *)(&d), 1, sizeof(double));
  return *this;
}

inline
T2_Socket::T2_Socket(const T2_Socket& s): T2_Sock(s),
					  T2_IBStream(false), T2_OBStream(false)
{
  t2_err_quit(__FILE__,__LINE__,"don't use T2_Socket copy constructor");
}

inline T2_Socket&
T2_Socket::operator =(const T2_Socket& s)
{
  t2_err_quit(__FILE__,__LINE__,"don't use T2_Socket operator =");
  return *this;
}



//////////////////////////////////////////////////////////////////////////
//
// T2_SocketServer
//
// A listening TCP socket. 
//
// To ensure proper execution, one should not set T2_SocketServer to
// non-blocking.
//
//////////////////////////////////////////////////////////////////////////
class T2_SocketServer : public T2_Sock {
public:
  T2_SocketServer(): T2_Sock() {};
  ~T2_SocketServer() {};

  // start the server to listen at port p
  bool startServer(int p, int backlog=20);

  // query method
  short getLocalPort() const;		// return -1 if server not started

  //////
  // poll for a new connection and return true if a connection arrives, 
  // false if time-out or error occurred
  //	errval		reason
  //	-------		---------
  //	ENOTCONN	socket server not started (ie not listening)
  //	ETIMEDOUT	time out
  //	others		error occurred in some system call
  //////
  bool pollNewConnection(T2_Socket& s, struct timeval* timeout=NULL);
  bool pollNewConnection(T2_Socket& s, int nsec, int nusec);

  // wait for a new connection - a blocking call
  // returns false if the object does not have a valid listening socket file
  // or error occurrs; returns true otherwise
  bool acceptConnection(T2_Socket& s);

protected:
  struct sockaddr_in servaddr;

  // open a socket
  bool openSocket() { return T2_Sock::openSocket(AF_INET, SOCK_STREAM, 0); };

private:
  T2_SocketServer(const T2_SocketServer& s);
  T2_SocketServer& operator =(const T2_SocketServer& s);

};


inline bool
T2_SocketServer::pollNewConnection(T2_Socket& s, int nsec, int nusec)
{
  struct timeval tval;
  tval.tv_sec = nsec;
  tval.tv_usec = nusec;
  return pollNewConnection(s, &tval);
}

inline
T2_SocketServer::T2_SocketServer(const T2_SocketServer& s): T2_Sock(s)
{
  t2_err_quit(__FILE__,__LINE__,"don't use T2_SocketServer copy constructor");
}

inline T2_SocketServer&
T2_SocketServer::operator =(const T2_SocketServer& s)
{
  t2_err_quit(__FILE__,__LINE__,"don't use T2_SocketServer operator =");
  return *this;
}



#endif	/* ! _T2_SOCKET_H_ */
