
/* t2_tset.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/2/98
 * Template classes for sets with limited capacity.
 */

#ifndef _T2_TSET_H_
#define _T2_TSET_H_


#include	<iostream.h>
#include	"t2_header.h"


// forward declarations
template<class Type> class T2_TSetBStreamWriter;
template<class Type> class T2_TSetBStreamReader;
template<class Type> class T2_TSetBStrWriter;
template<class Type> class T2_TSetBStrReader;

/////////////////////////////////////////////////////////////////////////
//
// T2_TSet
//
// A template for a set of objects, of type Type.
// The element objects must have a constructor that takes no arguments,
// a copy constructor, an assignment operator (ie operator =),
// an equality test operator (ie operator ==),
// and can be written out to ostream (ie operator<<(ostream&, const Type&).
//
/////////////////////////////////////////////////////////////////////////
template <class Type>
class T2_TSet {
friend class T2_TSetBStreamWriter<Type>;
friend class T2_TSetBStreamReader<Type>;
friend class T2_TSetBStrWriter<Type>;
friend class T2_TSetBStrReader<Type>;
private:
   // do not use copy constructor and operator =
   T2_TSet(const T2_TSet<Type>& s);
   T2_TSet<Type>& operator =(const T2_TSet<Type>& t);

protected:
   u_int ndata,		// number of Type objects stored in data[]
         len;		// total length of array data[] (ie max set size)
   Type *data;

public:
   T2_TSet(const u_int l=0);
   ~T2_TSet() { delete [] data; };

   // query methods
   u_int getMaxSetSize() const { return len; };
   u_int getNumberElements() const { return ndata; };
   bool isFull() const { return (ndata == len)? true:false; };
   bool isEmpty() const { return (ndata == 0)? true:false; };

   // access methods
   const Type* getAllElements() const { return data; };
   Type* getAllElements() { return data; };
   const Type& operator[](u_int i) const { t2_Assert(i<ndata); return data[i];};
   Type& operator[](u_int i) { t2_Assert(i<ndata); return data[i];};
//   const Type& operator[](u_int i) const { return data[i];};

   // this is a method that returns a ref to the i-th position in the set,
   // regardless of whether there is any data in there;
   // this method is given to conveniently initialize all the place
   // holders when necessary
   const Type& getPosition(u_int i) const { t2_Assert(i<len); return data[i];};
   Type& getPosition(u_int i) { t2_Assert(i<len); return data[i];};

   // some operators
   T2_TSet<Type>& operator +=(const Type& d) { in(d); return *this; };
   T2_TSet<Type>& operator <<(const Type& d) { return (*this += d); };

   // modify the set
   // (note that the operator= and the copy constructor of Type are used
   // by the methods below, so their cost depends on the implementation of
   // the operator= and the copy constructor of Type)
   T2_TSet<Type>& emptySet() { ndata = 0; return *this; };
   Type out(u_int i=0);
   bool in(const Type& d);
   bool in(const T2_TSet<Type>& t);

   // increment the number of elements in the set, if not already full,
   // and return the index of the newly added non-initialized element 
   // through the given argument; if set was already full, returns false and
   // i is undefined;
   // if succeeded, this method together with operator [] is an easy way of 
   // adding to the set a "blank" (ie un-initialized) element and have access
   // to it without any memory copy
   // (works similarly as T2_VArray::appendELement() does)
   bool appendElement(u_int& i);

   // enlarge the maximum set size and keep all the elements;
   // method fails and does not change any thing if the new size is smaller
   void growMaxSetSize(u_int newlen);

   // throw away all elements and change the max set size
   u_int clear(u_int l);

   // output to ostream
//   ostream& write(ostream& os) const;
};


template <class Type>
inline
T2_TSet<Type>::T2_TSet(const u_int l): ndata(0), len(l)
{ t2_Alloc(data, l, Type); }

// do not use copy constructor
template <class Type>
inline
T2_TSet<Type>::T2_TSet(const T2_TSet<Type>& s) : ndata(s.ndata), len(s.len)
{
  t2_Alloc(data, len, Type);
  for (u_int i=0; i<ndata; i++)
     data[i] = s.data[i];
  t2_err_quit(__FILE__, __LINE__, "do not use T2_TSet copy constructor");
}

// do not use operator=
template <class Type>
inline T2_TSet<Type>&
T2_TSet<Type>::operator =(const T2_TSet<Type>&)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_TSet operator=");
  return *this;
}

template <class Type>
inline Type
T2_TSet<Type>::out(u_int i)
{
  t2_Assert(!isEmpty() && i<ndata);
  Type r = data[i];
  data[i] = data[--ndata];
  return r;
}

template <class Type>
inline bool
T2_TSet<Type>::in(const Type& d)
{
  if (isFull())
     return false;
  data[ndata++] = d;
  return true;
}

template <class Type>
inline bool
T2_TSet<Type>::in(const T2_TSet<Type>& t)
{
  // make sure enough space first before appending new elements
  if (isFull())
     return false;
  if ((getNumberElements() + t.getNumberElements()) > getMaxSetSize())
     return false;
  for (u_int i=0; i<t.getNumberElements(); i++)
     data[ndata++] = t.data[i];
  return true;
}

template <class Type>
inline bool
T2_TSet<Type>::appendElement(u_int& i)
{
  if (isFull())
     return false;
  i = ndata++;
  return true;
}

// enlarge the maximum set size and keep all the elements;
// method fails and does not change any thing if the new size is smaller
template <class Type>
inline void
T2_TSet<Type>::growMaxSetSize(u_int newlen)
{
  if (len >= newlen) return;		// no change of size
  Type *d = data;
  t2_Alloc(data, newlen, Type);
  for (u_int i=0; i<ndata; i++)
     data[i] = d[i];
  len = newlen;
  delete [] d;
}

template <class Type>
inline u_int
T2_TSet<Type>::clear(u_int newlen)
{
  ndata = 0;
  if (len != newlen) {
     delete [] data;
     t2_Alloc(data, newlen, Type);
     len = newlen;
  }
  return len;
}

//template <class Type>
//inline ostream&
//T2_TSet<Type>::write(ostream& os) const
//{
//  os << "T2_TSet[" << ndata << "] =";
//  for (u_int i=0; i<ndata; i++)
//     os << " " << data[i];
//  os << endl;
//  return os;
//}

//template <class Type>
//inline
//ostream& operator << (ostream& os, const T2_TSet<Type>& s)
//{ return s.write(os); }



/////////////////////////////////////////////////////////////////////////
//
// T2_TSetOStream
//
// A template class for printing out a T2_TSet<Type> to an ostream.
// The data element type is expected to provide operator <<(ostream&)
//
/////////////////////////////////////////////////////////////////////////
template <class Type>
class T2_TSetOStream {
public:
  T2_TSetOStream(const T2_TSet<Type>& s): tset(s) {};
   T2_TSetOStream(const T2_TSetOStream<Type>& s): tset(s.tset) {};

   T2_TSetOStream<Type>& operator =(const T2_TSetOStream<Type>& t);

   // output to ostream
   ostream& write(ostream& os) const;

protected:
  const T2_TSet<Type>& tset;

private:
};


template <class Type>
inline T2_TSetOStream<Type>& 
T2_TSetOStream<Type>::operator =(const T2_TSetOStream<Type>& t)
{
  tset = t.tset;
  return *this;
}

template <class Type>
inline ostream&
T2_TSetOStream<Type>::write(ostream& os) const
{
  u_int ndata = tset.getNumberElements();
  os << "T2_TSet[" << ndata << "] =";
  for (u_int i=0; i<ndata; i++)
     os << " " << tset[i];
  os << endl;
  return os;
}

template <class Type>
inline ostream&
operator <<(ostream& os, const T2_TSetOStream<Type>& s)
{
  return s.write(os);
}



/////////////////////////////////////////////////////////////////////////
//
// T2_TSetSearch
//
// A template class that adds searching to T2_TSet<Type>.
// The data element type is expected to provide operator ==(const Type&).
//
/////////////////////////////////////////////////////////////////////////
template <class Type>
class T2_TSetSearch: public T2_TSet<Type> {
public:
   T2_TSetSearch(const u_int l=0): T2_TSet<Type>(l) {};

   // search methods,
   // returns true and set index if found, returns false otherwise
   bool search(const Type& t) const;
   bool search(const Type& t, u_int& k) const;

protected:
private:
   // do not use copy constructor and operator =
   T2_TSetSearch(const T2_TSetSearch<Type>& s);
   T2_TSetSearch<Type>& operator =(const T2_TSetSearch<Type>& t);
};


// a search method,
// returns true if found, returns false otherwise
template <class Type>
inline bool
T2_TSetSearch<Type>::search(const Type& t) const
{
  for (u_int i=0; i<getNumberElements(); i++)
     if (t == data[i]) return true;
  return false;
}

// a search method,
// returns true and set index if found, returns false otherwise
template <class Type>
inline bool
T2_TSetSearch<Type>::search(const Type& t, u_int& k) const
{
  for (u_int i=0; i<getNumberElements(); i++)
     if (t == data[i]) { k = i; return true; };
  return false;
}

// do not use copy constructor
template <class Type>
inline
T2_TSetSearch<Type>::T2_TSetSearch(const T2_TSetSearch<Type>& s):
			T2_TSet<Type>(s)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_TSetSearch copy constructor");
}

// do not use operator =
template <class Type>
inline T2_TSetSearch<Type>&
T2_TSetSearch<Type>::operator =(const T2_TSetSearch<Type>& s)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_TSetSearch operator =");
  return *this;
}




#endif	/* _T2_TSET_H_ */

