
/* $Id: t2_txtfile.h,v 1.5 2000/03/14 23:24:58 chialin Exp $ */

/* t2_txtfile.h */

#ifndef _T2_TXTFILE_H_
#define _T2_TXTFILE_H_

#include	<fstream.h>
#include	<ctype.h>
#include	<strings.h>
#include	<string.h>
#include	"t2_header.h"

////////////////////////////////////////////////////////////////////////////
//
// T2_TxtFile
//
// A class that is used to access text files with comments. A comment
// is all the texts in a line after a special "mark", eg '#'.
// The constructor takes an argument that specifies the leading char for
// a comment and another argument that specifies the max length of
// a line. Note the this class should not allow seek, as seeking to random
// points in the file may disrupt the way this class works (for example,
// if one seeks to the middle of a comment line, then this class would
// not be able to detect that this is actually a comment). The only supported 
// method similar to seek is rewind, which always seek to the beginning
// of the file. However, there is a seek method provided, and callers
// can use them at their own risk.
//
////////////////////////////////////////////////////////////////////////////
class T2_TxtFile {
public:
  T2_TxtFile(char m, size_t s=512);
  ~T2_TxtFile() { if (ifs) ifs.close(); delete [] filename; delete [] line; };

  operator const void*() const { return (valid == true)? this:NULL; };

  // query methods
  const char* getFileName() const { return filename; };
  size_t getMaxLineLength() const { return maxlinelen; };
  char getCommentMark() const { return mark; };

  // get the line number for the current line in the line buffer
  // (note that this does not work any more if seek() was called)
  u_int getCurrentLineNumber() const { return lineno; };

  // return the offset from the beginning of the file for the current
  // line buffer;
  // note that the way to find out the offset of a string s returned by
  // getNextUncommentString(), just do the following *before* any other
  // call that would advance the read pointer:
  //	offset = getCurrentLineOffset() + (s - getLine());
  off_t getCurrentLineOffset() const { return linepos; };

  // open and close
  void open(const char* fname, bool remember_filename = false);
  void close();
  bool eof() const { return eof_f; };

  // rewind to beginning of file
  void rewind();

  // return the next uncommented string in the file;
  // return NULL if eof is reached;
  // if argument offset is not null, the offset of the returned string from
  // the beginning of the file is stored in *offset
  const char* getNextUncommentString();

  // return the start of the next uncommented string in the given buf
  // (no read from the file !!);
  // return NULL if no uncommented strings is found in the given buf
  const char* getStartOfNextUncommentString(const char* buf) const;

  // return the line buffer and the length of line stored in the buffer
  const char* getLine() const { return line; };
  size_t getLineLength() const { return linelen; };

  // return the next line, or the next non-empty line, that does not start with
  // mark (ie lines that start with mark would be skipped)
  // note that a line that has mark as the 1st non-white-spaced char of the
  // line will not be skipped by getNextLine(), but would be skipped by 
  // getNextNonemptyLine()
  void getNextLine();
  void getNextNonemptyLine();

  // skip the remaining of the current line in buffer
  void skipLine() { cur = NULL; };

  // allocate a char array, write the next uncomment string into that char
  // array, and return the char array; return NULL if no more uncomment
  // string is available;
  // note that the caller is responsible for deleting the char array at the end
  char* copyNextUncommentString();

  // copy the next non-comment string into the given char array, including
  // the '\0' char at the end; returns true if a non-comment string is
  // available, false otherwise; note that buf[] is assumed to be large
  // enough to hold the string
  bool copyNextUncommentString(char* buf);

  // reads
  T2_TxtFile& operator >> (char& c);
  T2_TxtFile& operator >> (short& s);
  T2_TxtFile& operator >> (unsigned short& s);
  T2_TxtFile& operator >> (int& i);
  T2_TxtFile& operator >> (unsigned int& i);
  T2_TxtFile& operator >> (long& i);
  T2_TxtFile& operator >> (unsigned long& i);
  T2_TxtFile& operator >> (float& f);
  T2_TxtFile& operator >> (double& d);
  T2_TxtFile& operator >> (char* c)
			{ copyNextUncommentString(c); return *this; };

  // seek to some random offset from the beginning of the file;
  // note that you should use this function at your own risk, as it
  // may make this class mistreate comments as non-comments
  void seek(off_t offset);

protected:
public:
  ifstream ifs;
  const char mark;		// mark for beginning of comment
  char* filename;		// filename associated with the object

  const size_t maxlinelen;	// max length of a line
  size_t linelen;		// length of line in line buffer
  char *line,			// the line buffer
       *cur;			// cur line pointer in the buffer
  bool eof_f;			// eof flag
  u_int lineno;			// current line number
  off_t linepos;		// offset for the beginning of the current line
  bool valid;

private:
};


inline void 
T2_TxtFile::close()
{
  ifs.close();
  delete [] filename;
  filename = NULL;
  eof_f = true;
  cur = NULL;
  linelen = 0;
  lineno=0;
  linepos = 0;
  valid = false;
}

inline void 
T2_TxtFile::open(const char* fname, bool remember_filename)
{
  close();
  if (fname == NULL)
     return;
  ifs.open(fname, ios::in); 
  if (!ifs) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: cannot open '%s'", fname);
     return;
  }
  if (remember_filename == true) {
     const size_t len = strlen(fname)+1;
     t2_Alloc(filename, len, char);
     memcpy(filename, fname, len);
  }
  cur = NULL; 
  linelen = 0;
  eof_f = false;
  lineno = 0;
  linepos = 0;
  valid = true;
}

inline void
T2_TxtFile::rewind()
{
  if (valid == false)
     return;

  // note that seekg might not work when ifs.eof() is true, therefore
  // need to do ifs.clear() first before doing seekg
  ifs.clear();

  ifs.seekg((streampos) 0, ios::beg);
  cur = NULL; 
  linelen = 0;
  eof_f = false;
  lineno = 0;
  linepos = 0;
}

inline void
T2_TxtFile::seek(off_t offset)
{
  if (!ifs)
     return;
  ifs.seekg(offset, ios::beg);
  cur = NULL;
  linelen = 0;
  eof_f = false;
  lineno = 0;			// no longer makes sense
  linepos = offset;
}



#endif	/* ! _T2_TXTFILE_H_ */
