
/* $Id: qspec.C,v 1.23 2000/11/11 03:27:37 chialin Exp $ */

/* qspec.C */


#include	<iostream.h>
#include	<string.h>
#include	"t2_qspec.h"



////////////////////////////////////////////////////////////////
//
// some functions and variables for reading old versions of
// binary qbatch files;
// note that since the version numbers do not increase
// monotonically, larger version number does not mean later
// version;
//
////////////////////////////////////////////////////////////////

// latest version number for storing a qbatch in a binary file
static const u_int qbatch_version = 1001;

// the current version of qbatch being read from a binary file
static u_int curread_version = 0;

u_int t2_getLatestQBatchOutputBinaryFileVersion()
{
  return qbatch_version;
}

// check if read for the current version is supported
bool
bread_version(T2_IBStream& ibf)
{
  ibf >> curread_version;
  if (curread_version != t2_getLatestQBatchOutputBinaryFileVersion()) {
     // try to accomodate old qbatch formats
     switch (curread_version) {
	case 1000:
		t2_say(cout,
		       "encounters old qbatch version " << curread_version
		       << " (current version = "
		       << t2_getLatestQBatchOutputBinaryFileVersion()
		       << ")" << endl);
		break;			// still able to read this format

	default:
		t2_err_ret(__FILE__, __LINE__,
			   "T2_QBatch::bread expects version %d, but encounters version %d",
			   t2_getLatestQBatchOutputBinaryFileVersion(),
			   curread_version);
		return false;
     }
  }

  // able to read the given format
  return true;
}



////////////////////////////////////////////////////////////////
//
// T2_QSpecDataset methods
//
////////////////////////////////////////////////////////////////
ostream&
T2_QSpecDataset::write(ostream& os) const
{
  os << "     dataset id = " << getDatasetID()
     << ", iterator id = " << getIteratorID() << endl
     << "     index id = " << getIndexID()
     << ", index arg size = " << getIndexConstructorArg().getDataSize() << endl
     << "     index arg = " << getIndexConstructorArg() << endl
     << "     project function id = " << getProjID()
     << ", prj arg size = " << getProjConstructorArg().getDataSize() << endl
     << "     prj arg = " << getProjConstructorArg() << endl
     << "     query box = " << getQueryBox();
  return os;
}


#ifdef _T2_BSTREAM_H_

T2_OBStream&
T2_QSpecDataset::bwrite(T2_OBStream& obf) const
{
  obf << ds_id << ds_iter
      << index_id << index_consarg
      << prj_id << prj_consarg
      << qregion;
  return obf;
}

T2_IBStream&
T2_QSpecDataset::bread(T2_IBStream& ibf)
{
  t2_Assert(ibf);
  ibf >> ds_id >> ds_iter
      >> index_id >> index_consarg
      >> prj_id >> prj_consarg
      >> qregion;
  return ibf;
}

#endif  /* _T2_BSTREAM_H_ */



////////////////////////////////////////////////////////////////
//
// T2_QSpecOutput methods
//
////////////////////////////////////////////////////////////////
ostream&
T2_QSpecOutput::write(ostream& os) const
{
  switch (type) {
    case t2_oUnknown:
		os << "  T2_QSpecOutput has type t2_oUnknown" << endl;
		break;

    case t2_oBFile:
		os << "  T2_QSpecOutput has type t2_oBFile" << endl;
		break;

    case t2_oSocket:
		os << "  T2_QSpecOutput has type t2_oSocket" << endl;
		break;

    case t2_oMC:
		os << "  T2_QSpecOutput has type t2_oMC" << endl;
		break;

    default:
		os << "  T2_QSpecOutput has unrecognized type " << (int) type
		   << endl;
  }

  if (useT2Protocol())
     os << "     T2 protocol enabled" << endl;
  else
     os << "     T2 protocol disabled" << endl;

  switch (getOutputEndian()) {
     case T2_BigEndian:
		os << "     output endian = big endian" << endl; break;
     case T2_LittleEndian:
		os << "     output endian = little endian" << endl; break;
     default:
		os << "     output endian = unknown" << endl; break;
  }

  os << "     host = " << ((hostname == NULL)? "NULL" : hostname)
     << ", port = " << port << endl;
  os << "     file prefix = " << ((prefix == NULL)? "NULL" : prefix) << endl;
  os << "     client = " << ((client == NULL)? "NULL" : client)
     << ", nclients = " << nclients << ", ofbtag = " << ofbtag
     << ", datatag = " << datatag << endl;

  return os;
}



#ifdef _T2_BSTREAM_H_

// write to a binary file
T2_OBStream&
T2_QSpecOutput::bwrite(T2_OBStream& obf) const
{
  size_t len;
  switch (type) {
     case t2_oUnknown:
		t2_err_ret(__FILE__, __LINE__,
			"T2_QSpecOutput writing type t2_oUnknown into a binary file");
		obf << (u_char) 0;
		break;

     case t2_oBFile:
		obf << (u_char) 1;
		obf << uset2protocol << endian;
		if (prefix == NULL) {
		   obf << 0;
		}
		else {
		   len = strlen(prefix);
		   obf << len;
		   obf.bwrite(prefix, len, sizeof(char));
		}
		break;

     case t2_oSocket:
		obf << (u_char) 2;
		obf << uset2protocol << endian;
		if (hostname == NULL) {
		   obf << 0;
		}
		else {
		   len = strlen(hostname);
		   obf << len;
		   obf.bwrite(hostname, len, sizeof(char));
		}
		obf << port;
		break;

     case t2_oMC:
		obf << (u_char) 3;
		obf << uset2protocol << endian;
		if (client == NULL) {
		   obf << 0;
		}
		else {
		   len = strlen(client);
		   obf << len;
		   obf.bwrite(client, len, sizeof(char));
		}
		obf << nclients << ofbtag << datatag;
		break;

     default:
		t2_err_quit(__FILE__, __LINE__,
			"T2_QSpecOutput::bwrite encountering unknown type %d",
			(int) type);
  }

  return obf;
}


// read from a binary file
T2_IBStream&
T2_QSpecOutput::bread(T2_IBStream& ibf)
{
  t2_Assert(ibf);

  size_t len;
  u_char t;
  ibf >> t;

  switch (curread_version) {
     case qbatch_version:
		if (t != 0)			// only read if not t2_oUnknown
		   ibf >> uset2protocol >> endian;
		break;

     case 1000:
		uset2protocol = true;		// no such field in ver 1000
		endian = T2_BigEndian;		// no such field in ver 1000
		break;

     default:
		t2_err_quit(__FILE__, __LINE__,
			"T2_QSpecOutput::bread encountering unknown version %d",
			curread_version);
  }

  switch (t) {
     case 0:
		type = t2_oUnknown;
		delete [] prefix;
		prefix = NULL;
		delete [] hostname;
		hostname = NULL;
		delete [] client;
		client = NULL;
		break;

     case 1:
		type = t2_oBFile;
		delete [] prefix;
		ibf >> len;
		if (len == 0)
		   prefix = NULL;
		else {
		   t2_Alloc (prefix, len+1, char);
		   ibf.bread(prefix, len, sizeof(char));
		   prefix[len] = '\0';
		}
		break;

     case 2:
		type = t2_oSocket;
		delete [] hostname;
		ibf >> len;
		if (len == 0)
		   hostname = NULL;
		else {
		   t2_Alloc (hostname, len+1, char);
		   ibf.bread(hostname, len, sizeof(char));
		   hostname[len] = '\0';
		}
		ibf >> port;
		break;

     case 3:
		type = t2_oMC;
		delete [] client;
		ibf >> len;
		if (len == 0)
		   client = NULL;
		else {
		   t2_Alloc (client, len+1, char);
		   ibf.bread(client, len, sizeof(char));
		   client[len] = '\0';
		}
		ibf >> nclients >> ofbtag >> datatag;
		break;

     default:
		t2_err_quit(__FILE__, __LINE__,
			"T2_QSpecOutput::bread encountering unknown type %d",
			t);
		type = t2_oUnknown;
  }

  return ibf;
}

#endif  /* _T2_BSTREAM_H_ */




////////////////////////////////////////////////////////////////
//
// T2_QSpec methods
//
////////////////////////////////////////////////////////////////
ostream&
T2_QSpec::write(ostream& os) const
{
  os << "T2_QSpec: this=" << (int) this << endl;
  os << "  acc id = " << getAccID()
     << ", iterator id = " << getAccNavigatorID()
     << ", acc arg size = " << getAccConstructorArg().getDataSize() << endl
     << "  acc arg = " << getAccConstructorArg() << endl
     << "  aggr id = " << getAggrID()
     << ", aggr arg size = " << getAggrConstructorArg().getDataSize() << endl
     << "  aggr arg = " << getAggrConstructorArg() << endl
     << "  output box = " << output_box << endl;
  switch (endian) {
     case T2_BigEndian:		os << "  user arg in big endian" << endl;
				break;
     case T2_LittleEndian:	os << "  user arg in little endian" << endl;
				break;
     default:			os << "  user arg in unknown endian" << endl;
				break;
  }
  os << "  total number of datasets = " << dataset.getNumberElements() << endl;
  for (u_int i=0; i<dataset.getNumberElements(); i++) {
     os << "  dataset " << i << endl;
     os << dataset[i] << endl;
  }
  os << output;

  return os;
}


#ifdef _T2_BSTREAM_H_

T2_OBStream&
T2_QSpec::bwrite(T2_OBStream& obf) const
{
  obf << endian << output_box
      << acc_id << acc_navid << acc_consarg
      << aggr_id << aggr_consarg
      << output;

  obf << getNumberDatasets();
  for (u_int i=0; i<getNumberDatasets(); i++)
     obf << dataset[i];
  return obf;
}

T2_IBStream&
T2_QSpec::bread(T2_IBStream& ibf)
{
  t2_Assert(ibf);
  ibf >> endian >> output_box
      >> acc_id >> acc_navid >> acc_consarg
      >> aggr_id >> aggr_consarg;
  ibf >> output;

  u_int n;
  ibf >> n;
  setNumberDatasets(n);
  for (u_int i=0; i<n; i++)
     ibf >> dataset[i];
  return ibf;
}

#endif  /* _T2_BSTREAM_H_ */



////////////////////////////////////////////////////////////////
//
// T2_QBatch methods
//
////////////////////////////////////////////////////////////////

#ifdef _T2_BSTREAM_H_

T2_OBStream&
T2_QBatch::bwrite(T2_OBStream& obf) const
{
  obf << t2_getLatestQBatchOutputBinaryFileVersion();

  u_int n = getNumberQueries();
  obf << n;
  for (u_int i=0; i<n; i++)
     obf << batch[i];
  return obf;
}


T2_IBStream&
T2_QBatch::bread(T2_IBStream& ibf)
{
  t2_Assert(ibf);

  if (bread_version(ibf) == false) {	// doesn't support the version
     valid = false;
     return ibf;
  }

  valid = true;

  u_int n;
  ibf >> n;  t2_Assert (ibf);
  setNumberQueries(n);
  for (u_int i=0; i<n; i++)
     ibf >> batch[i];

  curread_version = 0;
  return ibf;
}


#endif  /* _T2_BSTREAM_H_ */



////////////////////////////////////////////////////////////////
//
// T2_QPtrBatch methods
//
////////////////////////////////////////////////////////////////
T2_QPtrBatch::~T2_QPtrBatch()
{
  const u_int n = getNumberQueries();
  for (u_int i=0; i<n; i++) {
     if (batch[i] != NULL)
	delete batch[i];
  }
}


bool
T2_QPtrBatch::addQuerySpec(T2_QSpec& q, int where)
{
  const u_int n = getNumberQueries();
  if (where < 0) {			// find an empty spot
     for (where=0; (u_int)where<n; where++)
	if (getQuerySpec(where) == NULL)
	   break;
  }
  if ((u_int) where >= n) {		// invalid spot or batch is full

     t2_say(cout, "oops, where = " << where << ", and n = " << n << endl);

     return false;
  }

  batch[where] = &q;
  return true;
}


#ifdef _T2_BSTREAM_H_

T2_OBStream&
T2_QPtrBatch::bwrite(T2_OBStream& obf) const
{
  obf << t2_getLatestQBatchOutputBinaryFileVersion();

  u_int n = getNumberNonNullQueries();
  obf << n;
  for (u_int i=0; i<n; i++)
     if (batch[i] != NULL)
	obf << *(batch[i]);
  return obf;
}


T2_IBStream&
T2_QPtrBatch::bread(T2_IBStream& ibf)
{
  t2_Assert(ibf);

  if (bread_version(ibf) == false) {	// doesn't support the version
     valid = false;
     return ibf;
  }

  valid = true;

  u_int n;
  ibf >> n;  t2_Assert (ibf);
  setNumberQueries(n);
  for (u_int i=0; i<n; i++) {
     T2_QSpec*& qspec = batch[i];
     t2_Assert (qspec == NULL);
     qspec = new T2_QSpec;
     ibf >> *qspec;
  }

  curread_version = 0;
  return ibf;
}

#endif  /* _T2_BSTREAM_H_ */

