
/* sock.c */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/12/98
 * Many codes were taken from "Unix Network Programming" by Stevens
 */


#include	"t2_header.h"
#include	"t2_sockdefs.h"



/************************************************************************
 *
 * functions for systems that only support IPv4
 *
 ************************************************************************/
#ifndef T2_SYS_SUPPORT_INET_PTON
int
inet_pton(int family, const char* strptr, void *addrptr)
{
  if (strptr == NULL || addrptr == NULL) {
     errno = EINVAL;
     return 0;
  }

  if (family == AF_INET) {
     struct in_addr in_val;
     if (inet_aton(strptr, &in_val)) {			/* succeeds ! */
	memcpy(addrptr, &in_val, sizeof(struct in_addr));
	return 1;
     }
     return 0;						/* inet_aton failed */
  }
  errno = EAFNOSUPPORT;
  return -1;
}
#endif	/* ! T2_SYS_SUPPORT_INET_PTON */


#ifndef T2_SYS_SUPPORT_INET_NTOP
const char*
inet_ntop(int family, const void* addrptr, char* strptr, size_t len)
{
  const u_char *p = (const u_char *) addrptr;

  if (addrptr == NULL || strptr == NULL) {
     errno = EINVAL;
     return NULL;
  }

  if (family == AF_INET) {
     char temp[INET_ADDRSTRLEN];
/*     snprintf(temp, sizeof(temp), "%d.%d.%d.%d", p[0], p[1], p[2], p[3]); */
     sprintf(temp, "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);
     if (strlen(temp) >= len) {
	errno = ENOSPC;
	return NULL;
     }
     strcpy(strptr, temp);
     return strptr;
  }
  errno = EAFNOSUPPORT;
  return NULL;
}
#endif	/* ! T2_SYS_SUPPORT_INET_NTOP */




/************************************************************************
 *
 * sock_ntop
 * 
 * A protocol-independent version of ntop functions. It looks into
 * the structure and calls the appropriate function to return the
 * presentation format of the address.
 *
 * Returns non-null pointer if OK, NULL on error
 *
 ************************************************************************/
char*
sock_ntop(const struct sockaddr *sa, socklen_t salen)
{
  char portstr[7];
  static char str[128];

  if (sa == NULL)
     return NULL;

  switch (sa->sa_family) {
    case AF_INET: {
		struct sockaddr_in *sin = (struct sockaddr_in *) sa;
		if (inet_ntop(AF_INET,&(sin->sin_addr),str,sizeof(str)) == NULL)
		   return NULL;
		if (ntohs(sin->sin_port) != 0) {
		   sprintf(portstr, ".%d", ntohs(sin->sin_port));
		   strcat(str, portstr);
		}
		return str;
		}
		break;

    default:
		t2_err_ret(__FILE__, __LINE__,
			  "sock_ntop sees unrecognized sa_family %d",
			  (int) sa->sa_family);
		return NULL;
  }

  return NULL;
}



/************************************************************************
 *
 * printhostent
 * 
 * A function that prints to stdout the content of the given struct
 * hostent.
 *
 ************************************************************************/
void
printhostent(struct hostent *hptr)
{
  char **pptr;
  char str[INET6_ADDRSTRLEN];

  if (hptr == NULL) {
     printf("printhostent gets a NULL pointer\n");
     return;
  }

  printf("offical hostname = %s\n", hptr->h_name);

  for (pptr = hptr->h_aliases; *pptr != NULL; pptr++)
     printf("\talias: %s\n", *pptr);

  switch (hptr->h_addrtype) {
  case AF_INET:
        pptr = hptr->h_addr_list;
        for (; *pptr != NULL; pptr++)
            printf("\taddress: %s\n",
                   inet_ntop(hptr->h_addrtype, *pptr, str, sizeof(str)));
  break;

  default:
        printf("\tunknown address type\n");
        break;
  }
}



/************************************************************************
 *
 * testFcntlFlag, setFcntlFlag and unsetFcntlFlag
 * 
 * testFcntlFlag returns 1 if a particular feature is turned on, 0 if
 * it's off, or -1 if fcntl fails.
 *
 * setFcntlFlag and unsetFcntlFlag set (turn on) and unset (turn off)
 * a feature through fcntl. They return 0 if ok, -1 on error.
 *
 ************************************************************************/
int
testFcntlFlag(int fd, int feature)
{
  int flags;
  if ((flags = fcntl(fd, F_GETFL, 0)) < 0) {
     t2_err_ret(__FILE__, __LINE__, "fcntl(F_GETFL) error");
     return -1;
  }

  if (flags & feature)
     return 1;
  return 0;
}

int
setFcntlFlag(int fd, int feature)
{
  int flags;
  if ((flags = fcntl(fd, F_GETFL, 0)) < 0) {
     t2_err_ret(__FILE__, __LINE__, "fcntl(F_GETFL) error");
     return -1;
  }
  if (fcntl(fd, F_SETFL, flags | feature) < 0) {
     t2_err_ret(__FILE__, __LINE__, "fcntl(F_SETFL) error");
     return -1;
  }
  return 0;
}

int
unsetFcntlFlag(int fd, int feature)
{
  int flags;
  if ((flags = fcntl(fd, F_GETFL, 0)) < 0) {
     t2_err_ret(__FILE__, __LINE__, "fcntl(F_GETFL) error");
     return -1;
  }
  if (fcntl(fd, F_SETFL, flags & ~feature) < 0) {
     t2_err_ret(__FILE__, __LINE__, "fcntl(F_SETFL) error");
     return -1;
  }
  return 0;
}



/************************************************************************
 *
 * hstrerror
 *
 * A function that prints out the error msg according to values set
 * in extern int h_errno, which are set by gethostbyname and gethostbyaddr.
 *
 * This function does not seem to be defined in AIX.
 *
 ************************************************************************/
#ifndef T2_SYS_SUPPORT_HSTRERROR
const char*
hstrerror(int e)
{
  switch (e) {
     case 0:			return ("no error");
     case HOST_NOT_FOUND:	return ("Unknown host");
     case TRY_AGAIN:		return ("Hostname lookup failure");
     case NO_RECOVERY:		return ("Unknown non recoveral error");
     case NO_DATA:		return ("No address associated with name");
  }
  return ("unknown error");
}

#endif	/* ! T2_SYS_SUPPORT_HSTRERROR */
