
/* $Id: t2_conscat.h,v 1.2 2000/02/23 22:08:50 chialin Exp $ */

/* t2_conscat.h */

#ifndef _T2_CONSCAT_H_
#define _T2_CONSCAT_H_

#include	<iostream.h>
#include	<fstream.h>
#include	<string.h>
#include	<sys/types.h>
#include	<regex.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_txtfile.h"
#include	"t2_charptr.h"


////////////////////////////////////////////////////////////////////////
//
// T2_ConsCatalogEntry
//
// An entry in a constructor catalog file, which stores information
// about all registered user-defined constructor functions for
// user-defined classes, such as the datasets, the indexes,
// projections fuctions, aggregation functions, etc. Most importantly,
// the constructor catalog also contains the id number for each of those
// constructors registered.
//
// The file should be generated by a T2 utility program.
//
////////////////////////////////////////////////////////////////////////
class T2_ConsCatalogEntry {
public:
  T2_ConsCatalogEntry();
  ~T2_ConsCatalogEntry() { delete [] name; delete [] description; };

  // clear up all pointers and release memory
  void clear() { delete [] name; delete [] description;
		 name = NULL; description = NULL; };

  // input and output method
  T2_TxtFile& read(T2_TxtFile& file);
  ostream& write(ostream& os) const;

  T2_UDFType type;
  u_int id;
  char *name,
       *description;
  T2_Array<T2_CharPtr> iterators;		// iterator names
						// (only valid for dataset
						// constructor functions)

  // try not to use copy constructor and operator =
  T2_ConsCatalogEntry(const T2_ConsCatalogEntry&);
  T2_ConsCatalogEntry& operator = (const T2_ConsCatalogEntry&);

  // allocate and copy string
  static void str_alloc_cpy(char*& dest, const char* src);

private:
};


inline void
T2_ConsCatalogEntry::str_alloc_cpy(char*& dest, const char* src)
{
  delete [] dest;
  if (src == NULL) {
     dest = NULL;
     return;
  }
  size_t len = strlen(src) + 1;
  t2_Alloc(dest, len, char);
  memcpy(dest, src, len);
}

inline
T2_ConsCatalogEntry::T2_ConsCatalogEntry(): type(T2_UDF_Unknown), id(0),
					  name(NULL), description(NULL),
					  iterators(0)
{
}

inline ostream&
operator << (ostream& os, const T2_ConsCatalogEntry& e)
{
  return e.write(os);
}

inline 
T2_ConsCatalogEntry::T2_ConsCatalogEntry(const T2_ConsCatalogEntry& e)
			: type(e.type), id(e.id),
			  iterators(e.iterators.getNumberElements())
{
  str_alloc_cpy(name, e.name);
  str_alloc_cpy(description, e.description);

  for (u_int i=0; i<e.iterators.getNumberElements(); i++) {
    char*& itp = iterators[i];
    const char* e_itp = e.iterators[i];
    str_alloc_cpy(itp, e_itp);
  }
}

inline T2_ConsCatalogEntry& 
T2_ConsCatalogEntry::operator = (const T2_ConsCatalogEntry& e)
{
  type = e.type;
  id = e.id;
  str_alloc_cpy(name, e.name);
  str_alloc_cpy(description, e.description);

  iterators.clear(e.iterators.getNumberElements());
  for (u_int i=0; i<e.iterators.getNumberElements(); i++) {
    char*& itp = iterators[i];
    const char* e_itp = e.iterators[i];
    str_alloc_cpy(itp, e_itp);
  }
  return *this;
}


  

////////////////////////////////////////////////////////////////////////
//
// T2_ConsCatalogWriter
//
// A file that creates a constructor catalog file, which stores info
// about user-defined constructor functions for user-defined classes,
// such as the datasets, the projection functions, the aggregation
// functions, etc. Most importantly, the front-end meta file contains
// the id number for each of those constructors registered.
//
////////////////////////////////////////////////////////////////////////
class T2_ConsCatalogWriter {
public:
  T2_ConsCatalogWriter(const char* f=NULL);
  ~T2_ConsCatalogWriter() {};

  // open and close file
  void open(const char* fname);
  void close();

  operator const void* () const { return (valid==true)? this:NULL; };

  bool addEntry(T2_UDFType t, u_int id, const char* name, const char* descr,
		const T2_VArray<T2_CharPtr>& iterators);

protected:
  ofstream file;
  bool valid;				// set to true if a file is open

private:
};



////////////////////////////////////////////////////////////////////////
//
// T2_ConsCatalogReader
//
// A file that reads a constructor catalog file, which stores info
// about user-defined constructor functions for user-defined classes,
// such as the datasets, the projection functions, the aggregation
// functions, etc. Most importantly, the front-end meta file contains
// the id number for each of those constructors registered.
// Methods are provided to allow for search of constructors using
// their names.
//
////////////////////////////////////////////////////////////////////////
class T2_ConsCatalogReader {
public:
  T2_ConsCatalogReader(const char* f=NULL);
  ~T2_ConsCatalogReader() { file.close(); if (re_init == true) regfree(&re); };

  // open and close file
  void open(const char* fname);
  void close() { file.close(); };

  operator const void* () const { return (const void*) file; };

  // init regular-expression search function; if 2nd argument is 
  // T2_UDF_Unknown, then no restriction on constructor type,
  // otherwise search is restricted to the given specific type;
  // return false if initialization fails
  bool fetchRegExpInit(const char* pattern, T2_UDFType t = T2_UDF_Unknown,
		       int flag = REG_EXTENDED|REG_NOSUB|REG_ICASE);

  // get the next entry that satisfies the given constraints;
  // return false if no more match is found
  bool fetchRegExpName(T2_ConsCatalogEntry& e);

  // fetch constructor entry for exact match;
  // returns true if a match is found, false otherwise;
  // since names are assumed to be unique, only the 1st one found is
  // returned;
  bool fetchExactMatch(T2_ConsCatalogEntry& e,
		       const char* name, const T2_UDFType t = T2_UDF_Unknown);

protected:
  T2_TxtFile file;

  // for search
  T2_UDFType search_type;
  regex_t re;
  bool re_init;			// has re ever been initialized?

private:
};



#endif	/* ! _T2_CONSCAT_H_ */
