
/* t2_feqpool.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/25/98
 * T2 front-end query pool.
 */


#ifndef _T2_FEQPOOL_H_
#define _T2_FEQPOOL_H_


#include	"t2_header.h"
#include	"t2_tqueue.h"
#include	"t2_qspec.h"


// forward declaration
struct T2_QSpecStatusEntry;
class T2_FEQPool;
class T2_FE;

//////////////////////////////////////////////////////////////////////////////
//
// T2_QSpecStatus
//
// The status of a query specification submitted to T2_FE. It can be in
// one of three states:
//	inactive	there is no pending qspec associated with this object
//	pending		the associated qspec is submitted to T2_FE, but not
//			yet sent back to the back-end
//	activated	the associated qspec is submitted to T2_FE and is
//			forwarded to the back-end already
//
//////////////////////////////////////////////////////////////////////////////
enum T2_QSpecStatusEnum {
		t2_qsInactive = -1,		// not a pending qspec
		t2_qsPending = 0,		// submitted but not yet sent
						// to the back-end
		t2_qsActivated = 1,		// already submitted to the
						// back-end
		t2_qsSuccessful,		// query completed and does
						// intersect with input data
		t2_qsUnsuccessful,		// query completed but did not
						// intersect with any input data
		t2_qsCanceled			// query canceled by back-end
};

class T2_QSpecStatus {
friend struct T2_QSpecStatusEntry;
friend class T2_FEQPool;
friend class T2_FE;
public:
  T2_QSpecStatus(): status(t2_qsInactive) {};
  T2_QSpecStatus(const T2_QSpecStatus& s): status(s.status) {};
  ~T2_QSpecStatus() {};

  T2_QSpecStatus& operator =(const T2_QSpecStatus& s);

protected:
  T2_QSpecStatusEnum status;

  void setStatus(T2_QSpecStatusEnum s) { status = s; };

  // check the status; returns -1 if inactive, 0 if pending, 1 if activated
  int test() const { return (int) status; };

private:
};


inline T2_QSpecStatus&
T2_QSpecStatus::operator =(const T2_QSpecStatus& s)
{
  status = s.status;
  return *this;
}



//////////////////////////////////////////////////////////////////////////////
//
// T2_QSpecStatusEntry
//
//////////////////////////////////////////////////////////////////////////////
struct T2_QSpecStatusEntry {
public:
  T2_QSpec *qspec;
  T2_QSpecStatus *status;

  T2_QSpecStatusEntry() { qspec = NULL; status = NULL; };
  T2_QSpecStatusEntry(T2_QSpec* q, T2_QSpecStatus* s) { qspec = q; status = s;};
  T2_QSpecStatusEntry(const T2_QSpecStatusEntry& e);

  T2_QSpecStatusEntry& operator =(const T2_QSpecStatusEntry& e);
};


inline 
T2_QSpecStatusEntry::T2_QSpecStatusEntry(const T2_QSpecStatusEntry& e)
{
  qspec = e.qspec;
  status = e.status;
}

inline T2_QSpecStatusEntry&
T2_QSpecStatusEntry::operator =(const T2_QSpecStatusEntry& e)
{
  qspec = e.qspec;
  status = e.status;
  return *this;
}


//////////////////////////////////////////////////////////////////////////////
//
// T2_FEQPool
//
//////////////////////////////////////////////////////////////////////////////
class T2_FEQPool {
public:
  T2_FEQPool(u_int l);
  ~T2_FEQPool();

  // is the pool full or empty?
  bool isFull() const { return pool.isFull(); };
  bool isEmpty() const { return pool.isEmpty(); };
  u_int getMaxPoolSize() const { return pool.getMaxQueueSize(); };
  u_int getNumberElements() const { return pool.getNumberElements(); };

  // register a query specification; returns true if submission succeeded,
  // false otherwise
  bool submitQuerySpec(T2_QSpec* q, T2_QSpecStatus* s);

  // retrieve a query out of the pool, returns true if succeeds, false otherwise
  bool retrieveQuerySpec(T2_QSpec*& q, T2_QSpecStatus*& s);

  // cancel a submitted query if possible
  // returns true if cancel succeeds, false otherwise
  bool cancelQuerySpec(T2_QSpecStatus* s);

  // search for a submitted query, returns index if found, -1 if not
  int searchQuerySpec(T2_QSpecStatus* s) const;

protected:
  T2_TQueue<T2_QSpecStatusEntry> pool;

private:
};


inline
T2_FEQPool::T2_FEQPool(u_int l): pool(l)
{
}

inline
T2_FEQPool::~T2_FEQPool()
{
}



#endif	/* ! _T2_FEQPOOL_H_ */
