
/* $Id: dsinfo.C,v 1.3 2000/02/24 17:06:37 chialin Exp $ */

/* dsinfo.C */


#include	<stdio.h>
#include	<ctype.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	"t2_dsinfo.h"


// define if want to write comments when creating dataset info file
//#define T2_DSINFO_WRITE_COMMENT


///////////////////////////////////////////////////////////////////////////
//
// a function used to ignore the comments when T2_DSINFO_WRITE_COMMENT
// is not defined
//
///////////////////////////////////////////////////////////////////////////
static const char t2_dsinfo_empty_string[] = "";
inline const char*
t2_dsinfo_comment(const char* s)
{
#  ifdef T2_DSINFO_WRITE_COMMENT
	return s;
#  else
	return t2_dsinfo_empty_string;
#  endif
}


///////////////////////////////////////////////////////////////////////////
//
// T2_DSInfoFileWriter methods
//
///////////////////////////////////////////////////////////////////////////
void
T2_DSInfoFileWriter::open(const char* fname)
{
  close();
  if (fname == NULL)
     return;

  // check if file already exists
  ifstream ifs(fname, ios::in);
  if (ifs) {
     ifs.close();
     char bakfname[512];
     sprintf(bakfname, "%s.bak", fname);
     if (rename(fname, bakfname) != 0) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: file '%s' already exists; cannot rename to '%s'",
		fname, bakfname);
	return;
     }
     t2_say(cout, "WARNING: Rename " << fname << " to " << bakfname << endl);
  }

  file.open(fname, ios::in | ios::out);
  if (!file) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: cannot open file '%s'", fname);
     return;
  }

  // save first max_ndigits_ndatasets bytes to write total number of 
  // datasets later
  for (u_int i=0; i<max_ndigits_ndatasets; i++)
     file << " ";
  file << t2_dsinfo_comment("  # total number of datasets") << endl << endl;

  valid = true;
}


void
T2_DSInfoFileWriter::close()
{
  if (valid == true) {
     // go back to beginning of file to write number of datasets added
     file.clear();
     file.seekg((streampos) 0, ios::beg);
     file << ndatasets;
     ndatasets = nindexes = 0;
     valid = false;
     file.close();
  }
}


bool
T2_DSInfoFileWriter::addEntry(T2_DSRegistrationEntry& e, T2_DSID& ds_id,
			      T2_Array<T2_IndexID>& idx_ids,
			      T2_Array<T2_CharPtr>& iterators)
{
  if (!file)
     return false;

  u_int i;

  // assign id's
  ds_id = ndatasets++;
  t2_Assert (idx_ids.getNumberElements() == e.indexes.getNumberElements());
  for (i=0; i<idx_ids.getNumberElements(); i++)
     idx_ids[i] = nindexes++;

  // write info
  t2_Assert (e.dataset_name != NULL);
  file << t2_dsinfo_comment("# dataset entry\n")
       << e.dataset_name << t2_dsinfo_comment("  # dataset name") << endl
       << ds_id << t2_dsinfo_comment("   # dataset id") << endl
       << t2_stringEmptyWrap(e.description)
       << t2_dsinfo_comment("  # description") << endl
       << t2_stringEmptyWrap(e.blob_fname)
       << t2_dsinfo_comment("  # blob filename") << endl
       << iterators.getNumberElements()
       << t2_dsinfo_comment("  # number of iterators") << endl;
  for (i=0; i<iterators.getNumberElements(); i++) {
     char *& p = iterators[i];
     t2_Assert (p != NULL);
     file << p << t2_dsinfo_comment("  # iterator name") << endl;
  }
  file << e.indexes.getNumberElements() 
       << t2_dsinfo_comment("  # number of indexes") << endl;
  t2_Assert (e.indexes.getNumberElements() == idx_ids.getNumberElements());
  for (i=0; i<e.indexes.getNumberElements(); i++) {
     const T2_IndexRegistrationEntry& ie = e.indexes[i];
     t2_Assert (ie.index_name != NULL);
     file << ie.index_name << t2_dsinfo_comment("  # index name") << endl
	  << idx_ids[i] << t2_dsinfo_comment("  # index id") << endl
	  << t2_stringEmptyWrap(ie.description) 
	  << t2_dsinfo_comment("  # index description") << endl;
  }
  file << endl;

  return true;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_IndexInfoEntry methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_IndexInfoEntry::read(T2_TxtFile& file)
{
  if (!file)
     return false;
  t2_Assert (file.getFileName() != NULL);

  // get index name
  const char* string = file.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no index name is given in file '%s'.",
		file.getFileName());
     return false;
  }
  str_alloc_cpy(idx_name, string);
  const u_int lineno = file.getCurrentLineNumber();

  // get index id
  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no index id for index %s at line %d in file '%s'.",
		idx_name, lineno, file.getFileName());
     return false;
  }
  int i, atoi_err;
  i = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || i < 0) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid index id '%s' at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), file.getFileName());
     return false;
  }
  idx_id = i;

  // get description, if any
  file.getNextLine();
  if (file.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: unexpectedly sees eof when looking for description for index '%s' given at line %d in file '%s'.",
		idx_name, lineno, file.getFileName());
     return false;
  }
  if ((string = file.getStartOfNextUncommentString(file.getLine())) != NULL) {
     // description is not null, so find start of comment, if any, and ignore it
     char* p = (char *) string;
     while (*p != '\0' && *p != file.getCommentMark())
        p++;
     *p = '\0';
     str_alloc_cpy(idx_description, string);
  }
  file.getNextLine();

  return true;
}


ostream&
T2_IndexInfoEntry::write(ostream& os) const
{
  os << "index id= " << idx_id
     << "	name= " << t2_stringNullWrap(idx_name) << endl
     << "	description= " << t2_stringNullWrap(idx_description)
     << endl;
  return os;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_DSInfoEntry methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_DSInfoEntry::read(T2_TxtFile& file)
{
  clear();

  if (!file)
     return false;
  t2_Assert (file.getFileName() != NULL);

  const char* string = file.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no dataset name is given in file '%s'.",
		file.getFileName());
     return false;
  }
  str_alloc_cpy(dataset_name, string);
  const u_int lineno = file.getCurrentLineNumber();

  // get dataset id
  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no dataset id for dataset %s at line %d in file '%s'.",
		dataset_name, lineno, file.getFileName());
     return false;
  }
  int i, atoi_err;
  i = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || i < 0) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid dataset id '%s' at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), file.getFileName());
     clear();
     return false;
  }
  ds_id = (T2_DSID) i;

  // get description, if any
  file.getNextLine();
  if (file.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: unexpectedly sees eof when looking for description for dataset '%s' given at line %d in file '%s'.",
		dataset_name, lineno, file.getFileName());
     clear();
     return false;
  }
  if ((string = file.getStartOfNextUncommentString(file.getLine())) != NULL) {
     // description is not null, so find start of comment, if any, and ignore it
     char* p = (char *) string;
     while (*p != '\0' && *p != file.getCommentMark())
        p++;
     *p = '\0';
     str_alloc_cpy(dataset_description, string);
  }
  file.getNextLine();

  // get blob data filename, if any
  if (file.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: unexpectedly sees eof when looking for blob filename for dataset '%s' given at line %d in file '%s'.",
		dataset_name, lineno, file.getFileName());
     clear();
     return false;
  }
  if ((string = file.getStartOfNextUncommentString(file.getLine())) != NULL) {
     // filename is not null, so find end of string
     char* p = (char *) string;
     while (*p != '\0' && *p != file.getCommentMark() && !isspace((int)*p))
        p++;
     *p = '\0';
     str_alloc_cpy(blob_fname, string);

     // check file status
     struct stat blob_stat;
     if (stat(string, &blob_stat) != 0) {	// error occurs
	t2_err_ret(__FILE__, __LINE__,
		"WARNING: stat returns error for blob file '%s'.",
		string);
	// ignore, and treate as if no blob file is given
     }
     else
	blob_fsize = blob_stat.st_size;
  }
  file.getNextLine();

  // get number of iterators
  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: unexpectedly sees eof when looking for number of iterators for dataset '%s' given at line %d in file '%s'.",
		dataset_name, lineno, file.getFileName());
     clear();
     return false;
  }
  int niters = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || niters < 0) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid number of iterators at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), file.getFileName());
     clear();
     return false;
  }
  iternames.clear(niters);
  for (i=0; i<niters; i++) {			// get iterator name
     char*& itname = iternames[i];
     if ((string = file.getNextUncommentString()) == NULL) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: only got %d out of %d iterator names before seeing eof in file '%s'.",
		i, niters, file.getFileName());
	clear();
	return false;
     }
     str_alloc_cpy(itname, string);
  }

  // get number of index entries
  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: unexpectedly sees eof when looking for number of indexes for dataset '%s' given at line %d in file '%s'.",
		dataset_name, lineno, file.getFileName());
     clear();
     return false;
  }
  int nidxs = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || nidxs < 0) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid number of indexes at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), file.getFileName());
     clear();
     return false;
  }
  indexes.clear(nidxs);
  for (i=0; i<nidxs; i++) {		// get the index entry
     T2_IndexInfoEntry& ie = indexes[i];
     if (ie.read(file) == false) {
	clear();
	return false;
     }
  }

  return true;
}

ostream&
T2_DSInfoEntry::write(ostream& os) const
{
  os << "dataset id= " << ds_id
     << "	name= " << t2_stringNullWrap(dataset_name) << endl
     << "	description= " << t2_stringNullWrap(dataset_description) << endl
     << "	blob data size= " << blob_fsize
     << "    filename=" << t2_stringNullWrap(blob_fname) << endl;
  os << "iterators:";
  for (u_int i=0; i<iternames.getNumberElements(); i++)
     os << " " << t2_stringNullWrap(iternames[i]);
  os << endl;
  os << "indexes:" << endl;
  for (u_int j=0; j<indexes.getNumberElements(); j++)
    os << indexes[j];
  return os;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_DSInfoFileReader methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_DSInfoFileReader::fetchRegExpInit(const char* pattern, int flags)
{
  file.rewind();

  // get number of datasets
  const char* string = file.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no number of datasets is given in file '%s'.",
		file.getFileName());
     return false;
  }
  int i, atoi_err;
  i = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || i < 0) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: invalid number of datasets '%s' at line %d in file '%s'.",
	   string, file.getCurrentLineNumber(), file.getFileName());
     return false;
  }
  ndatasets = (u_int) i;

  if (re_init == true) {
     regfree(&re);
  }
  else
     re_init = true;
  if (regcomp(&re, pattern, flags) != 0)
       return false;   			// report error

  cur = 0;
  return true;
}


bool
T2_DSInfoFileReader::fetchRegExpName(T2_DSInfoEntry& e)
{
  if (cur >= ndatasets)
     return false;

  t2_Assert (file.eof() == false);

  do {
     t2_Assert (file.eof() == false);
     cur++;

     if (e.read(file) == false) {	// some thing went wrong
	cur = ndatasets;		// no more subsequent calls
	return false;
     }
     if (regexec(&re, e.getDatasetName(), (size_t) 0, NULL, 0) == 0)
	return true;			// successful match
  }
  while (cur < ndatasets);

  e.clear();				// no match
  return false;
}


bool
T2_DSInfoFileReader::fetchExactMatch(T2_DSInfoEntry& e, const char* name)
{
  file.rewind();

  // get number of datasets
  const char* string = file.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no number of datasets is given in file '%s'.",
		file.getFileName());
     return false;
  }
  int i, atoi_err;
  i = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || i < 0) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: invalid number of datasets '%s' at line %d in file '%s'.",
	   string, file.getCurrentLineNumber(), file.getFileName());
     return false;
  }
  ndatasets = (u_int) i;

//  while (file.eof() == false) {
  for (u_int k=0; k<ndatasets; k++) {
     if (e.read(file) == false)		// some thing went wrong
	return false;

     if (strcmp(name, e.getDatasetName()) == 0)
	return true;			// successful match
  }

  e.clear();				// no match
  return false;
}
