
/* $Id: fe.C,v 1.8 2000/03/16 00:49:33 chialin Exp $ */

/* fe.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/26/98
 * front-end methods
 */


#include	"t2_fe.h"


///////////////////////////////////////////////////////////////////////////
//
// T2_FE methods
//
///////////////////////////////////////////////////////////////////////////
T2_FE::T2_FE(u_int l, u_int qwin):
		sockserver(), beconnects(0), nconnected(0),
		be_endian(T2_UnknownEndian), my_endian(t2_checkEndianess()),
		pool(l), max_qtokens(qwin)
{
  // qtokens is initialized to max_qtokens in T2_FE::startup()
}


T2_FE::~T2_FE()
{
  shutdown();
}


bool
T2_FE::startup(u_int be_port)
{
  shutdownBEConnections();			// shutdown old connections

  qtokens = max_qtokens;// the back-end will send an ack for each qbatch
			// submitted (currently, only back-end node 0 does
			// this); to avoid sendQBatch from waiting for
			// acks, the front-end is allowed to starting checking
			// for acks after the qtoken-th qbatch is submitted;
			// that is, the front-end will send qbatch 1, qbatch 2,
			// ... qbatch qtokens without waiting for an ack,
			// and then wait for ack from back-end for qbatch 1
			// when the front-end submits qbatch qtokens+1, wait
			// for ack for qbatch 2 when the front-end submits 
			// qbatch qtokens+2, and so on

  return sockserver.startServer(be_port);	// start socket server
}


void
T2_FE::shutdown()
{
  if (getNumberConnectedBackEnds() > 0) {
     bool r = true;
     for (; qtokens<max_qtokens && r == true; qtokens++)
	r = recvQueryAck();
  }

  if (qtokens < max_qtokens) {
     t2_say(cout, "recvQueryAck return false in T2_FE::shutdown()" << endl);
  }

  sockserver.closeSocket();
  shutdownBEConnections();
}


// error values returned by method connectBackEnd();
// return the back-end node id when succeeds
static const int 
	RECV_CONNECT_ERROR	= -1,	// error while recving connect packet
	ZERO_NUMNODES		= -2,	// back-end claims zero back-end node
	INCONSIST_NUMNODES	= -3,	// back-ends claim different numnodes
	INVALID_MYNODE		= -4,	// invalid mynode
	DUP_BACK_END		= -5,	// duplicate back-end nodes
	SEND_ACCEPT_ERROR	= -6;	// error while sending accept packet

int
T2_FE::connectBackEnd(T2_BEConnection& s, u_char& endian)
{
  t2_Assert (s.getSocket().connected() == true);

  T2_ConnectPacket p;
  int t = s.recvConnectPacket(p, my_endian);
  if (t != T2_ConnectPacketTag) {
	t2_err_ret(__FILE__, __LINE__,
		  "readConnectPacket recvs wrong packet with tag = %d", t);
	return RECV_CONNECT_ERROR;
  }

  u_int nbe_expected = getTotalNumberBackEnds();
  if (nbe_expected == 0) {		// don't know # back-ends yet
	t2_Assert (getNumberConnectedBackEnds() == 0);
	int numnodes = p.getNumberNodes();
	t2_Assert (numnodes >= 0);
	if (numnodes == 0) {			// not sure if this will happen
	   char ipaddr[INET6_ADDRSTRLEN];
           if (s.getSocket().getPeerAddress(ipaddr, INET6_ADDRSTRLEN) == false)
	   {
		ipaddr[0] = '?';
		ipaddr[1] = '\0';
	   }
	   t2_err_ret(__FILE__, __LINE__,
	      "Back-end node from host %s claims there is zero back-end node",
	      ipaddr);
	   return ZERO_NUMNODES;
	}
	beconnects.clear(numnodes);
  }
  else if (nbe_expected != (u_int) p.getNumberNodes()) {	// confused ...
	char ipaddr[INET6_ADDRSTRLEN];
        if (s.getSocket().getPeerAddress(ipaddr, INET6_ADDRSTRLEN) == false) {
	   ipaddr[0] = '?';
	   ipaddr[1] = '\0';
	}
	t2_err_ret(__FILE__, __LINE__,
	     "Expecting %d nodes but connect packet from host %s says %d nodes",
	     nbe_expected, ipaddr, p.getNumberNodes());
	return INCONSIST_NUMNODES;
  }

  int mynode = p.getMyNode();
  if (mynode < 0 || mynode >= p.getNumberNodes()) {
     t2_err_ret(__FILE__, __LINE__,
		"numnodes = %d, but mynode = %d", p.getNumberNodes(), mynode);
     return INVALID_MYNODE;
  }

  T2_BEConnection& con = beconnects[mynode];
  if (!con.getSocket()) {		// a new connection
     con.transfer(s);
  }
  else {				// this connection is already connected
	char ipaddr1[INET6_ADDRSTRLEN],
	     ipaddr2[INET6_ADDRSTRLEN];
        if (s.getSocket().getPeerAddress(ipaddr1, INET6_ADDRSTRLEN) == false) {
	   ipaddr1[0] = '?';
	   ipaddr1[1] = '\0';
	}
        if (con.getSocket().getPeerAddress(ipaddr2, INET6_ADDRSTRLEN) == false){
	   ipaddr2[0] = '?';
	   ipaddr2[1] = '\0';
	}
	t2_err_ret(__FILE__, __LINE__,
		"Host %s connected as back-end %d, now host %s claims to be the same back-end",
		ipaddr1, p.getMyNode(), ipaddr2);
        return DUP_BACK_END;
  }

  T2_AcceptPacket ack;
  if (con.sendAcceptPacket(ack) == false) {	// send accept fails
	t2_err_ret(__FILE__, __LINE__,
		  "Send accept packet to back-end node %d error",
		  p.getMyNode());
	return SEND_ACCEPT_ERROR;
  }

  nconnected++;
  t2_Assert (nconnected <= (u_int) getTotalNumberBackEnds());

  // get back-end node's endian
  endian = p.getMyEndian();

  return mynode;
}


int
T2_FE::connectAllBackEnds(struct timeval* timeout)
{
  if (!sockserver) {
     return -1;
  }

  t2_Assert (nconnected <= (u_int) getTotalNumberBackEnds());

  while (nconnected < (u_int) getTotalNumberBackEnds() ||
	 getTotalNumberBackEnds()==0)
  {
     T2_BEConnection s;
     T2_Socket& sock = s.getSocket();
     bool r = sockserver.pollNewConnection(sock, timeout);
     if (r == false) {					// error occurs
        if (sockserver.getErrorValue() == ETIMEDOUT)	// time out
	   return 0;

	t2_err_ret(__FILE__, __LINE__,
		   "ERROR: T2_SockServer::acceptConnection error");
//	shutdownBEConnections();
	return -1;
     }

     u_char endian;
     int c;
     if ((c = connectBackEnd(s, endian)) < 0) {
       switch (c) {
	case RECV_CONNECT_ERROR:
	case INVALID_MYNODE:
	case SEND_ACCEPT_ERROR:
					// close the new connection
					sock.closeSocket(); return -1;
	case ZERO_NUMNODES:
	case INCONSIST_NUMNODES:
	case DUP_BACK_END:
					shutdownBEConnections(); return -1;
	default:
					t2_err_quit(__FILE__, __LINE__,
					       "unexpected return value %d", c);
					return -1;
       }		// switch
     }
     else {		// if ((c = connectBackEnd(s, endian)) < 0)
       t2_Assert (endian == T2_BigEndian || endian == T2_LittleEndian);
       if (be_endian == T2_UnknownEndian)	// first back-end to connect
	  be_endian = endian;
       else if (be_endian != endian) {		// inconsistent endian-ness
	  t2_err_ret(__FILE__, __LINE__,
		"ERROR: back-end node %d is %s but others are %s",
		c, endian2string(endian), endian2string(be_endian));
	  return -1;
       }

       // check if back-end node uses the same endian as the front-end
       if (my_endian != be_endian) {
	  t2_Assert (beconnects[c].getSocket().outputConvertEndian() == false);
	  t2_Assert (beconnects[c].getSocket().inputConvertEndian() == false);
	  beconnects[c].getSocket().setOutputConvertEndian(true);
       }
     }
  }			// while

  return nconnected;
}


int
T2_FE::pollBackEndConnection(struct timeval* timeout)
{
  if (!sockserver) {
     return -1;
  }

  T2_BEConnection s;
  T2_Socket& sock = s.getSocket();
  bool r = sockserver.pollNewConnection(sock, timeout);
  if (r == false) {					// error occurs
     if (sockserver.getErrorValue() == ETIMEDOUT)	// time out
	return 0;

     t2_err_ret(__FILE__, __LINE__, "T2_SockServer::pollNewConnection error");
//     shutdownBEConnections();
     return -1;
  }

  u_char endian;
  int c;
  if ((c = connectBackEnd(s, endian)) < 0) {
     switch (c) {
	case RECV_CONNECT_ERROR:
	case ZERO_NUMNODES:
	case INCONSIST_NUMNODES:
	case INVALID_MYNODE:
	case DUP_BACK_END:
	case SEND_ACCEPT_ERROR:
					sock.closeSocket(); return -1;
	default:
					t2_err_quit(__FILE__, __LINE__,
					   "ERROR: unexpected return value %d",
					   c);
					return -1;
     }
  }
  else {		// if ((c = connectBackEnd(s, endian)) < 0)
       t2_Assert (endian == T2_BigEndian || endian == T2_LittleEndian);
       if (be_endian == T2_UnknownEndian)	// first back-end to connect
	  be_endian = endian;
       else if (be_endian != endian) {		// inconsistent endian-ness
	  t2_err_ret(__FILE__, __LINE__,
		"ERROR: back-end node %d is %s but others are %s",
		c, endian2string(endian), endian2string(be_endian));
	  return -1;
       }
  }

  return 1;
}


bool
T2_FE::sendQBatch(T2_QBatch& q)
{
  t2_Assert (getTotalNumberBackEnds() > 0 &&
	     getTotalNumberBackEnds() == getNumberConnectedBackEnds());

  if (qtokens > 0) {		// no need to recv ack from back-end yet
      --qtokens;
  }
  else {
      // qtokens has reached zero, so must wait for ack from back-end
      bool rr = recvQueryAck();
      if (rr == false)
	 return false;
  }

  bool r = true;
  for (int i=0; i<getTotalNumberBackEnds(); i++) {
     if (beconnects[i].sendQBatch(q) == false)
        r = false;
  }

  return r;
}


bool
T2_FE::sendQBatch(T2_QPtrBatch& q)
{
  t2_Assert (getTotalNumberBackEnds() > 0 &&
	     getTotalNumberBackEnds() == getNumberConnectedBackEnds());

  if (qtokens > 0) {		// no need to recv ack from back-end yet
      --qtokens;
  }
  else {
      // qtokens has reached zero, so must recv ack from back-end
      bool rr = recvQueryAck();
      if (rr == false)
	 return false;
  }

  return sendQBatch_NoWaitForBEAck(q);
}


bool
T2_FE::sendQBatch_NoWaitForBEAck(T2_QPtrBatch& q)
{
  bool r = true;
  for (int i=0; i<getTotalNumberBackEnds(); i++) {
     if (beconnects[i].sendQBatch(q) == false)
        r = false;
  }

  return r;
}


bool
T2_FE::recvQueryAck()
{
  T2_QueryAckPacket ack;
  int t = ack.recv(beconnects[0].getSocket());	// only recv ack from node 0
  if (t != T2_QueryAckTag)
     return false;

# ifdef T2_ASSERTION
     u_int cnt = 0;
     beconnects[0].getSocket() >> cnt;
     t2_say2(cout, "T2_FE::recvQueryAck recvs counter = " << cnt 
	           << " (qtokens = " << qtokens << ")" << endl);
# endif

  return true;
}


void
T2_FE::do_work()
{
  t2_Assert (getTotalNumberBackEnds() > 0 &&
	     getTotalNumberBackEnds() == getNumberConnectedBackEnds());

  // this method is supposed to dump as many qbatches to the back-end as
  // possible;
  // for now, I'll just dump a query batch, which contains one query, to the
  // back-end per invocation

  if (pool.isEmpty())
     return;

  T2_QPtrBatch qbatch(1);		// a batch of a single query
  T2_QSpec *q = NULL;
  T2_QSpecStatus *s = NULL;
  while (!pool.isEmpty()) {
     bool r = pool.retrieveQuerySpec(q, s);
     t2_Assert (r == true);
     t2_Assert (q != NULL);
     t2_Assert (s != NULL);
     t2_Assert (s->test() != 1);
     if (s->test() == 0)
	break;
  }

  bool r = qbatch.addQuerySpec(*q, 0);
  t2_Assert (r == true);
  r = sendQBatch(qbatch);
  t2_Assert (r == true);

  s->setStatus(t2_qsActivated);
}
