
/* $Id: frontend.C,v 1.8 2000/03/16 00:49:33 chialin Exp $ */

/* frontend.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 10/16/98
 * front-end methods
 */


#include	"t2_fe.h"


// The strategy for endian-handling is to always have the sender do
// the appropriate conversion.
// In the current implementation, I simply sets the covert-endian flag
// for the outgoing-channel of a socket connection (eg between the application
// front-end and T2 front-end, between the T2 front-end and a T2 back-end node,
// between a T2 back-end node and the client), and does byte-swapping whenever
// I write to the socket some thing other than char's. This should work fine
// for all the stuff that T2 code needs to parse (eg the number of query
// specifications in a query batch when submitting a query batch).
// On the other hand, the strategy for user-args for various user-defined
// cosntructor functions is to have the application front-end (ie the process
// generating the query specifications) generate the user-args in the 
// endian-ness that the T2 back-end uses, which could be different from
// the endian-ness for the T2 front-end. Therefore, to avoid the socket code
// blindly swap the bytes of a user-arg, I always have to write a usr-arg 
// into a socket as a char array.

///////////////////////////////////////////////////////////////////////////
//
// T2_AppToFEProtocol methods
//
///////////////////////////////////////////////////////////////////////////
// Methods app_connect() and fe_connectAck() define the protocol between 
// application front-end and T2 front-end when application front-end 
// connects to the T2 front-end and T2 front-end accepts the connection.
// Here is how the protocol works (FE: front-end, BE: back-end).
//	App-FE			T2-FE
//	connects
//				send: connect ack, T2-FE-endian, T2-BE-endian
//	read: ack,
//	      T2-FE-endian,
//	      T2-BE-endian
//	send: app-FE-endian
//				read: app-FE-endian
//				send: number-BE-nodes
//
// The protocol when T2 front-end refuses a connection from the application
// front-end is simply returning a t2_ackBusy packet, as implemented by
// fe_busyAck().
//
T2_AppToFEAck
T2_AppToFEProtocol::app_connect(T2_Socket& sock, const u_char my_endian,
				u_char& t2fe_endian, u_char& t2be_endian,
				u_int& numnodes)
{
  t2_Assert (sock.connected());
  T2_AppToFEAck ack = readFEAck(sock);
  t2_Assert (ack == t2_ackBusy || ack == t2_ackConnect);
  switch (ack) {
     case t2_ackConnect:
		break;

     case t2_ackBusy:
		return t2_ackBusy;

     default:
		t2_err_ret(__FILE__, __LINE__,
		      "ERROR: readFEAck returns unexpected ack %d", (int)ack);
		return t2_ackError;
  }

  // connect succeeded !!

  // read endianess
  t2_Assert (sock.inputConvertEndian() == false);
  t2_Assert (sock.outputConvertEndian() == false);
  sock >> t2fe_endian >> t2be_endian;
  if (sock.getNumberOfBytesLastRead() == 0) {
     t2_err_ret(__FILE__, __LINE__, 
	"ERROR: socket read error during connection protocol to T2 front-end.");
     return t2_ackError;
  }
  t2_Assert ((int) t2fe_endian == T2_BigEndian ||
	     (int) t2fe_endian == T2_LittleEndian);
  t2_Assert ((int) t2be_endian == T2_BigEndian ||
	     (int) t2be_endian == T2_LittleEndian);

  // ask socket to do byte-swapping if T2 front-end uses different endian
  if (t2fe_endian != my_endian)
     sock.setOutputConvertEndian(true);

  // send my-endian
  t2_Assert ((int) my_endian == T2_BigEndian ||
	     (int) my_endian == T2_LittleEndian);
  sock << my_endian;
  if (sock.getNumberOfBytesLastWritten() == 0) {
     t2_err_ret(__FILE__, __LINE__, 
	"ERROR: socket write error during connection protocol to T2 front-end.");
     return t2_ackError;
  }

  // read number of T2 back-end nodes
  sock >> numnodes;
  if (sock.getNumberOfBytesLastRead() < (ssize_t) sizeof(numnodes)) {
     t2_err_ret(__FILE__, __LINE__, 
	"ERROR: socket read error during connection protocol to T2 front-end.");
     return t2_ackError;
  }

  return t2_ackConnect;
}


void
T2_AppToFEProtocol::fe_busyAck(T2_Socket& sock)
{
  writeFEAck(sock, t2_ackBusy);
}


bool
T2_AppToFEProtocol::fe_connectAck(T2_Socket& sock, const u_char t2fe_endian,
				const u_char t2be_endian, const u_int numnodes,
				u_char& appfe_endian)
{
  // send ack
  writeFEAck(sock, t2_ackConnect);

  // send endian-ness
  t2_Assert ((int) t2fe_endian == T2_BigEndian ||
	     (int) t2fe_endian == T2_LittleEndian);
  t2_Assert ((int) t2be_endian == T2_BigEndian ||
	     (int) t2be_endian == T2_LittleEndian);
  sock << t2fe_endian << t2be_endian;
  if (sock.getNumberOfBytesLastWritten() == 0) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: socket write error during connection protocol with an application front-end.");
	return false;
  }

  // read application front-end's endianness
  t2_Assert (sock.inputConvertEndian() == false);
  t2_Assert (sock.outputConvertEndian() == false);
  sock >> appfe_endian;
  if (sock.getNumberOfBytesLastRead() == 0) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: socket read error during connection protocol with an application front-end.");
	return false;
  }
  t2_Assert ((int) appfe_endian == T2_BigEndian ||
	     (int) appfe_endian == T2_LittleEndian);

  // ask socket to do byte-swapping if T2 front-end uses different endian
  if (t2fe_endian != appfe_endian)
     sock.setOutputConvertEndian(true);

  // send number of back-end nodes
  sock << numnodes;
  if (sock.getNumberOfBytesLastWritten() < (ssize_t) sizeof(numnodes)) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: socket write error during connection protocol with an application front-end.");
	return false;
  }

  // connection succeeded
  return true;
}


void
T2_AppToFEProtocol::app_disconnect(T2_Socket& s)
{
  t2_Assert (s.connected());
  writeAppCommand(s, t2_acmdDisconnect);
}


void
T2_AppToFEProtocol::fe_disconnectAck(T2_Socket& s)
{
}


bool
T2_AppToFEProtocol::app_ping(T2_Socket& s)
{
  t2_Assert (s.connected());
  writeAppCommand(s, t2_acmdPing);
  T2_AppToFEAck ack = readFEAck(s);
  if (ack != t2_ackPing)
     return false;

  return true;
}


void
T2_AppToFEProtocol::fe_pingAck(T2_Socket& s)
{
  t2_Assert (s.connected());
  writeFEAck(s, t2_ackPing);
}


bool
T2_AppToFEProtocol::app_submitQBatch(T2_Socket& s, T2_QBatch& qbatch)
{
  t2_Assert (s.connected());
  writeAppCommand(s, t2_acmdSubmit);

  s << qbatch;
  return true;
}


bool
T2_AppToFEProtocol::app_sendInquiry(T2_Socket& s, T2_FEInquiry& inq)
{
  t2_Assert (s.connected() == true);
  if (writeAppCommand(s, t2_acmdInquiry) == false ||
      inq.bwrite(s) == false)
  {
     return false;
  }

  return true;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_QBatchList methods
//
///////////////////////////////////////////////////////////////////////////
T2_QPtrBatch&
T2_QBatchList::appendEmptyQBatch()
{
  T2_QPtrBatch* q = new T2_QPtrBatch;
  t2_Assert (q != NULL);

  qlist.appendList(q);
//  return *q;
  t2_Assert (qlist.getTailData() == q);
  return *(qlist.getTailData());
}


T2_QPtrBatch*
T2_QBatchList::getAndRemoveQBatch()
{
  t2_Assert (qlist.isEmpty() == false);
  t2_Assert (qlist.curAtHead() == true);

  T2_QPtrBatch* q = qlist.getCurData();
  t2_Assert (q != NULL);

  qlist.deleteCurWrap();
  t2_Assert (qlist.curAtHead() == true);

  return q;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_FrontEnd methods
//
///////////////////////////////////////////////////////////////////////////
T2_FrontEnd::~T2_FrontEnd()
{
}


bool
T2_FrontEnd::connectT2FrontEndByHostname(const char* hostname, short port)
{
  errval = t2_feNoError;
  disconnectT2FrontEnd();

  if (sock.connectByName(hostname, port) == false) {
     errno = sock.getErrorValue();
     errval = t2_feConnectFailure;
     return false;
  }

  return waitforConnectAck();
}


bool
T2_FrontEnd::connectT2FrontEndByAddress(const char* ipaddr, short port)
{
  errval = t2_feNoError;
  disconnectT2FrontEnd();

  if (sock.connectByAddress(ipaddr, port) == false) {
     errno = sock.getErrorValue();
     errval = t2_feConnectFailure;
     return false;
  }
  return waitforConnectAck();
}


void
T2_FrontEnd::disconnectT2FrontEnd()
{
  errval = t2_feNoError;
  nbackends = 0;
  t2fe_endian = T2_UnknownEndian;
  if (sock.connected() == false)		// do nothing if not connected
     return;

  protocol.app_disconnect(sock);
  sock.closeSocket();
}


bool
T2_FrontEnd::waitforConnectAck()
{
  t2_Assert (sock.connected() == true);
  t2_Assert (errval == t2_feNoError);

  switch (protocol.app_connect(sock, my_endian, t2fe_endian, t2be_endian,
			       nbackends))
  {
     case t2_ackBusy:
		errval = t2_feT2Busy;
		sock.closeSocket();
		return false;
		break;

     case t2_ackConnect:
		break;

     case t2_ackUnexpectedDisconnect:
		errval = t2_feUnexpectedDisconnect;
		sock.closeSocket();
		return false;
		break;

     default:
		errval = t2_feSysError;
		sock.closeSocket();
		return false;
		break;
  }

  // connection succeeded !!
  return true;
}


bool
T2_FrontEnd::pingT2FrontEnd()
{
  errval = t2_feNoError;
  if (sock.connected() == false) {
     errval = t2_feNotConnected;
     return false;
  }
  if (protocol.app_ping(sock) == false) {
     errval = t2_fePingAckError;
     return false;
  }
  return true;
}


bool
T2_FrontEnd::submitQBatch(T2_QBatch& qbatch)
{
  errval = t2_feNoError;
  t2_Assert (qbatch.isValid() == true);
  if (sock.connected() == false) {
     errval = t2_feNotConnected;
     return false;
  }

//  qbatch.bwrite(sock);
  return protocol.app_submitQBatch(sock, qbatch);
}



const char*
T2_FrontEnd::errorValToString(T2_FrontEndError e)
{
  static char msg[128];

  switch (e) {
	case t2_feNoError:
		return "no error";

	case t2_feSysError:	
		return "T2_FrontEnd software error";

	case t2_feNotConnected:
		return "currently not connected to T2";

	case t2_feConnectFailure:
		return "failed to connect to T2";

	case t2_feUnexpectedDisconnect:
		return "connection closed unexpectedly by T2";

	case t2_feT2Busy:
		return "T2 is too busy";

	case t2_fePingAckError:
		return "unrecognized ack from T2 for ping";

	case t2_feEInvalid:
		return "invalid argument";

	case t2_feWriteError:
		return "write error";

	case t2_feReadError:
		return "read error";
  }

  sprintf(msg, "unrecognized T2_FrontEndError value %d", (int) e);
  return msg;
}


bool
T2_FrontEnd::inquireFunction(const char* pattern, bool exact_match,
			     const T2_UDFType func_type, const int fields,
			     T2_FEFunctionInquiryResults& results)
{
  errval = t2_feNoError;
  results.getStatus() = T2_FEInquiry::results_ok;

  if (sock.connected() == false) {
     errval = t2_feNotConnected;
     results.getStatus() = T2_FEInquiry::bad_socket_error;
     return false;
  }
  if (pattern == NULL) {
     errval = t2_feEInvalid;
     results.getStatus() = T2_FEInquiry::invalid_inquiry;
     return false;
  }

  // set up an appropriate inquiry object
  T2_FEInquiry inq;
  inq.getInquiryType() = T2_FEInquiry::inquiry_function;
  inq.getFunctionType() = func_type;
  inq.setNameNoCopy(pattern);
  inq.getExactMatchFlag() = exact_match;
  inq.getFieldFlag() = (u_short) fields;
  inq.getTotalBlobDataSizeLimit() = 0;

  t2_say5(cout, "about to send inquiry to T2 front-end" << endl << inq << endl);

  // ssend inquiry to T2 front-end
  if (protocol.app_sendInquiry(sock, inq) == false) {  	// send failed
     errval = t2_feWriteError;
     results.getStatus() = T2_FEInquiry::bad_socket_error;
     return false;
  }

  t2_say5(cout, "gonna call result_reader.initiate" << endl);

  // wait for reply
  T2_FEInquiryResultsReader result_reader(sock);
  if (result_reader.initiate() == false) {
     errval = t2_feReadError;
     results.getStatus() = T2_FEInquiry::bad_socket_error;
     t2_err_ret(__FILE__, __LINE__, "ERROR: initiate result transfer fails");
     return false;
  }

  t2_say5(cout, "gonna call result_reader.readResults" << endl);

  if (result_reader.readResults(results) == false) {
     errval = t2_feReadError;
     t2_err_ret(__FILE__, __LINE__, "ERROR: result transfer fails");
     return false;
  }

  t2_say5(cout, "return from result_reader.readResults" << endl);

  return true;
}


bool
T2_FrontEnd::inquireDataset(const char* pattern, bool exact_match,
			    const int fields, const size_t max_totblobsize,
			    T2_FEDatasetInquiryResults& results)
{
  errval = t2_feNoError;
  results.getStatus() = T2_FEInquiry::results_ok;

  if (sock.connected() == false) {
     errval = t2_feNotConnected;
     results.getStatus() = T2_FEInquiry::bad_socket_error;
     return false;
  }
  if (pattern == NULL) {
     errval = t2_feEInvalid;
     results.getStatus() = T2_FEInquiry::invalid_inquiry;
     return false;
  }

  // set up an appropriate inquiry object
  T2_FEInquiry inq;
  inq.getInquiryType() = T2_FEInquiry::inquiry_dataset;
  inq.getFunctionType() = T2_UDF_Unknown;
  inq.setNameNoCopy(pattern);
  inq.getExactMatchFlag() = exact_match;
  inq.getFieldFlag() = (u_short) fields;
  inq.getTotalBlobDataSizeLimit() = max_totblobsize;

  t2_say5(cout, "inside inquireDataset: inquiry is" << endl << inq << endl);

  // ssend inquiry to T2 front-end
  if (protocol.app_sendInquiry(sock, inq) == false) {  	// send failed
     errval = t2_feWriteError;
     results.getStatus() = T2_FEInquiry::bad_socket_error;
     return false;
  }

  t2_say5(cout, "about to call initiate" << endl);

  // wait for reply
  T2_FEInquiryResultsReader result_reader(sock);
  if (result_reader.initiate() == false) {
     errval = t2_feReadError;
     results.getStatus() = T2_FEInquiry::bad_socket_error;
     t2_err_ret(__FILE__, __LINE__, "ERROR: initiate result transfer fails");
     return false;
  }

  t2_say5(cout, "after initiate, about to call readResults" << endl);

  if (result_reader.readResults(results) == false) {
     errval = t2_feReadError;
     t2_err_ret(__FILE__, __LINE__, "ERROR: result transfer fails");
     return false;
  }

  t2_say5(cout, "after readResults" << endl);

  return true;
}
