
/* $Id: t2fe.C,v 1.8 2000/03/16 00:49:33 chialin Exp $ */

/* t2fe.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/22/98
 * A sequential pure T2 front-end process
 * interface).
 */


#include	<unistd.h>
#include	<strings.h>
#include	<sys/types.h>
#include	<sys/time.h>
#include	"t2_signal.h"
#include	"t2_qspec.h"
#include	"t2_fe.h"
#include	"t2_conscat.h"
#include	"t2_dsinfo.h"


/////////////////////////////////////////////////////////////////////////////
//
// some help function prototypes and variables
//
/////////////////////////////////////////////////////////////////////////////
enum usageError { eUNKNOWN, eBADARG, eNOARG };
static void printusage(const char*);
static void badusage(usageError, const char*, char);

const u_int default_client_port = 9987,
	    default_client_served = 3,
	    default_max_num_clients = 5,
	    default_be_port	= 9988,
	    default_be_nqry	= 10;
const int   default_be_timeout	= -1;

void
t2fe_action(T2_FE& fe, T2_SocketServer& listensock,
	    T2_ConsCatalogReader& conscat, T2_DSInfoFileReader& dsinfo,
	    T2_TSet<T2_AppFrontEnd*>& free_afe, int clients_served,
	    const u_int max_num_qlists);

T2_AppToFECommand
processAppCommand(T2_AppToFEProtocol& protocol, T2_AppFrontEnd& app,
		  T2_ConsCatalogReader& conscat, T2_DSInfoFileReader& dsinfo,
		  T2_FE& fe);
bool
process_inquiry(T2_Socket& sock, T2_ConsCatalogReader& conscat,
                T2_DSInfoFileReader& dsinfo);


/////////////////////////////////////////////////////////////////////////////
//
// main
//
/////////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  int client_port			// port that app front-end connect to
		= default_client_port,
      clients_served
		= default_client_served,
      max_num_clients			// max # of app front-ends to accept
		= default_max_num_clients,
      be_port				// port that back-end connects to
		= default_be_port,
      be_nqry 				// # query batches to send to back-end
		= default_be_nqry,
      be_timeout			// # secs to wait for all back-end nodes
		= default_be_timeout;	// to connect

  const char* qfile = NULL;
  u_int qtokens = 3;

  const char *conscat_file = NULL,	// constructor function file
	     *dsinfo_file = NULL;	// dataset information file

  // deal with the command-line arguments
  int i;
  for (i=0; i<argc; i++) {
     if (argv[i][0] == '-') {
	switch(argv[i][1]) {
	case 'h':
		printusage(argv[0]); return 0; break;

	case 'p':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		client_port = atoi(argv[i]);
		if (client_port <= 0)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		break;

	case 'f':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		conscat_file = argv[i];
		break;

	case 'd':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		dsinfo_file = argv[i];
		break;

	case 'c':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		clients_served = atoi(argv[i]);
		if (clients_served <= 0)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		break;

	case 'a':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		max_num_clients = atoi(argv[i]);
		if (max_num_clients <= 0)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		break;

	case 'b':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		be_port = atoi(argv[i]);
		if (be_port <= 0)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		break;

	case 'n':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		be_nqry = atoi(argv[i]);
		if (be_nqry < 0)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		break;

	case 'q':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		qfile = argv[i];
		break;

	case 't':
		i++;
		if (i >= argc)
		   { badusage(eNOARG, argv[0], argv[i-1][1]); return -1; }
		be_timeout = atoi(argv[i]);
		break;

	default:
		badusage(eUNKNOWN, argv[0], argv[i][1]); return -1; break;
        }
     }
  }

  if (conscat_file == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no constructor catalog file was specified");
     return -1;
  }

  if (dsinfo_file == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no dataset information file was specified");
     return -1;
  }

  T2_ConsCatalogReader conscat(conscat_file);
  if (!conscat) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open constructor function catalog file '%s'.",
		conscat_file);
     return -1;
  }

  T2_DSInfoFileReader dsinfo(dsinfo_file);
  if (!dsinfo) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open dataset information file '%s'.",
		dsinfo_file);
     return -1;
  }

  cout << "Expecting to serve at least " << clients_served
       << " clients before terminating" << endl;
  cout << "Will serve at any time at most " << max_num_clients << " clients"
       << endl;
  if (qfile != NULL) {
     cout << "Will read query batch file " << qfile << " and send the query"
	  << endl
          << "batch in the file to the back-end " << be_nqry << " times"
	  << endl;
  }


  // ignore SIGPIPE
  t2_signal(SIGPIPE, SIG_IGN);


  T2_FE fe(200, qtokens);		// front-end with max 200 query entries
  if (fe.startup(be_port) == false) {		// front-end startup error
     t2_err_ret(__FILE__, __LINE__, "fe.startup error");
     return -1;
  }

  T2_SocketServer listensock;
  if (qfile == NULL) {			// need to listen for clients
     if (listensock.startServer(client_port) == false) {
        t2_err_ret(__FILE__, __LINE__,
		   "cannot start T2_SocketServer at port %d", client_port);
	return -1;
     }
  }

  u_int nbe;
  if (be_timeout >= 0) {
     cout << "Now expecting back-end nodes to connect to port " << be_port 
          << " in " << be_timeout << " sec" << endl;
     nbe = fe.connectAllBackEnds(be_timeout);	// connect to back-end
  }
  else {
     cout << "Now blocking until back-end nodes connect to port " << be_port 
          << endl;
     nbe = fe.connectAllBackEnds();		// connect to back-end
  }

  if (nbe < 0) {
     t2_err_ret(__FILE__, __LINE__, "fe.connectAllBackEnd error");
     return -1;
  }
  t2_Assert (nbe <= (u_int) fe.getTotalNumberBackEnds());
  if (nbe != (u_int) fe.getTotalNumberBackEnds()) {
     cout << endl << "only " << nbe << " out of "
	  << fe.getTotalNumberBackEnds()
	  << " expected back-end nodes are connected, shutting down" << endl
	  << "the back-end and runs in test mode" << endl;
    fe.shutdown();
  }
  if (nbe == 0) {		// no back-end connects
     cout << "Time's up. No back-end connects to front-end,"
	  << " so now runs in test mode,"
	  << endl
	  << "where T2 query submissions would just be dropped,"
	  << " but inquries are honored."
	  << endl;
  }

  // report T2 back-end information
  cout << endl << fe.getNumberConnectedBackEnds()
       << " back-end nodes are connected, and they are "
       << endian2string(fe.getBackEndsEndian()) << " nodes"
       << endl;
  cout << "now awaits clients to connect to port " << client_port
       << endl << endl;


  if (qfile == NULL) {
     T2_AppFrontEnd *afe = new T2_AppFrontEnd[max_num_clients];
     T2_TSet<T2_AppFrontEnd*> afeptr(max_num_clients);
     for (int j=0; j<max_num_clients; j++)
	afeptr.in(afe + j);
     t2fe_action(fe, listensock, conscat, dsinfo, afeptr,
		 clients_served, clients_served);
     delete [] afe;
  }
  else {
     T2_IBFile qibf(qfile);
     if (!qibf) {
	t2_err_ret(__FILE__, __LINE__,
		  "Cannot open query batch file %s", qfile);
	return -1;
     }

     T2_QBatch q;
     qibf >> q;
     qibf.close();

     if (q.isValid() == false) {		// qbatch read error
	t2_err_ret(__FILE__, __LINE__, "Query batch read error");
     }
     else {

	// the following uses the new sendQBatch interface
	for (int j=0; j<be_nqry; j++) {
	  cout << "submitting query batch file " << j << "-th time" << endl;
	  bool r = fe.sendQBatch(q);
	  if (r == false) {
	     t2_err_ret(__FILE__, __LINE__, "fe.sendQBatch error");
	     break;
	  }
	}			// for (u_int j=0; j<be_nqry; j++)
     }				// if (q.isValid() == false) else

  }				// if (qfile == NULL) else


  cout << endl << "now shutting down the back-end" << endl;
  fe.shutdown();

  return 0;
}



/////////////////////////////////////////////////////////////////////////////
//
// help functions
//
/////////////////////////////////////////////////////////////////////////////
void
printusage(const char* pname)
{
  cout << "Usage: " << pname
    << " [-h] |"
    << " -d <dataset info file>"
    << " -f <constructor catalog file>" << endl
    << " [-p <port number for application front-ends to connect>]"
    << endl
    << " [-a <max number of application front-ends accept at a time>]" << endl
    << " [-c <# of clients to serve before terminated>]" << endl
    << " [-b <port number for back-end to connect>]" << endl
    << " [-q <query batch file>]"
    << " [-n <# query batches>]"
    << endl
    << " [-t <# secs to wait for back-end to connect>]"
    << endl
    << endl;

  cout << "Where:" << endl;
  cout << "-h prints this message" << endl;
  cout << "-d specifies the dataset information file" << endl;
  cout << "-f specifies the constructor function catalog file" << endl;
  cout << "-p specifies the port number for application front-end to connect to"
       << endl
       << "   and must be > 0 (default = " << default_client_port  << ")" 
       << endl;
  cout << "-c specifies the number for applications to serve before terminated"
       << endl
       << "   and must be >= 0 (default = " << default_client_served  << ")" 
       << endl;
  cout << "-a specifies the max number of application front-ends to accept at a time"
       << endl
       << "   and must be > 0 (default = " << default_max_num_clients  << ")" 
       << endl;
  cout << "-b specifies the port number for T2 back-end to connect to" << endl
       << "   and must be > 0 (default = " << default_be_port  << ")" << endl;
  cout << "-q specifies the query batch file" << endl
       << "   (default = NULL)" << endl;
  cout << "-n when -q is used to specify the query batch file, -n specifies the"
       << endl
       << "   number of times the query batch file is read and the query batches"
       << endl
       << "   stored in the file are sent to the back-end before shutting down;"
       << endl
       << "   when -q is not specified, -n specifies the number of times an empty"
       << endl
       << "   query batch file is sent to the back-end before shutting down;"
       << endl
       << "   value must be > 0 (default = " << default_be_nqry  << ")"
       << endl;
  cout << "-t specifies the number of seconds to wait for all back-end nodes to"
       << endl
       << "   connect; will block and wait for all back-end nodes to connect if < 0"
       << endl
       << "   (default = " << default_be_timeout << " sec)"
       << endl;
}


void
badusage(usageError e, const char* p, char a)
{
  switch (e) {
    case eUNKNOWN:
        cerr << "Unknown parameter for -" << a << " in command line" << endl;
        break;
    case eBADARG:
        cerr << "Bad argument for parameter -" << a << " in command line"
             << endl;
        break;
    case eNOARG:
        cerr << "Missing argument for parameter -" << a << " in command line"
             << endl;
        break;
    default:
        cerr << "Some kind of error in the parameter -" << a
             << " in command line" << endl;
  }
  printusage(p);
}



///////////////////////////////////////////////////////////////////////////
//
// set fd_set
//
///////////////////////////////////////////////////////////////////////////
inline void
set_fdset(int fd, fd_set& fdset, int& nfds)
{
  if (fd <= 0)
     return;
  t2_Assert (!FD_ISSET(fd, &fdset));
  FD_SET(fd, &fdset);
  if (fd >= nfds)
     nfds = fd + 1;
}


inline void
reset_fd(fd_set& fdset, int& nreadfds, int be_fd)
{
  FD_ZERO(&fdset);
  if (be_fd > 0) {
     FD_SET(be_fd, &fdset);
     nreadfds = be_fd + 1;
  }
  else
     nreadfds = 0;
}


///////////////////////////////////////////////////////////////////////////
//
// real action
//
//	fe		communicates with the back-end
//	listensock	listens for new clients to connect
//	free_afe	a list of free entries for T2_AppFrontEnd
//			(need one entry per connected client)
//	clients_served	min # of clients served before shutting down
//	max_num_qlists	max # of lists of pending queries
//	qtokens		max # of qbatches to send to the back-end
//			before waiting for acks from the back-end
//
///////////////////////////////////////////////////////////////////////////
void
t2fe_action(T2_FE& fe, T2_SocketServer& listensock,
	    T2_ConsCatalogReader& conscat, T2_DSInfoFileReader& dsinfo,
	    T2_TSet<T2_AppFrontEnd*>& free_afe, int clients_served,
	    const u_int max_num_qlists)
{
  t2_Assert (fe.getMax_QTokens() > 0);
  u_int qtokens = fe.getMax_QTokens();

  T2_AppToFEProtocol protocol;
  T2_TDBList<T2_AppFrontEnd*> clients;

  T2_TDBList<T2_QBatchList*> qlists;

  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = 0;

  const int num_benodes = fe.getNumberConnectedBackEnds();
  int be_fd = -1;
  if (num_benodes > 0) {
    T2_Socket& be_sock = fe.getBackEndConnectionSocket(0);
    t2_Assert (be_sock.connected() == true);
    be_fd = be_sock.getFileDescriptor();
    t2_Assert (be_fd > 0);
  }

  int nreadfds = 0;
  fd_set readfds;
  reset_fd(readfds, nreadfds, be_fd);

  // get endian-ness information
  const u_char t2fe_endian = t2_checkEndianess(),
	       t2be_endian = (num_benodes > 0) ? 
				fe.getBackEndsEndian() : T2_BigEndian;

const double proctime = 5.;
double dt = -1.;


  bool notdone = true;

  do {

    // poll for new connection
    T2_Socket s;
    if (listensock.pollNewConnection(s,0,0) == true) {	// a new client connects

#	ifdef T2_VERBOSE
	{
	   char peer[128];
	   bool r = s.getPeerAddress(peer, 128);
	   t2_Assert (r == true);
	   cout << "--- a new client from " << peer << " ask for connection" 
		<< endl;
	}
#	endif


	if (free_afe.isEmpty() ==  true || 		// busy, try again
	    qlists.getListSize() == max_num_qlists)	// too many qlists
	{
	   protocol.fe_busyAck(s);
	   s.closeSocket();
	   t2_say(cout, "connection rejected due to busy" << endl);
	}
	else {						// accept connection

	   u_char app_endian;
	   if (protocol.fe_connectAck(s, t2fe_endian, t2be_endian, num_benodes,
				      app_endian) == false)
	   {  // connection fails during connection protocol
	      s.closeSocket();
	   }
	   else {
	      // connection succeeds
	      t2_Assert (qlists.getListSize() < max_num_qlists);
	      t2_Assert ((int) app_endian == T2_BigEndian ||
			 (int) app_endian == T2_LittleEndian);

	      // create a new client
	      T2_AppFrontEnd* app = free_afe.out();
	      t2_Assert (app != NULL);
	      t2_Assert (app->getQueryList() == NULL);
	      app->setAppFrontEndEndian(app_endian);

	      // deal with the socket
	      T2_Socket& appsock = app->getSocket();
	      t2_Assert (appsock.connected() == false);
	      appsock.transfer(s);
	      set_fdset(appsock.getFileDescriptor(), readfds, nreadfds);
	      clients.appendList(app);

	      // deal with the query list
	      T2_QBatchList* ql = new T2_QBatchList;
	      app->setQueryList(ql);
	      qlists.appendList(ql);
	      ql->setAppFrontEnd(app);

	      t2_say(cout, "connection accepted" << endl);
	      t2_say(cout, "now have " << clients.getListSize() 
			   << " clients" << endl);

	   }		// if (protocol.fe_connectAck() == false) else
	}		// if (free_afe.isEmpty() ==  true ...) else
    }			// a new client connects
    else {		// if (listensock.pollNewConnection(s,0,0) == true)

	t2_Assert (listensock.getErrorValue() == ETIMEDOUT);

    }			// if (listensock.pollNewConnection(s,0,0) == true) else


    // looking for commands from connected clients, or acks from back-end
    if (nreadfds > 0) {
	fd_set old_readfds;
	bcopy((const void *) &readfds, (void *) &old_readfds, sizeof(fd_set));

	int n;
	if ((n = select(nreadfds, &readfds, NULL, NULL, &timeout)) < 0) {
	   t2_err_ret(__FILE__, __LINE__, "select error");
	   break;
	}

	if (n == 0) {				// no fd is readable
	   bcopy((const void*) &old_readfds, (void *) &readfds, sizeof(fd_set));
	}
	else {					// some fd is readable

	  if (be_fd > 0) {			// back-end is connected
	    if (FD_ISSET(be_fd, &readfds)) {	// back-end ack is ready
	       bool r = fe.recvQueryAck();
	       t2_Assert (r == true);		// need to handle the case where
						// r == false
	       qtokens++;
	    }
	    else {
	       FD_SET(be_fd, &readfds);
	    }
	  }

	  if (clients.isEmpty() == false) {
	     bool r = clients.moveHead();	// check for clients
	     t2_Assert (r == true);

	     const u_int nclients = clients.getListSize();
	     for (u_int i=0; i<nclients; i++) {
		T2_AppFrontEnd* app = clients.getCurData();
		t2_Assert (app != NULL);
		T2_Socket& appsock = app->getSocket();
		int fd = appsock.getFileDescriptor();
		if (FD_ISSET(fd, &readfds)) {
		   switch (processAppCommand(protocol,*app,conscat,dsinfo,fe))
		   {
		     case t2_acmdError:
			t2_say(cout, "processAppCommand returns Error" << endl);

		     case t2_acmdUnexpectedDisconnect:
			{
			  T2_QBatchList *ql = app->getQueryList();
			  t2_Assert (ql != NULL);
			  ql->clearAppFrontEnd();
			  t2_Assert (app->clearQueryList() == NULL);
			}

			appsock.closeSocket();
			clients.deleteCurWrap();
			t2_Assert (free_afe.isFull() == false);
			free_afe << app;

			FD_CLR(fd, &readfds);

			t2_say(cout,
			       "processAppCommand returns UnexpectedDisconnect or Error" << endl);
			t2_say(cout, "now have " << clients.getListSize() 
				     << " clients" << endl);
			clients_served--;
			break;

		     case t2_acmdDisconnect:
			{
			  T2_QBatchList *ql = app->getQueryList();
			  t2_Assert (ql != NULL);
			  ql->clearAppFrontEnd();
			  t2_Assert (app->clearQueryList() == NULL);
			}

			appsock.closeSocket();
			clients.deleteCurWrap();
			t2_Assert (free_afe.isFull() == false);
			free_afe << app;

			FD_CLR(fd, &readfds);

			clients_served--;
			break;

		     default:
			clients.moveNext();
		   }			// switch

		}			// if (FD_ISSET(fd, &readfds))
		else {
		   FD_SET(fd, &readfds);
		   clients.moveNext();
		}			// if (FD_ISSET(fd, &readfds)) else

	     }				// for (u_int i=0; i<nclients; i++)


	     t2_Assert (nreadfds > 1);
	     if (!FD_ISSET(nreadfds-1, &readfds)) {  // need to adjust nreadfds
		if (clients.getListSize() == 0) {
		   reset_fd(readfds, nreadfds, be_fd);
		}
		else {
		   for (int j=nreadfds-2; j>0; j--) {
		      if (FD_ISSET(j, &readfds)) {
			nreadfds = j+1;
			break;
		      }
		   }
		}
	     }				// if (!FD_ISSET(nreadfds-1, &readfds))

	  }				// if (clients.isEmpty() == false)

	}				// if (n == 0)

    }					// if (nreadfds > 0)

    // send up to qtokens query batches to back-end
    if (qlists.isEmpty() == false && qtokens > 0) {

       const u_int nn = qlists.getListSize();
       for (u_int i=0; i<nn; i++) {
	  T2_QBatchList* ql = qlists.getCurData();
	  t2_Assert (ql != NULL);
	  if (ql->isEmpty() == true) {
	     if (ql->getAppFrontEnd() == NULL) {  	// no associated client
		qlists.deleteCurWrap();
		delete ql;
	     }
	     else {
	        bool r = qlists.moveNextWrap();
	        t2_Assert (r == true);
	     }
	     continue;
	  }

	  T2_QPtrBatch *qbatch = ql->getAndRemoveQBatch();
	  t2_Assert (qbatch != NULL);

	  if (be_fd > 0) {		// only send query if backend exists
	     bool r = fe.sendQBatch_NoWaitForBEAck(*qbatch);
	     t2_Assert (r == true);

	     t2_say(cout, "--- just sent a qbatch to backend (" << (int) ql
			<< "), qbatchlist.size = " << ql->getNumberQBatches()
			<< endl);
	     t2_say3(cout, *qbatch << endl);
	  }

if (be_fd < 0 && dt < 0.)
   dt = t2_dclock() + proctime;


	  delete qbatch;

	  if (ql->isEmpty()) {		// QBatchList becomes empty
	     T2_AppFrontEnd *app = ql->getAppFrontEnd();
	     if (app == NULL) { 	// no associated client
		qlists.deleteCurWrap();
		delete ql;
	     }
	  }
	  else {
	     bool r = qlists.moveNextWrap();
	     t2_Assert (r == true);
	  }

	  if (be_fd > 0) {		// only consume tokens if backend exists
	     if (--qtokens == 0)
	        break;
	  }

       }				// for (u_int i=0; i<nn; i++)
    }					// if (qlists.isEmpty() == false)


if (be_fd < 0 && dt > 0) {
  if (t2_dclock() >= dt) {
     qtokens++;
     if (qtokens < fe.getMax_QTokens())
	dt += proctime;
     else
	dt = -1.;
//     cout << "--- now time = " << t2_dclock()
//	  << ", new dt = " << dt << ", qtokens becomes "
//	  << qtokens << endl;
  }
}


    if (clients_served <= 0		// enough number of clients served
	&& qlists.isEmpty() == true)	// no more pending queries
    {
	notdone = false;
	cout << endl
	     << "no more clients to serve and no more queries to process"
	     << endl;
    }

  } while (notdone == true);

  t2_Assert (qtokens <= fe.getMax_QTokens());
  fe.setQTokens(qtokens);
}


///////////////////////////////////////////////////////////////////////////
//
// processAppCommand
//
///////////////////////////////////////////////////////////////////////////
T2_AppToFECommand
processAppCommand(T2_AppToFEProtocol& protocol, T2_AppFrontEnd& app,
		  T2_ConsCatalogReader& conscat, T2_DSInfoFileReader& dsinfo,
		  T2_FE& fe)
{
  T2_Socket& sock = app.getSocket();
  T2_QBatchList *qlist = app.getQueryList();
  t2_Assert (qlist != NULL);

# ifdef T2_VERBOSE
     char peer[128];
     bool r = sock.getPeerAddress(peer, 128);
     t2_AssertOrDo (r == true,
		    cout << "errval = " << sock.getErrorValue() << endl);
# endif

  T2_AppToFECommand cmd;
  switch (cmd = protocol.readAppCommand(sock)) {
	case t2_acmdSubmit:
		{
		  T2_QPtrBatch& qbatch = qlist->appendEmptyQBatch();
		  sock >> qbatch;

		  t2_say(cout, "--- received query batch from socket " 
				<< sock.getFileDescriptor()
				<< " (" << (int) qlist 
				<< "), QBatchList.Size = " 
				<< qlist->getNumberQBatches() << endl);
		  t2_say3(cout, qbatch << endl);
		}
		if (sock.getErrorValue() == EPIPE)
		   cmd = t2_acmdUnexpectedDisconnect;
		break;

	case t2_acmdDisconnect:

		t2_say(cout, "--- client from " << peer 
				<< " ask for disconnected" << endl);
		protocol.fe_disconnectAck(sock);
		break;

	case t2_acmdPing:

		t2_say(cout, "--- client from " << peer << " just pinged" 
				<< endl);
		protocol.fe_pingAck(sock);
		if (sock.getErrorValue() == EPIPE)
		   cmd = t2_acmdUnexpectedDisconnect;
		break;

	case t2_acmdUnexpectedDisconnect:
		break;

	case t2_acmdInquiry:		// an inquiry on the meta data
		t2_say(cout, "--- client from " << peer
				<< " just sent an inquiry" << endl);
		if (process_inquiry(sock, conscat, dsinfo) == false)
		   if (sock.getErrorValue() == EPIPE)
		      cmd = t2_acmdUnexpectedDisconnect;
		break;

	case t2_acmdError:
		break;

	default:
		t2_err_quit(__FILE__, __LINE__,
			    "unrecognized command %d", (int)cmd);
  }

  return cmd;
}

