/*
  File name: hilbert2d.c 

  Used to find the Hilbert mapping of a point in 2-dimensional space.

  Source: 
     "Fractals for Secondary Key Retrieval", C. Faloutsos, S. Roseman
     University of Maryland, Computer Science Dept., Tech. Rept. CS-TR-2242.

  $Id: hilbert2d.c,v 1.1 2000/02/23 17:11:42 chialin Exp $
*/
#ifndef HILBERT2D_C
#define HILBERT2D_C

#include <stdlib.h>
#include <stdio.h>
#include "hilbert.h"

/* 2D Hilbert Curve --- another version */
static int _RotationTable[4]   = {3,0,0,1};
static int _SenseTable[4]      = {-1,1,1,-1};
static int _QuadTable[4][2][2] = {{{0,1},{3,2}},
				  {{1,2},{0,3}},
				  {{2,3},{1,0}},
				  {{3,0},{2,1}}};
 
/* Returns the index of a point (x,y) in hilbert curve ordering */
unsigned int hilbert2Dshort(unsigned short x, unsigned short y) 
{
  int rotation = 0;
  int sense    = 1;
  unsigned int side, k, num = 0;
  int quad, xbit, ybit;
 
  /* IMPORTANT: side should be a power of 2 */
  side = (unsigned int) 65536;
  for (k=side; k > 0; k >>= 1) { 
    xbit = x/k;
    ybit = y/k;
    x -= k*xbit;
    y -= k*ybit;
    quad = _QuadTable[rotation][xbit][ybit];
    if (sense == -1)
      num += k*k*(3-quad);
    else
      num += k*k*quad;
    rotation += _RotationTable[quad];
    if (rotation >= 4)
      rotation -= 4;
    sense *=_SenseTable[quad];
  }
  return num;
}

/* Comparison routines for quick sort */
int hsf2d_compare_short(const void *a, const void *b)
{
#define H2DCASTS(x) ((_thsf2d_order_short*)x)
  if (H2DCASTS(a)->hsf_idx>H2DCASTS(b)->hsf_idx)
    return 1;
  if (H2DCASTS(a)->hsf_idx<H2DCASTS(b)->hsf_idx)
    return -1;
  return 0;
}

/* ================================================================== */

static unsigned int first_column[4]       = {0,1,2,3};
static unsigned int second_column[4]      = {0,1,3,2};
static unsigned int inv_second[4]         = {0,1,3,2};
static unsigned int third_column[4][2]    = {
  {0,1},{0,2},{0,2},{3,2}
};
static unsigned int fourth_column[4]      = {1,2,2,1};
static int fifth_column[4][2][2] = {
  {{0,1},{1,0}},
  {{1,0},{0,1}},
  {{1,0},{0,1}},
  {{0,-1},{-1,0}}
};

int hilbert2D(unsigned int x, unsigned int y, unsigned int hn[])
{
  unsigned char input[2], output[2], rotated[32][2];
  int           i;
  unsigned int  j,n,k,l;
  int  cmat[2][2] = {{0,1},{1,0}};
  int  cnew[2][2];

  for (i=31;i>=0;i--) {
    input[0] = (x&(1<<i)) ? 1 : 0;
    input[1] = (y&(1<<i)) ? 1 : 0;

    for (j=0;j<2;j++) {
      for (k=0;k<2;k++) 
	if (cmat[j][k]!=0) break;

      if (j==2 || k==2) {
	fprintf(stderr, "There is an error in the matrix.\n");
	fflush(stderr);
	return 0;
      }
      
      if (cmat[j][k]==1)
	output[j] = input[k];
      else
	output[j] = 1-input[k];
    }
    k = (unsigned int) (((int)output[0])*2+((int)output[1]));
    j = inv_second[k];
    rotated[i][1] = (first_column[j]&(1<<1)) ? 1 : 0;
    rotated[i][0] = (first_column[j]&(1<<0)) ? 1 : 0;

    cnew[0][0] = (fifth_column[j][0][0] * cmat[0][0] +
		  fifth_column[j][0][1] * cmat[1][0]);
    cnew[0][1] = (fifth_column[j][0][0] * cmat[0][1] +
		  fifth_column[j][0][1] * cmat[1][1]);

    cnew[1][0] = (fifth_column[j][1][0] * cmat[0][0] +
		  fifth_column[j][1][1] * cmat[1][0]);
    cnew[1][1] = (fifth_column[j][1][0] * cmat[0][1] +
		  fifth_column[j][1][1] * cmat[1][1]);

    cnew[2][0] = (fifth_column[j][2][0] * cmat[0][0] +
		  fifth_column[j][2][1] * cmat[1][0]);
    cnew[2][1] = (fifth_column[j][2][0] * cmat[0][1] +
		  fifth_column[j][2][1] * cmat[1][1]);
    
    for (j=0;j<2;j++)
      for (k=0;k<2;k++)
	cmat[j][k] = cnew[j][k];
  }

  for (i=0;i<2;i++) hn[i] = 0;

  k = 0; n = 1; 
  for (i=0;i<32;i++) {
    for (j=0;j<2;j++) {
      hn[n] |= (((unsigned int)rotated[i][j])<<k);
      k++;
      if (k==32) {
	n--; k = 0;
      }
    }
  }

  return 1;
}


int hsf2d_compare(const void *a, const void *b)
{
#define H2DCAST(x) ((_thsf2d_order*)x)
  if (H2DCAST(a)->v[0]>H2DCAST(b)->v[0]) return  1;
  if (H2DCAST(a)->v[0]<H2DCAST(b)->v[0]) return -1;
  if (H2DCAST(a)->v[1]>H2DCAST(b)->v[1]) return  1;
  if (H2DCAST(a)->v[1]<H2DCAST(b)->v[1]) return -1;
  return 0;
}



#endif /* HILBERT2D_C */


