/*
  File name: hilbert3d.c 

  Used to find the Hilbert mapping of a point in 3-dimensional space.

  Source: 
     "Fractals for Secondary Key Retrieval", C. Faloutsos, S. Roseman
     University of Maryland, Computer Science Dept., Tech. Rept. CS-TR-2242.

  $Id: hilbert3d.c,v 1.1 2000/02/23 17:11:42 chialin Exp $
*/
#ifndef HILBERT3D_C
#define HILBERT3D_C

#include <stdlib.h>
#include <stdio.h>
#include "hilbert.h"

static unsigned int first_column[8]       = {0,1,2,3,4,5,6,7};
static unsigned int second_column[8]      = {0,1,3,2,6,7,5,4};
static unsigned int inv_second[8]         = {0,1,3,2,7,6,4,5};
static unsigned int third_column[8][2]    = {
  {0,1},{0,2},{0,4},{5,4},
  {0,2},{3,2},{0,4},{5,4}
};
static unsigned int fourth_column[8]      = {1,2,4,1,2,1,4,1};
static int fifth_column[8][3][3] = {
  {{0,0,1},{0,1,0},{1,0,0}},
  {{0,1,0},{0,0,1},{1,0,0}},
  {{1,0,0},{0,1,0},{0,0,1}},
  {{0,0,-1},{0,1,0},{-1,0,0}},
  {{0,1,0},{1,0,0},{0,0,1}},
  {{0,0,-1},{0,-1,0},{1,0,0}},
  {{1,0,0},{0,1,0},{0,0,1}},
  {{0,0,-1},{0,1,0},{-1,0,0}}
};

int hilbert3D(unsigned int x, unsigned int y, unsigned int z, 
	      unsigned int hn[])
{
  unsigned char input[3], output[3], rotated[32][3];
  int           i;
  unsigned int  j,n,k,l;
  int           cmat[3][3] = {{0,0,1},{0,1,0},{1,0,0}};
  int           cnew[3][3];

  for (i=31;i>=0;i--) {
    input[0] = (x&(1<<i)) ? 1 : 0;
    input[1] = (y&(1<<i)) ? 1 : 0;
    input[2] = (z&(1<<i)) ? 1 : 0;

    for (j=0;j<3;j++) {
      for (k=0;k<3;k++) 
	if (cmat[j][k]!=0) break;

      if (j==3 || k==3) {
	fprintf(stderr, "There is an error in the matrix.\n");
	return 0;
      }
      
      if (cmat[j][k]==1)
	output[j] = input[k];
      else
	output[j] = 1-input[k];
    }

    k = (unsigned int) (((int)output[0])*4+
			((int)output[1])*2+
			((int)output[2]));
    j = inv_second[k];
    rotated[i][2] = (first_column[j]&(1<<2)) ? 1 : 0;
    rotated[i][1] = (first_column[j]&(1<<1)) ? 1 : 0;
    rotated[i][0] = (first_column[j]&(1<<0)) ? 1 : 0;

    cnew[0][0] = (fifth_column[j][0][0] * cmat[0][0] +
		  fifth_column[j][0][1] * cmat[1][0] +
		  fifth_column[j][0][2] * cmat[2][0]);
    cnew[0][1] = (fifth_column[j][0][0] * cmat[0][1] +
		  fifth_column[j][0][1] * cmat[1][1] +
		  fifth_column[j][0][2] * cmat[2][1]);
    cnew[0][2] = (fifth_column[j][0][0] * cmat[0][2] +
		  fifth_column[j][0][1] * cmat[1][2] +
		  fifth_column[j][0][2] * cmat[2][2]);

    cnew[1][0] = (fifth_column[j][1][0] * cmat[0][0] +
		  fifth_column[j][1][1] * cmat[1][0] +
		  fifth_column[j][1][2] * cmat[2][0]);
    cnew[1][1] = (fifth_column[j][1][0] * cmat[0][1] +
		  fifth_column[j][1][1] * cmat[1][1] +
		  fifth_column[j][1][2] * cmat[2][1]);
    cnew[1][2] = (fifth_column[j][1][0] * cmat[0][2] +
		  fifth_column[j][1][1] * cmat[1][2] +
		  fifth_column[j][1][2] * cmat[2][2]);

    cnew[2][0] = (fifth_column[j][2][0] * cmat[0][0] +
		  fifth_column[j][2][1] * cmat[1][0] +
		  fifth_column[j][2][2] * cmat[2][0]);
    cnew[2][1] = (fifth_column[j][2][0] * cmat[0][1] +
		  fifth_column[j][2][1] * cmat[1][1] +
		  fifth_column[j][2][2] * cmat[2][1]);
    cnew[2][2] = (fifth_column[j][2][0] * cmat[0][2] +
		  fifth_column[j][2][1] * cmat[1][2] +
		  fifth_column[j][2][2] * cmat[2][2]);

    
    for (j=0;j<3;j++)
      for (k=0;k<3;k++)
	cmat[j][k] = cnew[j][k];
  }

  for (i=0;i<3;i++) hn[i] = 0;

  k = 0; n = 2; 
  for (i=0;i<32;i++) {
    for (j=0;j<3;j++) {
      hn[n] |= (((unsigned int)rotated[i][j])<<k);
      k++;
      if (k==32) {
	n--; k = 0;
      }
    }
  }

  return 1;
}

int hsf3d_compare(const void *a, const void *b)
{
#define H3DCAST(x) ((_thsf3d_order*)x)
  if (H3DCAST(a)->v[0]>H3DCAST(b)->v[0]) return  1;
  if (H3DCAST(a)->v[0]<H3DCAST(b)->v[0]) return -1;
  if (H3DCAST(a)->v[1]>H3DCAST(b)->v[1]) return  1;
  if (H3DCAST(a)->v[1]<H3DCAST(b)->v[1]) return -1;
  if (H3DCAST(a)->v[2]>H3DCAST(b)->v[2]) return  1;
  if (H3DCAST(a)->v[2]<H3DCAST(b)->v[2]) return -1;
  return 0;
}

#endif /* HILBERT3D_C */

