
/* $Id: t2dld_mover.h,v 1.10 2000/02/21 07:27:41 chialin Exp $ */

/* t2dld_mover.h */

#ifndef _T2DLD_MOVER_H_
#define _T2DLD_MOVER_H_

#include	"t2dld_header.h"
#include	"t2_bits.h"
#include	"t2_txtfile.h"
#include	"t2dld_base.h"
#include	"t2_rtree.h"


/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_MoverInstruction
//
/////////////////////////////////////////////////////////////////////////////
class T2DLD_MoverInstruction {
public:
  T2DLD_MoverInstruction(u_int l=200): clusters(l) {};
  ~T2DLD_MoverInstruction() {};

  // avoid to use copy constructor and operator =, since they are very
  // inefficient;
  T2DLD_MoverInstruction(const T2DLD_MoverInstruction& m);
  T2DLD_MoverInstruction& operator = (const T2DLD_MoverInstruction& m);

  // query methods
  u_int getNumberDataClusters() const { return clusters.getNumberElements(); };

  // const access methods
  const T2DLD_DataCluster* operator[] (u_int i) const { return clusters[i]; };

  // append a data cluster pointer to the end of the instruction
  T2DLD_MoverInstruction& operator << (const T2DLD_DataCluster* c);

  // insert a data cluster pointer to the i-th position of the instruction
  // and push instruction i, i+1, i+2, ... one position back;
  // method fails and returns false if i >= number of data cluster pointers
  // already in the instruction
  bool insert(u_int i, const T2DLD_DataCluster* c);

  // output to ostream
  ostream& write(ostream& os) const;

protected:
  T2_VArray<const T2DLD_DataCluster*> clusters;

private:
};


inline T2DLD_MoverInstruction&
T2DLD_MoverInstruction::operator << (const T2DLD_DataCluster* c)
{
  t2_Assert (c != NULL);
  clusters << c;
  return *this;
}

inline ostream&
operator << (ostream& os, const T2DLD_MoverInstruction& m)
{
  return m.write(os);
}



/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_Mover
//
/////////////////////////////////////////////////////////////////////////////
class T2DLD_Mover {
public:
  T2DLD_Mover();
  ~T2DLD_Mover();

  // const access methods
  u_int getLogicalDiskID() const { return did; };

  // query methods for the current dataset being loaded
  u_int getNumberDimensions() const { return ndims; };

  // initialize with d = assigned logical disk id
  bool init(u_int d);

  // prepare to load a dataset into T2 disks;
  // fid is the dataset file id
  // return true on success, false otherwise
  bool prepareLoadOneDataset(u_int ndims, u_int fid, const char* datasetfname,
			     const char* indexfname);

  // move data according to given instructions starting with instructs[start];
  // tapefiles give the filenames of the input tape files,
  // ds_id and diskfile_prefix give the dataset id and the file prefix for 
  // the output disk files, idx_id and diskindex_prefix give the index id
  // and the file prefix for the on-disk index for the dataset,
  // and instructs[start], instructs[start+1], ...,
  // instructs[start+getNumberDisks()-1] gives the instructions for moving
  // data onto disks.
  bool moveData(const T2_VArray<char*>& tapefiles, 
		const T2DLD_MoverInstruction& instr);

  // tell the mover that no more data to move for this dataset
  bool endLoadOneDataset();

  // output to ostream
  ostream& write(ostream& os) const;

  // do not use copy constructor and operator =
  T2DLD_Mover(const T2DLD_Mover& m);
  T2DLD_Mover& operator = (const T2DLD_Mover& m);

protected:
  u_int did;			// assigned logical disk id

  // the information regarding the current dataset the mover is loading
  int ds_fd;			// file descriptor for dataset file
  u_int ds_fid;			// dataset file id for the dataset file
  T2_RTree rtree;		// R-tree index
  u_int ndims;			// number of dims for bounding boxes
  off_t ds_offset;		// current offset of dataset file

  // close fd and rtree
  bool close();

private:
};


inline bool
T2DLD_Mover::close()
{
  if (ds_fd > 0) {
     ::close(ds_fd);
     ds_fd = -1;
  }
  if ((const void*) rtree != NULL) {
     if (rtree.close() == 0) {
	t2_err_ret(__FILE__, __LINE__, "ERROR: fail to close rtree");
	return false;
     }
  }
  t2_Assert (!rtree);
  return true;
}

inline 
T2DLD_Mover::T2DLD_Mover()
	: did(0), ds_fd(-1), ds_fid(0), rtree(), ndims(0), ds_offset(0)
{
}

inline 
T2DLD_Mover::~T2DLD_Mover()
{
  close();
}

inline bool
T2DLD_Mover::init(u_int d)
{
  did = d;
  return true;
}

inline bool
T2DLD_Mover::endLoadOneDataset()
{
  return close();
}

inline 
T2DLD_Mover::T2DLD_Mover(const T2DLD_Mover& m)
	: did(m.did), ds_fd(m.ds_fd), ds_fid(m.ds_fid), rtree(),
	  ndims(m.ndims), ds_offset(m.ds_offset)
{
  t2_err_quit(__FILE__, __LINE__,
		"ERROR: do not use T2DLD_Mover copy constructor");
}

inline T2DLD_Mover&
T2DLD_Mover::operator =(const T2DLD_Mover& m)
{
  t2_err_quit(__FILE__, __LINE__,
		"ERROR: do not use T2DLD_Mover operator =");
  return *this;
}

inline ostream&
operator << (ostream& os, const T2DLD_Mover& m)
{
  return m.write(os);
}



#endif	/* ! _T2DLD_MOVER_H_ */
