
/* $Id: dld-mover.C,v 1.13 2000/02/23 17:10:18 chialin Exp $ */

/* dld-mover.C */

#include	<stdio.h>
#include	<string.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	"t2dld_mover.h"

#ifndef T2_SYS_SUPPORT_USLEEP
extern "C" int usleep(u_int);
#endif

extern int _g_mynode;

////////////////////////////////////////////////////////////////////////////
//
// a helper function to rename a file from foo to foo.bak
//
// note that the given file is assumed to exist -- that is, no checking
// is done in this function
//
////////////////////////////////////////////////////////////////////////////
bool
rename_bak(const char* fname) 
{
  t2_Assert (fname != NULL);
  char bakfname[512];
  sprintf(bakfname, "%s.bak", fname);
  if (rename(fname, bakfname) != 0)	// error
     return false;
  return true;
}


////////////////////////////////////////////////////////////////////////////
//
// T2DLD_MoverInstruction methods
//
////////////////////////////////////////////////////////////////////////////
bool
T2DLD_MoverInstruction::insert(u_int i, const T2DLD_DataCluster* c)
{
  t2_Assert (c != NULL);
  const u_int nclusters = getNumberDataClusters();
  if (i >= nclusters)			// index i out of range
     return false;

  const T2DLD_DataCluster *incoming_p = c;
  for (u_int j=i; j<nclusters; j++) {
     const T2DLD_DataCluster *outgoing_p = clusters[j];
     clusters[j] = incoming_p;
     incoming_p = outgoing_p;
  }
  clusters << incoming_p;
  return true;
}


T2DLD_MoverInstruction::T2DLD_MoverInstruction(const T2DLD_MoverInstruction& m)
		: clusters(m.clusters.segmentSize())
{
  for (u_int i=0; i<m.getNumberDataClusters(); i++) {
    t2_Assert (m[i] != NULL);
    clusters << m[i];
  }
}


T2DLD_MoverInstruction&
T2DLD_MoverInstruction::operator = (const T2DLD_MoverInstruction& m)
{
  clusters.clear();
  for (u_int i=0; i<m.getNumberDataClusters(); i++) {
    t2_Assert (m[i] != NULL);
    clusters << m[i];
  }
  return *this;
}


ostream&
T2DLD_MoverInstruction::write(ostream& os) const
{
  for (u_int i=0; i<getNumberDataClusters(); i++) {
     t2_Assert (clusters[i] != NULL);
     os << *(clusters[i]);
  }
  return os;
}



////////////////////////////////////////////////////////////////////////////
//
// some help functions for accessing tape files
//
////////////////////////////////////////////////////////////////////////////
int
open_tapefile(const char* path, int flag)
{
  int fd;
  for (u_int xx=0; xx<10; xx++) {
     if ((fd = open(path, flag)) > 0) {
        return fd;
     }
     usleep(100);
  }
  return -1;
}


ssize_t
read_tapefile(int fd, char* buf, off_t offset, size_t s)
{
  if (fd < 0)
     return -1;

  lseek(fd, offset, SEEK_SET);
  return read(fd, buf, s);
}


////////////////////////////////////////////////////////////////////////////
//
// T2DLD_Mover methods
//
////////////////////////////////////////////////////////////////////////////
bool
T2DLD_Mover::
prepareLoadOneDataset(u_int dim, u_int fid,
		      const char* datasetfname, const char* indexfname)
{
  close();			// close dataset file and index file, if open
  ds_fid = fid;
  ndims = dim;
  ds_offset = 0;

  if (datasetfname != NULL) {	// open dataset file
     ifstream ifs(datasetfname, ios::in);
     if (ifs) {			// file exists
	ifs.close();
	if (rename_bak(datasetfname) == false) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: file '%s' already exists; cannot rename it to '%s.bak'",
		datasetfname, datasetfname);
	   return false;
	}
	t2_say(cout, "WARNING: dataset file " << datasetfname
			<< " exists, renamed to " << datasetfname << ".bak" 
			<< endl);
     }
     ds_fd = open(datasetfname, O_WRONLY | O_CREAT | O_TRUNC, 0644);
     if (ds_fd < 0) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot create dataset file '%s'.", datasetfname);
	return false;
     }
  }

  if (indexfname != NULL) {	// open index file
     ifstream ifs(indexfname, ios::in);
     if (ifs) {			// file exists
	ifs.close();
	if (rename_bak(indexfname) == false) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: file '%s' already exists; cannot rename it to '%s.bak'",
		indexfname, indexfname);
	   return false;
	}
	t2_say(cout, "WARNING: index file " << indexfname
			<< " exists, renamed to " << indexfname << ".bak" 
			<< endl);
     }
     char ifaux_name[256];
     sprintf(ifaux_name, "%s.aux", indexfname);
     ifs.open(ifaux_name, ios::in);
     if (ifs) {			// file exists
	ifs.close();
	if (rename_bak(ifaux_name) == false) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: file '%s' already exists; cannot rename it to '%s.bak'",
		ifaux_name, ifaux_name);
	   return false;
	}
	t2_say(cout, "WARNING: index file " << ifaux_name
			<< " exists, renamed to " << ifaux_name << ".bak" 
			<< endl);
     }

     // hard-coded to use R-tree; initialize R-tree
     t2_Assert (!rtree);
     rtree.initNumberDimensions(ndims);

     // open index file 
     if (rtree.create(indexfname) == 0) {	// create index fails
	t2_err_quit(__FILE__, __LINE__,
		"ERROR: cannot create index file '%s'.", indexfname);
	close();
	return false;
     }
  }

  return true;
}


bool
T2DLD_Mover::moveData(const T2_VArray<char*>& tapefiles, 
		      const T2DLD_MoverInstruction& instr)
{
  u_int i;

  if (ds_fd < 0) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: no output dataset file is open.");
     return false;
  }
  if (!rtree) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: no R-tree is open.");
     return false;
  }

  // The tape files will be open as they are needed during actual data
  // transfer. The array in_fd[] will be used to store the file descriptors
  // for the tapefiles.
  const u_int ntapefiles = tapefiles.getNumberElements();
  T2_Array<int> in_fd(ntapefiles);
  for (i=0; i<ntapefiles; i++)
     in_fd[i] = 0;

  // read buffer -- will be allocated on-demand
  char* buf = NULL;
  size_t bufsize = 0;

  for (u_int n=0; n<instr.getNumberDataClusters(); n++) {
	const T2DLD_DataCluster* cluster = instr[n];
	t2_Assert (cluster != NULL);

	// move data from tapes to disks, one data block at a time
	size_t cluster_size = 0;
	for (u_int j=0; j<cluster->getNumberDataBlocks(); j++) {
	   const T2DLD_DataBlock& dblock = cluster->getDataBlock(j);

	   if (dblock.getData() != NULL)	// data already avaiable
		::write(ds_fd, dblock.getData(), dblock.getSize());
	   else {
		u_int in_fid = dblock.getDataFileID();
		t2_Assert (in_fid < ntapefiles);
		if (in_fd[in_fid] == 0) {	// tape file not open yet
		   if ((in_fd[in_fid]=open_tapefile(tapefiles[in_fid],O_RDONLY))
			  < 0)
		   {
		      // open failed
		      t2_err_quit(__FILE__, __LINE__,
			  "ERROR: cannot open tapefile %s", tapefiles[in_fid]);
		   }
		}

		// is buf large enough to hold data ?
		if (bufsize < dblock.getSize()) {
		   delete [] buf;
		   bufsize = dblock.getSize() + 256;
		   t2_Alloc(buf, bufsize, char);
		}

		ssize_t ss = read_tapefile(in_fd[in_fid], buf,
					   dblock.getOffset(),
					   dblock.getSize());

		// core dump for now if sizes don't match
		t2_Assert (ss == (ssize_t) dblock.getSize());
		::write(ds_fd, buf, ss);
	   }		// if (dblock.getData() != NULL) else

	   cluster_size += dblock.getSize();
	}		// for (j=0; j<cluster->getNumberDataBlocks(); j++)


	// update on-disk index entry
	if (rtree.insert(cluster->getMBR(),	// mbr
			 ds_fid,		// dataset file id
			 ds_offset, 		// offset
			 cluster_size,		// total cluster size
			 cluster->getUsrData())	// user data, if any
	   == 0)				// insertion fails
	{
	   t2_err_quit(__FILE__, __LINE__,
		  "ERROR: cannot insert R-tree entry for cluster %d", n);
	}


	// update offset counter
	ds_offset += cluster_size;

  }			// for (n=0; n<instr.getNumberDataClusters(); n++)

  // close any open tapefile
  for (i=0; i<ntapefiles; i++)
     if (in_fd[i] != 0)
        ::close(in_fd[i]);

  // release memory
  delete [] buf;

  return true;
}


ostream& 
T2DLD_Mover::write(ostream& os) const
{
  os << "mover assigned logical disk id= " << did << endl
     << "ds_fd=" << ds_fd << " ds_fid=" << ds_fid
     << " ndims=" << ndims << " ds_offset=" << ds_offset
     << endl;
  return os;
}

