/*
  File name: mc_regions.H

  Class for regions and set of regions. 
*/
#ifndef MC_REGIONS_H
#define MC_REGIONS_H

#include <stdlib.h>
#include "regionX.h"
#include "mc_assert.H"

/*
 * Region definition.
 */
class MC_Region {
private:
  inline void updateRegion(T_RegionType         type,
			   T_RegionDistribution dist,      
			   T_MemOp              memType,
			   int                  size,
			   void                 *userData) {
    this->type     = type;
    this->dist     = dist;
    this->memType  = memType;
    this->size     = size;
    this->userData = userData;
  };

  inline void updateRegion(int size, void *userData) {
    updateRegion(R_Enum,R_replicated,Mem_Copy,size,userData);
  };

  inline void updateRegion(MC_Region& rgn) {
    updateRegion(rgn.type,rgn.dist,rgn.memType,rgn.size,rgn.userData);
  };
public:
  // Member elements ... 
  T_RegionType         type;
  T_RegionDistribution dist;
  T_MemOp              memType;
  int                  size;
  void                 *userData;

  // Constructors ... 
  MC_Region() {
    updateRegion(R_Enum,R_replicated,Mem_Copy,0,0);
  };
  
  MC_Region(int size, void *userData) {
    updateRegion(R_Enum,R_replicated,Mem_Copy,size,userData);
  };

  MC_Region(T_RegionType         type,
	    T_RegionDistribution dist,      
	    T_MemOp              memType,
	    int                  size,
	    void                 *userData) {
    updateRegion(type,dist,memType,size,userData); 
  };

  MC_Region(MC_Region &rgn) {
    updateRegion(rgn.type,rgn.dist,rgn.memType,rgn.size,rgn.userData);
  };

  ~MC_Region() {  };

  MC_Region& operator=(MC_Region& region) {
    this->type     = region.type;
    this->dist     = region.dist;
    this->memType  = region.memType;
    this->size     = region.size;
    this->userData = region.userData;
    return *this;
  };
};

/*
 * Class definition for set of regions.
 */
class MC_SetOfRegions {
private:
  // This structure should be consistent with definition of MC_Region 
  // member elements. 
  struct Region {
    T_RegionType         type;
    T_RegionDistribution dist;
    T_MemOp              memType;
    int                  size;
    void                 *userData;
  } *rgn;

  int           iter, nelems;
  setOfRegion_s *sr;
  MC_Region     mc_reg;

  void allocSetOfRegions() { sr = Alloc_setOfRegion(); MC_ASSERT(sr!=NULL); };
  void freeSetOfRegions()  { 
    if (sr==NULL)
      return;
    for (reset();!end();advance()) {
      if (sr->regions[iter]!=NULL) 
	free(sr->regions[iter]);
    }
    Free_setOfRegion(sr); 
    sr=NULL;
  };

  int can_free;
public:
  MC_SetOfRegions() { allocSetOfRegions(); can_free=1; };

  MC_SetOfRegions(MC_SetOfRegions& setR) { this->sr = setR.sr; can_free=0; };
    
  ~MC_SetOfRegions() {
    if (can_free)
      freeSetOfRegions();
  };

  void addRegion(MC_Region& region) {
    // rgn is deallocated by free in Meta-Chaos so I should 
    // use malloc ... 
    rgn = (Region*)malloc(sizeof(Region));
    MC_ASSERT(rgn!=NULL);
    rgn->type     = region.type;
    rgn->dist     = region.dist;
    rgn->memType  = region.memType;
    rgn->size     = region.size;
    rgn->userData = region.userData;
    Add_Region_setOfRegion((region_s*)rgn,sr);
  };

  void addRegion(T_RegionType         type,
                 T_RegionDistribution dist,
                 T_MemOp              memType,
                 int                  size,
                 void                 *userData) {
    rgn = (Region*)malloc(sizeof(Region));
    MC_ASSERT(rgn!=NULL);
    rgn->type     = type;
    rgn->dist     = dist;
    rgn->memType  = memType;
    rgn->size     = size;
    rgn->userData = userData;
    Add_Region_setOfRegion((region_s*)rgn,sr);
  };

  void addRegion(int size, void *userData) {
    addRegion(R_Enum,R_replicated,Mem_Copy,size,userData);
  };

  int getSizeSetOfRegions() {
    if (sr!=NULL) 
      return (int)NbRegionSetOfRegion(sr);
    return 0;
  };

  setOfRegion_s* getSetOfRegions() {
    MC_ASSERT(sr!=NULL);
    return sr;
  };

  // Iterator methods .... 
  void reset() {
    iter = 0; nelems = getSizeSetOfRegions();
  };

  void advance() {
    if (iter<nelems) iter++;
  };

  int getRegionGlobalOffset() {
    if (sr!=NULL && iter<nelems) return sr->BeginGlobalOffset[iter];
    return -1;
  }

  MC_Region& getRegion() {
    if (sr!=NULL && iter<nelems) {
      rgn             = (Region*)(sr->regions[iter]);
      mc_reg.type     = rgn->type;
      mc_reg.dist     = rgn->dist;
      mc_reg.memType  = rgn->memType;
      mc_reg.size     = rgn->size;
      mc_reg.userData = rgn->userData;
    }
    return mc_reg;
  };

  int end() {
    if (iter>=nelems) return 1;
    return 0;
  };
};


#endif /* MC_REGIONS_H */
