
/* $Id: rtree.C,v 1.1 2000/03/08 02:54:46 chialin Exp $ */

/* rtree.C */

#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	"t2_txtfile.h"
#include	"t2_rtree.h"

/////////////////////////////////////////////////////////////////////////////
//
// get_aux_fname(fname, aux_fname): write into aux_fname the corresponding
// aux index filename for R-tree stored in file fname
//
/////////////////////////////////////////////////////////////////////////////
inline void
get_aux_fname(const char* fname, char *aux_fname)
{
  sprintf(aux_fname, "%s.aux", fname);
}


/////////////////////////////////////////////////////////////////////////////
//
// conversion from T2_Box into a double array
//
/////////////////////////////////////////////////////////////////////////////
static void
box2doublearray(const T2_Box& mbr, const u_int ndims, double* coord)
{
  t2_AssertOrDo (ndims == mbr.getNumberDimensions(),
		 cout << "ndims = " << ndims << " mbr.ndims = "
		      << mbr.getNumberDimensions() << endl);
  t2_Assert (coord != NULL);
  const T2_Point& low = mbr.getLow();
  const T2_Point& high = mbr.getHigh();

  for (u_int i=0, j=0; i<ndims; i++) {
     coord[j++] = low[i];
     coord[j++] = high[i];
  }
}


static void
doublearray2box(const u_int ndims, const double* coord, T2_Box& mbr)
{
  t2_Assert (ndims == mbr.getNumberDimensions());
  t2_Assert (coord != NULL);
  T2_Point& low = mbr.getLow();
  T2_Point& high = mbr.getHigh();

  for (u_int i=0, j=0; i<ndims; i++) {
     low[i] = coord[j++];
     high[i] = coord[j++];
  }
}


/////////////////////////////////////////////////////////////////////////////
//
// look for an uncommented string from file fd; linebuf is a char buffer of
// buflen bytes provided by the caller and mark is the special character that
// turns the rest of the line into comment;
// on success, this function returns the length of the string and put the
// string in linebuf; otherwise, it returns 0;
//
// if the string found has more than buflen bytes, then this function only
// returns the first buflen bytes;
// for correct operation, line[] should be longer than any non-comment
// string in the file.
//
/////////////////////////////////////////////////////////////////////////////
size_t
readNextUncommentString(int fd, char *line, size_t buflen, const char mark)
{
  t2_Assert (line != NULL);
  t2_Assert (buflen > 0);

  bool in_comment = false,
       found = false;
  char lastch = ' ';
  while (found == false && read(fd, line, 1) > 0) {
    if (in_comment == true) {           // look for end of line 
        if (line[0] == '\n') {
           in_comment = false;
           lastch = '\n';
        }
        continue;
    }

    if (isspace(line[0]))               // keep looking
        continue;

    if (line[0] == mark && isspace(lastch)) {   // beginning of comment
        in_comment = true;
        continue;
    }
    found = true;
  }

  if (found == true) {
     u_int i;
     char *p = line+1;
     for (i=1; i<buflen; i++, p++) {
        if (read(fd, p, 1) <= 0 || isspace(*p)) // end of file or end of string
           break;
     }  
     return i;
  }
  return 0;
}


/////////////////////////////////////////////////////////////////////////////
//
// parse_leafnode(char* buf, size_t usr_data_len,
//		  u_int& fid, off_t& offset, size_t& size, T2_UsrArg& arg)
//
// read (fid, offset, size) out of the array pointed to by buf, and then
// read usr_data_len bytes into arg
//
/////////////////////////////////////////////////////////////////////////////
static void
parse_leafnode(char* buf, size_t usr_data_len,
		u_int& fid, off_t& offset, size_t& size, T2_UsrArg& arg)
{
  t2_Assert (arg.getBufferSize() == 0);
  t2_Assert (buf != NULL);
  char *p = buf;

  // read fid, offset and cluster size from buf
  memcpy(&fid, p, sizeof(fid));			p += sizeof(fid);
  memcpy(&offset, p, sizeof(offset));		p += sizeof(offset);
  memcpy(&size, p, sizeof(size));		p += sizeof(size);

  // write user data from buf
  if (usr_data_len > 0) {			// user data exists
     size_t ss = arg.allocBuffer(usr_data_len);
     t2_Assert (ss == usr_data_len);
     t2_Assert (arg.getDataSize() == 0);
     memcpy(arg.getStartOfData(), p, usr_data_len);
     arg.setDataSize(usr_data_len);
  }
}


/////////////////////////////////////////////////////////////////////////////
//
// T2_RTree methods
//
/////////////////////////////////////////////////////////////////////////////
void
T2_RTree::initNumberDimensions(const unsigned int dim)
{
  ndims = dim;
  delete [] coord;
  delete [] qrycoord;
  if (dim == 0) {
     coord = qrycoord = NULL;
  }
  else {
     t2_Alloc (coord, dim*2, double);
     t2_Alloc (qrycoord, dim*2, double);
  }

  rtree.init(dim);
}


int
T2_RTree::open(const char* fname, int mode)
{
  if (rtree.open(fname, mode) == 0) {	// open fails
     return 0;
  }

  t2_Assert (max_leaf_size == 0);

  char aux_fname[512];			// get name of aux index file
  get_aux_fname(fname, aux_fname);

  switch (mode) {
	case W_CREATE_FILE: {
	  if (ndims == 0) {		// do not allow to create a 0-dim r-tree
	     t2_err_ret(__FILE__, __LINE__,
			"ERROR: cannot create an R-tree of 0-dim");
	     return 0;
	  }
	  ifstream ifs(aux_fname);
	  if (ifs) {			// aux index file exists
	     ifs.close();
	     unlink(aux_fname);
	  }
	  aux_ofs.open(aux_fname, ios::out);
	  if (!aux_ofs) {		// open fails
	     t2_err_ret(__FILE__, __LINE__,
			"ERROR: cannot create aux index file %s", aux_fname);
	     rtree.close();
	     return 0;
	  }
	  t2_Assert (max_leaf_size == 0);
	  break;
	}
	case W_APPEND_FILE: {
	  T2_TxtFile ifs('#');
	  ifs.open(aux_fname);
	  if (!ifs) {
	     t2_err_ret(__FILE__, __LINE__,
			"ERROR: cannot open aux index file %s", aux_fname);
	     rtree.close();
	     return 0;
	  }
	  ifs >> ndims >> max_leaf_size;
	  if (ifs.eof() == true) {
	     t2_err_ret(__FILE__, __LINE__,
			"ERROR: aux index file %s contains no data", aux_fname);
	     ifs.close();
	     rtree.close();
	     return 0;
	  }
	  ifs.close();
	  aux_ofs.open(aux_fname, iostream::out);  // overwrite old aux file
	  if (!aux_ofs) {		// open fails
	     t2_err_ret(__FILE__, __LINE__,
			"ERROR: cannot create aux index file %s", aux_fname);
	     rtree.close();
	     return 0;
	  }
	  initNumberDimensions(ndims);
	  break;
	}
	case W_READ_FILE: {
	  T2_TxtFile ifs('#');
	  ifs.open(aux_fname);
	  if (!ifs) {
	     t2_err_ret(__FILE__, __LINE__,
			"ERROR: cannot open aux index file %s", aux_fname);
	     rtree.close();
	     return 0;
	  }
	  ifs >> ndims >> max_leaf_size;
	  if (ifs.eof() == true) {
	     t2_err_ret(__FILE__, __LINE__,
		"ERROR: aux index file %s does not contain enough data",
		aux_fname);
	     ifs.close();
	     rtree.close();
	     return 0;
	  }
	  initNumberDimensions(ndims);
	  ifs.close();
	  break;
	}
  }					// switch (mode)

  if (max_leaf_size > buflen) {		// needs a larger buffer
     buflen = max_leaf_size;
     delete [] buf;
     t2_Alloc (buf, max_leaf_size, char);
  }

  return 1;
}


int
T2_RTree::close()
{
  if (rtree.close() == 0) {		// close fails
     return 0;
  }
  if (aux_ofs) {
     t2_Assert (rtree.getOpenMode() == W_CREATE_FILE || 
		rtree.getOpenMode() == W_APPEND_FILE);
     aux_ofs << ndims << " " << max_leaf_size;
     aux_ofs.close();
  }
  else {
     t2_Assert (rtree.getOpenMode() == W_READ_FILE);
  }
  max_leaf_size = 0;
  return 1;
}


int
T2_RTree::insert(const T2_Box& mbr, u_int fid, off_t offset, size_t size,
		 const T2_UsrArg& arg)
{
  box2doublearray(mbr, ndims, coord);

  // check if buf is large enough to hold every thing
  size_t leaf_realsize = leaf_size + arg.getDataSize();
  if (leaf_realsize > buflen) {
     delete [] buf;
     buflen = leaf_realsize + 256;		// allocate more than enough
     t2_Alloc(buf, buflen, char);
  }

  char *p = buf;

  // write fid, offset and cluster size into buf
  memcpy(p, &fid, sizeof(fid));			p += sizeof(fid);
  memcpy(p, &offset, sizeof(offset));		p += sizeof(offset);
  memcpy(p, &size, sizeof(size));		p += sizeof(size);

  // write user data into buf
  if (arg.getDataSize() > 0)
     memcpy(p, arg.getStartOfData(), arg.getDataSize());

  if (rtree.insert(coord, buf, leaf_realsize) == 0)	// insert fails
     return 0;

  if (leaf_realsize > max_leaf_size)
     max_leaf_size = leaf_realsize;
  return 1;
}


int 
T2_RTree::initSearch(const T2_Box& qb)
{
  box2doublearray(qb, ndims, qrycoord);
  return rtree.initSearch(qrycoord);
}


int
T2_RTree::getNext(T2_Box& mbr, u_int& fid, off_t& offset, size_t& size,
		 T2_UsrArg& arg)
{
  size_t leaf_realsize = buflen;
  u_int dim;
  int t = rtree.getNext(coord, dim, buf, leaf_realsize);
  if (t <= 0)			// error or no more data found
     return t;

  t2_Assert (dim == ndims);
  if (mbr.getNumberDimensions() != ndims)
     mbr.setNumberDimensions(ndims);
  doublearray2box(ndims, coord, mbr);

  arg.releaseBuffer();

  // read fid, offset, size and arg from buf
  t2_Assert (leaf_realsize >= leaf_size);
  parse_leafnode(buf, leaf_realsize - leaf_size, fid, offset, size, arg);

  return 1;
}



/////////////////////////////////////////////////////////////////////////////
//
// T2_StaticRTree methods
//
/////////////////////////////////////////////////////////////////////////////
void
T2_StaticRTree::initNumberDimensions(const unsigned int dim)
{
  ndims = dim;
  delete [] coord;
  delete [] qrycoord;
  if (dim == 0) {
     coord = qrycoord = NULL;
  }
  else {
     t2_Alloc (coord, dim*2, double);
     t2_Alloc (qrycoord, dim*2, double);
  }

  W_StaticRTree::init(dim);
}


int
T2_StaticRTree::open(const char *fname)
{
  if (W_StaticRTree::open(fname) == 0)		// open fails
     return 0;

  char aux_fname[512];				// get name of aux index file
  get_aux_fname(fname, aux_fname);

  T2_TxtFile aux_ifs('#');
  aux_ifs.open(aux_fname);
  if (!aux_ifs) {				// open fails
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open aux index file %s", aux_fname);
     W_StaticRTree::close();
     return 0;
  }

  size_t max_leaf_size;
  aux_ifs >> ndims >> max_leaf_size;		// get # dims, max leaf size
  if (aux_ifs.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: aux index file %s contains no data", aux_fname);
     W_StaticRTree::close();
     return 0;
  }

  if (ndims == 0) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: aux file claims 0-dim R-tree, which is not allowed");
     W_StaticRTree::close();
     return 0;
  }

  initNumberDimensions(ndims);
  if (buflen < max_leaf_size) {
     buflen = max_leaf_size;
     delete [] buf;
     t2_Alloc (buf, max_leaf_size, char);
  }

  return 1;
}


int
T2_StaticRTree::attach(int fd, int aux_fd)
{
  if (fd <= 0 || aux_fd < 0)
     return 0;

  if (W_StaticRTree::attach(fd) == 0) {		// attach fails
     t2_err_ret(__FILE__, __LINE__, "WARNING: attach fails");
     return 0;
  }

  // look for max leaf size
  char linebuf[512];
  size_t ss;
  lseek(aux_fd, 0, SEEK_SET);
  if ((ss = readNextUncommentString(aux_fd, linebuf, 511, '#')) == 0) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: aux index file contains no data");
     return 0;
  }
  linebuf[ss] = '\0';

  int dim = atoi(linebuf);
  if (dim <= 0) {
     t2_err_ret(__FILE__, __LINE__,
	     "ERROR: invalid number of dimensions %d read from aux index file",
	     dim);
     return 0;
  }
  initNumberDimensions(dim);

  if ((ss = readNextUncommentString(aux_fd, linebuf, 511, '#')) == 0) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: aux index file does not contain max leaf size");
     return 0;
  }
  linebuf[ss] = '\0';
  int max_leaf_size = atoi(linebuf);
  if (max_leaf_size <= 0) {
     t2_err_ret(__FILE__, __LINE__,
	     "ERROR: aux index file says max leaf size = %d", max_leaf_size);
     return 0;
  }
  if (buflen < (u_int) max_leaf_size) {
     buflen = max_leaf_size;
     delete [] buf;
     t2_Alloc (buf, max_leaf_size, char);
  }

  return 1;
}


int 
T2_StaticRTree::initSearch(const T2_Box& qb)
{
  box2doublearray(qb, ndims, qrycoord);
  return W_StaticRTree::initSearch(qrycoord);
}


int
T2_StaticRTree::getNext(T2_Box& mbr, u_int& fid, off_t& offset, size_t& size,
			T2_UsrArg& arg)
{
  size_t leaf_realsize = buflen;
  u_int dim;
  int t = W_StaticRTree::getNext(coord, dim, buf, leaf_realsize);
  if (t <= 0) {			// error or no more data found
     return t;
  }

  t2_Assert (dim == ndims);
  if (mbr.getNumberDimensions() != ndims)
     mbr.setNumberDimensions(ndims);
  doublearray2box(ndims, coord, mbr);

  arg.releaseBuffer();

  // read fid, offset, size and arg from buf
  t2_Assert (leaf_realsize >= leaf_size);
  parse_leafnode(buf, leaf_realsize - leaf_size, fid, offset, size, arg);

  return 1;
}

