
/* $Id: rtreeindex.C,v 1.1 2000/03/08 02:54:46 chialin Exp $ */

/* rtreeindex.C */

#include	"t2_rtreeindex.h"

/////////////////////////////////////////////////////////////////////////////
//
// Here is the relationship between the various R-tree classes and their
// wrappers. In the diagram below,  X ===> Y means class X is derived from
// class Y (ie class Y is a base class for class X), and X ---> Y means
// class X makes use of class Y by either defining a data member of class Y
// or by using a local variable of class Y in X's methods.
//
//	T2_StaticRTreeIndex
//	--->	T2_StaticRTree
//	===>	W_StaticRTree
//	--->	W_RTree
//	--->	W_GistM
//	===>	gist_m
//
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
//
// constructor function for class T2_StaticRTreeIndex
//
/////////////////////////////////////////////////////////////////////////////
T2_UDFRet
T2_StaticRTreeIndexConstructor(const T2_System& system, 
			       T2_UsrArgReader& arg,
			       const T2_Array<int>& fd,
			       T2_Index*& idxp)
{
  // ignore arg
  T2_StaticRTreeIndex* rtree = new T2_StaticRTreeIndex(fd);
  if (rtree == NULL || rtree->isValid() == false) {	// constructor fails
     delete rtree;
     idxp = NULL;
     return T2_UDFRet_ERROR;
  }

  idxp = (T2_Index *) rtree;
  return T2_UDFRet_OK;
}


/////////////////////////////////////////////////////////////////////////////
//
// T2_StaticRTreeIndex methods
//
/////////////////////////////////////////////////////////////////////////////
T2_StaticRTreeIndex::T2_StaticRTreeIndex(const T2_Array<int>& fd)
		: T2_Index(), rtrees(0), cur_tree(0)
{
  const u_int num_fds = fd.getNumberElements();

  // recall that we have two fds per rtree
  if (num_fds % 2 > 0) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: T2_StaticRTreeIndex gets %d file descriptors but expects an even number of file descriptors",
	   num_fds);
     valid = false;
     return;
  }

  rtrees.clear(num_fds/2);
  for (u_int i=0, j=0; i<num_fds; i+=2,j++) {
     T2_StaticRTree& rt = rtrees[j];
     if (rt.attach(fd[i], fd[i+1]) == 0) {	// attach fails
	t2_err_ret(__FILE__, __LINE__,
		  "ERROR: attach fails in T2_StaticRTreeIndex constructor for file descriptors %d and %d",
		  fd[i], fd[i+1]);
	for (u_int k=0; k<j; k++)
	   rtrees[k].close();
	valid = false;
	return;
     }
  }

  valid = true;
}


T2_UDFRet
T2_StaticRTreeIndex::fetchInit(const T2_System& system, const T2_Box& qr)
{
  cur_tree = 0;
  const u_int ntrees = rtrees.getNumberElements();
  for (u_int i=0; i<ntrees; i++) {
     if (rtrees[i].initSearch(qr) == 0) { 	// initSearch fails
        t2_err_ret(__FILE__, __LINE__,
		"ERROR: initSearch fails in T2_StaticRTreeIndex for tree %d",
		i);
        return T2_UDFRet_ERROR;
     }
  }
  return T2_UDFRet_OK;
}


T2_UDFRet
T2_StaticRTreeIndex::fetch(const T2_System& system, const T2_Box& qr,
			  bool& eod, T2_Box& mbr, T2_ClusterAuxInfoWriter& info,
			  T2_VArray<T2_BlockRequest>& chk)
{
  while (cur_tree < rtrees.getNumberElements()) {
     T2_StaticRTree& rt = rtrees[cur_tree];
     mbr.setNumberDimensions(0);
     u_int i_fid;
     off_t i_offset;
     size_t i_size;
     T2_UsrArg i_arg;
     int r;
     switch (r =rt.getNext(mbr, i_fid, i_offset, i_size, i_arg)) {
         case 0:	// error
			t2_err_ret(__FILE__, __LINE__,
				"ERROR: getNext returns error for rtree %d",
				cur_tree);
			cur_tree = 0;
			return T2_UDFRet_ERROR;

         case -1:	// no more data to find in this rtree, so move on to
			// the next one
			cur_tree++;

			break;

         case 1:	// data found
			{
			   T2_BlockRequest& breq = chk.appendElement();
			   breq.getFileID() = i_fid;
			   breq.getOffset() = i_offset;
			   breq.getNumberBytes() = i_size;

			   size_t ss = i_arg.getDataSize();
			   t2_Assert (i_arg.getBufferSize() == ss);
			   if (ss > 0) {
			      info.allocBuffer(ss);
			      info.write(i_arg.getStartOfData(), ss);
			   }
			}
			eod = false;
			return T2_UDFRet_OK;

         default:	// unknown return value from getNext
			t2_err_quit(__FILE__, __LINE__,
				"ERROR: getNext returns unrecognized value %d",
				r);
			return T2_UDFRet_ERROR;
     }
  }
  eod = true;
  return T2_UDFRet_OK;
}

