
/* t2_accmsg.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/3/98
 * aggregation function classes
 */


#ifndef _T2_ACCMSG_H_
#define _T2_ACCMSG_H_


#include	"t2_header.h"
#include	"t2_usrarg.h"


// forward declaration
class T2_QPlan;
class T2_QEnv;
class T2_System;

////////////////////////////////////////////////////////////////////////
//
// T2_AccMsgBuffer
//
////////////////////////////////////////////////////////////////////////
//typedef T2_UsrArg		T2_AccMsgBuffer;
//typedef T2_UsrArgWriter		T2_AccMsgBufferWriter;
typedef T2_UsrArg		T2_AccMsgRecvBuffer;
typedef T2_UsrArgReader		T2_AccMsgBufferReader;


////////////////////////////////////////////////////////////////////////
//
// T2_AccMsgSendBuffer
//
// This is a class that is used to store an out-going message for the
// global combine phase. It contains a T2_UrsArg (through inheritance),
// and a (data pointer, data size) pair. The idea is that the
// customization code T2_AggregateFuncObj::fillAccMsgBuffer() that
// accesses this class indirectly through the wrapper class
// T2_AccMsgBufferWriter could either choose to allocate a buffer 
// through the T2_UsrArg interface and write into the message buffer,
// or just pass a pointer, along with the proper data size, to this
// class and avoid a memory copy.
//
// At the time when a message is gonna be sent to the appropriate 
// processor, the method getSendBufferInfo() chooses between the 
// user data pointer and the T2_UsrArg buffer. Currently, it chooses
// the user data whenever the corresponding pointer (ie datap) is
// not NULL; otherwise, it chooses the T2_UsrArg buffer.
//
////////////////////////////////////////////////////////////////////////
class T2_AccMsgSendBuffer: public T2_AccMsgRecvBuffer {
public:
  T2_AccMsgSendBuffer();
  ~T2_AccMsgSendBuffer() {};	// no need to delete user data

  // const and non-const access methods to user data
  const char* getUserDataPtr() const { return datap; };
  size_t getUserDataSize() const { return size; };
  const char*& getUserDataPtr() { return datap; };
  size_t& getUserDataSize() { return size; };

  // this is the method that chooses between user data buffer (ie datap)
  // and the T2_AccMsgRecvBuffer buffer, and return the appropriate data size
  void getSendBufferInfo(const char*& b, size_t& s);

  // release both buffer
  void releaseBuffer() { datap = NULL; size = 0;
			 T2_AccMsgRecvBuffer::releaseBuffer(); }

  // do not use copy constructor and operator =
  T2_AccMsgSendBuffer(const T2_AccMsgSendBuffer& b);
  T2_AccMsgSendBuffer& operator =(const T2_AccMsgSendBuffer& b);

protected:
  const char * datap;		// user data pointer
  size_t size;			// user data size

private:
};


inline
T2_AccMsgSendBuffer::T2_AccMsgSendBuffer()
			: T2_AccMsgRecvBuffer(), datap(NULL), size(0)
{
}

inline void
T2_AccMsgSendBuffer::getSendBufferInfo(const char*& b, size_t& s)
{
  if (datap != NULL) {
     b = datap;
     s = size;
     t2_Assert (T2_AccMsgRecvBuffer::getDataSize() == 0);
     return;
  }

  // datap == NULL
  b = T2_AccMsgRecvBuffer::getStartOfData();
  s = T2_AccMsgRecvBuffer::getDataSize();
}

inline
T2_AccMsgSendBuffer::T2_AccMsgSendBuffer(const T2_AccMsgSendBuffer& b)
				: T2_AccMsgRecvBuffer(b), datap(NULL), size(0)
{
  t2_err_quit(__FILE__, __LINE__,
             "do not use T2_AccMsgSendBuffer copy constructor");
}

inline T2_AccMsgSendBuffer&
T2_AccMsgSendBuffer::operator =(const T2_AccMsgSendBuffer& b)
{
  t2_err_quit(__FILE__, __LINE__,
             "do not use T2_AccMsgSendBuffer operator =");
  return *this;
}



////////////////////////////////////////////////////////////////////////
//
// T2_AccMsgBufferWriter
//
// A msg buffer used by the customization code during global combine
// phase. In particular, T2_AggregateFuncObj::fillAccMsgBuffer()
// creates a msg for each of the back-end processors (including the local
// processor). For each processor, the cutomization code can either 
// allocate a buffer (using allocBuffer()) and write into that buffer
// (using the various write functions inherited from T2_UsrArgWriter),
// and use that buffer as the out-going msg; or alternatively, the 
// customization code can just hand over a pointer and a data size and 
// ask T2 to use them as the pointer to the msg buffer and the msg size. 
// In the latter case, 
// note that the user buffer that the given pointer points
// to must (1) reside in the accumulator object, which will not be freed
// until after output handling phase completes, and (2) be free'ed
// (ie deleted) when the accumulator object is free'ed to avoid data
// leaks.
//
////////////////////////////////////////////////////////////////////////
class T2_AccMsgBufferWriter : public T2_UsrArgWriter {
friend class T2_GlobalCombineWork;
friend void
execute_qplan_global_combine_send(T2_QPlan& qplan, T2_QEnv& qenv,
		const T2_System& system, T2_Array<T2_AccMsgSendBuffer>& msgbuf,
		bool verbose);
public:
  T2_AccMsgBufferWriter();
  ~T2_AccMsgBufferWriter() {};

  // access methods to user data
  const char* getUserDataPtr() const;
  size_t getUserDataSize() const;

  // set user data pointer and data size and ignore data in T2_UsrArgWriter
  void setUserDataBuffer(const char *p, size_t s);

  // throw away the old data of arg and allocate new space of s bytes in arg
  size_t allocBuffer(size_t s) { return T2_UsrArgWriter::allocBuffer(s); };

  // do not use copy constructor and operator =
  T2_AccMsgBufferWriter(const T2_AccMsgBufferWriter& w);
  T2_AccMsgBufferWriter& operator =(const T2_AccMsgBufferWriter& w);

protected:
  T2_AccMsgSendBuffer* sbuf;	// pointer to a send buffer

  // let T2_GlobalCombineWork::handler() connect the writer to a 
  // T2_AccMsgSendBuffer object, and disconnect from that object
  void open(T2_AccMsgSendBuffer&);
  void close();

private:
};


inline
T2_AccMsgBufferWriter::T2_AccMsgBufferWriter(): T2_UsrArgWriter(), sbuf(NULL)
{
}

inline void
T2_AccMsgBufferWriter::open(T2_AccMsgSendBuffer& b)
{
  T2_UsrArgWriter::open(b);
  sbuf = &b;
}

inline void
T2_AccMsgBufferWriter::close()
{
  T2_UsrArgWriter::close();
  sbuf = NULL;
}

inline const char*
T2_AccMsgBufferWriter::getUserDataPtr() const
{ 
  t2_Assert (sbuf != NULL);
  return sbuf->getUserDataPtr();
}

inline size_t
T2_AccMsgBufferWriter::getUserDataSize() const
{
  t2_Assert (sbuf != NULL);
  return sbuf->getUserDataSize();
}

inline void
T2_AccMsgBufferWriter::setUserDataBuffer(const char *p, size_t s)
{
  t2_Assert (sbuf != NULL);
  sbuf->getUserDataPtr() = p;
  sbuf->getUserDataSize() = s;
}

inline
T2_AccMsgBufferWriter::T2_AccMsgBufferWriter(const T2_AccMsgBufferWriter& w)
					: T2_UsrArgWriter(w)
{
  t2_err_quit(__FILE__, __LINE__,
             "do not use T2_AccMsgBufferWriter copy constructor");
}

inline T2_AccMsgBufferWriter&
T2_AccMsgBufferWriter::operator =(const T2_AccMsgBufferWriter& w)
{
  t2_err_quit(__FILE__, __LINE__,
             "do not use T2_AccMsgBufferWriter operator =");
  return *this;
}



#endif	/* _T2_ACCMSG_H_ */
