
/* $Id: t2_beconfig.h,v 1.4 2000/03/14 23:27:46 chialin Exp $ */

/* t2_beconfig.h */

#ifndef _T2_BECONFIG_H_
#define _T2_BECONFIG_H_


#include	<iostream.h>
#include	<fstream.h>
#include	<strings.h>
#include	<string.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_txtfile.h"
#include	"t2_bits.h"


///////////////////////////////////////////////////////////////////////////
//
// T2_BEConfigLocalDisk
//
// An entry in the back-end configuration file for a local disk of
// a back-end node.
//
///////////////////////////////////////////////////////////////////////////
class T2_BEConfigLocalDisk {
public:
  T2_BEConfigLocalDisk(): logdid(5) {};
  ~T2_BEConfigLocalDisk() {};

  // reset entry
  void clear() { phydid = 0; logdid.clear(); };

  // read entry from a file
  bool read(T2_TxtFile& file);

  // do not use copy constructor and operator =
  T2_BEConfigLocalDisk(const T2_BEConfigLocalDisk& b);
  T2_BEConfigLocalDisk& operator =(const T2_BEConfigLocalDisk& b);

  // data members
  u_int phydid;			// physical disk id
  T2_VArray<u_int> logdid;	// logical disk id's partitioned from this disk

protected:
private:
};


inline
T2_BEConfigLocalDisk::
T2_BEConfigLocalDisk(const T2_BEConfigLocalDisk& b): logdid(5)
{
  t2_err_quit(__FILE__, __LINE__,
	"ERROR: do not use T2_BEConfigLocalDisk copy constructor");
}

inline T2_BEConfigLocalDisk&
T2_BEConfigLocalDisk::operator =(const T2_BEConfigLocalDisk& b)
{
  t2_err_quit(__FILE__, __LINE__,
	"ERROR: do not use T2_BEConfigLocalDisk operator =");
  return *this;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_BEConfigEntry
//
// An entry in the back-end configuration file that corresponds to
// a back-end node.
//
///////////////////////////////////////////////////////////////////////////
class T2_BEConfigEntry {
public:
  T2_BEConfigEntry();
  ~T2_BEConfigEntry() { delete [] hostname; };

  // access methods
  const char* getHostName() const { return hostname; };
  u_int getLogicalProcessorID() const { return logpid; };
  const T2_VArray<T2_BEConfigLocalDisk>& getLocalDiskEntries() const
						{ return local_disks; };
  const T2_VArray<u_int>& getRemotePhysicalDiskID() const
						{ return remote_disks; };

  // read entry from a file;
  // returns 0 if error occurs, 1 if succeed,
  // return 2 if an empty record is read (ie nothing but the end-of-record mark
  u_int read(T2_TxtFile& file);

  // reset all fields
  void clear();

  // output to ostream
  ostream& write(ostream& os) const;

  // allocate space and copy a string into another string
  static void str_alloc_cpy(char*& dest, const char* src);

protected:
  char* hostname;
  u_int logpid;				// logical processor id
  T2_VArray<T2_BEConfigLocalDisk> local_disks;	// local disk entries
  T2_VArray<u_int> remote_disks;	// remote but accessible physical disk
					// id's
private:
};


inline
T2_BEConfigEntry::T2_BEConfigEntry()
		: hostname(NULL), logpid(0), local_disks(10), remote_disks(10)
{
}

inline void
T2_BEConfigEntry::clear()
{
  delete [] hostname;
  hostname = NULL;
  logpid = 0;
  local_disks.clear();
  remote_disks.clear();
}

inline void
T2_BEConfigEntry::str_alloc_cpy(char*& dest, const char* src)
{
  delete [] dest;
  size_t len = strlen(src) + 1;
  t2_Alloc(dest, len, char);
  memcpy(dest, src, len);
}

inline ostream&
operator << (ostream& os, const T2_BEConfigEntry& e)
{
  return e.write(os);
}



///////////////////////////////////////////////////////////////////////////
//
// T2_BEConfigFile
//
// A class that reads a back-end configuration flie.
//
///////////////////////////////////////////////////////////////////////////
class T2_BEConfigFile {
public:
  T2_BEConfigFile(const char* fname= NULL);
  ~T2_BEConfigFile() { close(); };

  // open and close file
  void open(const char*);
  void close();

  operator const void*() const { return (valid == true)? this:NULL; };

  // access methods
  const T2_BEConfigEntry* operator [] (u_int i) const { return entries[i]; };
  T2_BEConfigEntry* operator [] (u_int i)  { return entries[i]; };
  u_int getNumberEntries() const { return entries.getNumberElements(); };

  // get the smallest invalid id's (ie the smallest one that is not used)
  u_int getMinInvalidLogicalProcID() const { return min_invlogpid; };
  u_int getMinInvalidPhysicalDiskID() const { return min_invphydid; };
  u_int getMinInvalidLogicalDiskID() const { return min_invlogdid; };

  // check if the entries are correct, in the sense that all id's are defined
  bool check_content_correct() const;

  // return true if s is a end-of-record marker
  static bool check_eor_mark(const char* s);

  // output to ostream
  ostream& write(ostream& os) const;

protected:
  u_int min_invlogpid,			// min invalid logical processor id
	min_invphydid,			// min invalid physical disk id
	min_invlogdid;			// min invalid logical disk id
  T2_VArray<T2_BEConfigEntry*> entries;
  bool valid;

  // update the max id's against the given entry
  void update_id(const T2_BEConfigEntry& e);

private:
};


inline
T2_BEConfigFile::T2_BEConfigFile(const char* fname)
		: min_invlogpid(0), min_invphydid(0), min_invlogdid(0),
		  entries(16), valid(false)
{
  if (fname != NULL)
     open(fname);
}

inline ostream&
operator << (ostream& os, const T2_BEConfigFile& e)
{
  return e.write(os);
}



#endif	/* ! _T2_BECONFIG_H_ */
