
/* t2_catalog.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/5/98
 * a template class for T2 catalogs
 */

#ifndef _T2_CATALOG_H_
#define _T2_CATALOG_H_


#include	<fstream.h>
#include	<string.h>
#include	"t2_header.h"
#include	"t2_tarray.h"


//////////////////////////////////////////////////////////////////////////
//
// T2_Catalog
//
// A template class for T2 catalogs.
//
// The class Entry below is the class for an entry in the catalog,
// which consists of a number of data fields. It must be associated with
// the following operator:
//	ifstream& operator >>(ifstream& ifs, Entry& e)
// T2_Catalog assumes that each entry is associated with a unique id number.
// In the current implementation, I assume the various data fields
// for an entry is stored in the ascii format. This makes it easier
// to debug. To make the access more efficient, T2_Catalog assumes that each
// catalog has an index (*.cix), which lists for each id number the
// offset of the starting byte for its corresponding entry in the catalog.
// The index is assumed to be sorted on the id numbers, and the id numbers
// are assumed to be consecutive values starting with 0.
//
//////////////////////////////////////////////////////////////////////////
template <class Entry>
class T2_Catalog {
public:
  T2_Catalog();
//  T2_Catalog(const char* fname);
  virtual ~T2_Catalog();

  // query methods
  u_int getNumberEntries() const;

  // open the catalog
  bool open(const char* fname);
  void close() { catalog.close(); indexf.close(); };
  operator const void*() const { return (nofile)? 0 : this; };

  // read an entry from the catalog
  bool read(u_int i, Entry& e);

protected:
  ifstream catalog, indexf;
  T2_Array<off_t> index;
  bool nofile;

  // return the offset for a given id value from the index
  bool found(u_int i, off_t& f);

private:
  // do not use copy constructor and operator =
  T2_Catalog(const T2_Catalog<Entry>& c);
  T2_Catalog<Entry>& operator =(const T2_Catalog<Entry>& c);
};


template <class Entry>
inline
T2_Catalog<Entry>::T2_Catalog(): catalog(), indexf(), index(0), nofile(true)
{
}

//template <class Entry>
//inline
//T2_Catalog<Entry>::T2_Catalog(const char* fname): index(0)
//{
//  open(fname);
//}

template <class Entry>
inline
T2_Catalog<Entry>::~T2_Catalog()
{
}

template <class Entry>
inline u_int
T2_Catalog<Entry>::getNumberEntries() const
{
  if (nofile)
     return 0;
  return index.getNumberElements();
}

template <class Entry>
inline bool
T2_Catalog<Entry>::open(const char* fname)
{
  if (fname == NULL) {
     nofile = true;
     return false;
  }

  const size_t bufsize = 128;
  char namebuf[bufsize];
  const char cat_suffix[] = ".cat",
             idx_suffix[] = ".cix";
  const size_t suffix_len = 4;
  const size_t flen = strlen(fname);

  nofile = true;

  if (flen > bufsize - suffix_len - 1) {
     t2_err_ret(__FILE__, __LINE__, "Filename %s exceeds buffer length(%d)",
		fname, bufsize-suffix_len-1);
     return false;
  }

  strcpy(namebuf, fname);
  strcat(namebuf, cat_suffix);		// set catalog full filename
  catalog.open(namebuf, ios::in);
  if (!catalog) {			// open failed
     t2_err_ret(__FILE__, __LINE__, "T2_Catalog cannot open catalog %s",
		namebuf);
     return false;
  }

  strcpy(namebuf+flen, idx_suffix);	// set catalog index full filename
  indexf.open(namebuf, ios::in);
  if (!catalog) {			// open failed
     t2_err_ret(__FILE__, __LINE__, "T2_Catalog cannot open index %s",
		namebuf);
     catalog.close();
     return false;
  }

  // load index into memory -- assuming the index is small
  u_int n = 0;
  indexf >> n;				// get number of id numbers
  index.clear(n);
  {
     for (u_int i=0; i<n; i++) {
        if (indexf.eof()) {
	   t2_err_quit(__FILE__, __LINE__,
		      "T2_Catalog expects %d id's but encounters eof after reading %d id's from %s",
	    	      n, i, namebuf);
	}
	indexf >> index[i];
     }
  }

  nofile = false;
  return true;				// open succeeded !
}

template <class Entry>
inline bool
T2_Catalog<Entry>::found(u_int i, off_t& f)
{
  if (i >= index.getNumberElements())
     return false;
  f = index[i];
  return true;
}

template <class Entry>
inline bool
T2_Catalog<Entry>::read(u_int i, Entry& e)
{
  off_t f = 0;
  if (!found(i, f))
     return false;

  catalog.seekg(f);
  catalog >> e;
  return true;
}

template <class Entry>
inline
T2_Catalog<Entry>::T2_Catalog(const T2_Catalog<Entry>& c)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_Catalog copy constructor");
}

template <class Entry>
inline T2_Catalog<Entry>&
T2_Catalog<Entry>::operator =(const T2_Catalog<Entry>& c)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_Catalog operator =");
  return *this;
}




#endif	/* _T2_CATALOG_H_ */
