
/* $Id: t2_clusterreq.h,v 1.7 2000/01/11 20:58:58 chialin Exp $ */

/* t2_clusterreq.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/9/98
 * a cluster of chunk requests
 */


#ifndef _T2_CLUSTERREQ_H_
#define _T2_CLUSTERREQ_H_


#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_tlist.h"
#include	"t2_bits.h"
#include	"t2_shape.h"
#include	"t2_bbuf.h"
//#include	"t2_auxdata.h"
#include	"t2_usrarg.h"
#include	"t2_bcache.h"


// forward declaration
class T2_ClusterRequest;
class T2_DatasetEnv;

/////////////////////////////////////////////////////////////////////////
//
// T2_PhyBlockRequestStatus
//
// A physical block request can be in one of the following status:
//	idle
//	activated
//	pending
//	available
//	cached
//
/////////////////////////////////////////////////////////////////////////
enum T2_PhyBlockRequestStatus {
	t2_bIdle = 0,
	t2_bActivated,
	t2_bPending,
	t2_bAvailable,
	t2_bCached
};

				
/////////////////////////////////////////////////////////////////////////
//
// T2_PhyBlockOwnerInfo
//
// A class that records for a physical block request the information
// of the clusters that own that block request.
// Typical information includes :
//   numStaticOwners:	the total number of cluster requests that
//			owns the block request (it is "static" in the
//			sense that not all owners would need the block
//			at the same time); could be obtained by simply
//			inspecting the cluster requests from all queries;
//			note that in the current implementation of
//			T2_PhyBlockRequestList::initBlockRequests(),
//			a cluster request that would need the block
//			in two different iterations would only register
//			as once -- this therefore gives the maximum
//			number of clusters that would need to consume 
//			this block at once, which is different from
//			how many time the block would be consumed;
//			(if a count of how many times the block would be
//			consumed and a count of how many times the block
//			has been consumed are kept, the cache would know
//			better whether to cache a block or not)
//   owners:		a list of the pointers to the cluster requests
//			that need the block request right now (these are
//			the "current owners");
//			the length of this list should be bounded by
//			the value of numStaticOwners;
//
//  Check out the comment in class T2_PhyBlockRequest for how a
//  T2_PhyBlockOwnerInfo is used.
//
/////////////////////////////////////////////////////////////////////////
class T2_PhyBlockOwnerInfo {
public:
  T2_PhyBlockOwnerInfo();
  ~T2_PhyBlockOwnerInfo() {};

  // access numStaticOwners
  u_int getNumberStaticOwners() const { return numStaticOwners; };
  void addStaticOwner(T2_ClusterRequest* p) { incNumberStaticOwners(); };

  // access owners
  u_int getNumberCurrentOwners() const { return owners.getNumberElements(); };
  void resetCurrentOwnerSet() { owners.clear(getNumberStaticOwners()); };
  T2_ClusterRequest* getCurrentOwner(u_int i) const { return owners[i]; };
  void addCurrentOwner(T2_ClusterRequest* r);

protected:
  u_int numStaticOwners;

  T2_TSet<T2_ClusterRequest*> owners;

  // incrementing the number of static owners
  u_int incNumberStaticOwners() { return ++numStaticOwners; };

private:
  // do not use copy constructor and operator =
  T2_PhyBlockOwnerInfo(const T2_PhyBlockOwnerInfo& i);
  T2_PhyBlockOwnerInfo& operator =(const T2_PhyBlockOwnerInfo& i);

};


inline
T2_PhyBlockOwnerInfo::T2_PhyBlockOwnerInfo(): numStaticOwners(0), owners(0)
{
}

inline void
T2_PhyBlockOwnerInfo::addCurrentOwner(T2_ClusterRequest* r)
{
  t2_Assert (r != NULL);

  const u_int n = getNumberCurrentOwners();
  for (u_int i=0; i<n; i++)		// make sure no duplicates
     t2_Assert (owners[i] != r);

  t2_Assert (!owners.isFull());
  owners += r;
}

inline
T2_PhyBlockOwnerInfo::T2_PhyBlockOwnerInfo(const T2_PhyBlockOwnerInfo& i):
			numStaticOwners(0), owners(0)
{
  t2_err_quit(__FILE__, __LINE__, "do not use copy constructor");
}

inline T2_PhyBlockOwnerInfo&
T2_PhyBlockOwnerInfo::operator =(const T2_PhyBlockOwnerInfo& i)
{
  t2_err_quit(__FILE__, __LINE__, "do not use operator =");
  return *this;
}



/////////////////////////////////////////////////////////////////////////
//
// T2_PhyBlockRequest
//
// A read request for a contiguous block.
//
// A note on how the two data members, ownerinfo and refc, are used.
// ownerinfo:		used to store the number of "static" owners
//			and the list of "current owners" that touch
//			the block request during one of the following
//			status: idle, activated and pending;
//			ownerinf.owners is cleared when block becomes
//			idle, and records all the clusters that touch the
//			block request during the status of idle, activated
//			and pending; the number of elements in the list is
//			then used to initialize refc, which is used during
//			the following status: available and cached
//   refc:		reference count, which keeps track of the number
//			of current owners that will soon process, but just
//			haven't processed the block yet;
//			refc is initialized to the number of elements in
//			ownerinfo.owners when the block request becomes
//			available (ie when becomeAvailable is called),
//			and throuhtout the status of available and cached,
//			refc is incremented by 1 whenever touch is called
//			and decremented by 1 whenever release is called;
//			when refc reaches zero, the block request would
//			be put into the block cache
//
/////////////////////////////////////////////////////////////////////////
class T2_PhyBlockRequest {
public:
  T2_PhyBlockRequest();
  T2_PhyBlockRequest(int i, off_t f, size_t s);
  T2_PhyBlockRequest(const T2_PhyBlockRequest& r);

  // const access methods
  int getFileDescriptor() const { return fd; };
  off_t getOffset() const { return offset; };
  size_t getNumberBytes() const { return nbytes; };
  const T2_BBuffer& getBuffer() const { return buf; };
  u_int getNumberStaticOwners() const;
  bool hasNoBuffer() const { return buf.hasNoBuffer(); };

  // non-const access methods
  int& getFileDescriptor() { return fd; };
  off_t& getOffset() { return offset; };
  size_t& getNumberBytes() { return nbytes; };
  T2_PhyBlockRequest* getAddress() { return this; };
  T2_BBuffer& getBuffer() { return buf; };

  // operator =
  T2_PhyBlockRequest& operator =(const T2_PhyBlockRequest& r);

//  // methods for setting the pointers in owners[] and bits of iteration
//  void setOwner(T2_ClusterRequest* p);
//  bool setIteration(T2_ClusterRequest* p, T2_Iteration i);
//					// return false if iteration i not found

  // methods for accessing/updating the status
  T2_PhyBlockRequestStatus touch(T2_ClusterRequest* p);
  T2_PhyBlockRequestStatus release();
  T2_PhyBlockRequestStatus becomePending();	// activated -> pending
  T2_PhyBlockRequestStatus becomeAvailable();	// pending -> available
  T2_PhyBlockRequestStatus kill();		// (cached || available) -> idle
  T2_PhyBlockRequestStatus getStatus() const { return status; };

  // accessing refc
  u_int getRefCount() const { return refc; };

  // accessing ownerinfo
  T2_PhyBlockOwnerInfo& getOwnerInfo() { return ownerinfo; };

  // method for setting a static owner
  void addStaticOwner(T2_ClusterRequest* p) { ownerinfo.addStaticOwner(p); };

  // methods for accessing curent owner
  void resetCurrentOwnerSet() { ownerinfo.resetCurrentOwnerSet(); };

  // set the global cache object
  static void setBlockCache(T2_BlockCache* c);

protected:
  int fd;				// data file descriptor
  off_t offset;				// offset into the data file
  size_t nbytes;			// number of bytes to read

  T2_BBuffer buf;			// block buffer

  T2_PhyBlockRequestStatus status;	// current status
  friend class T2_LIO;
  void setStatus(T2_PhyBlockRequestStatus s) { status = s; };

  T2_PhyBlockOwnerInfo ownerinfo;

  u_int refc;



//  struct IterationBitVector;
//  struct IBVCmp;
//  friend struct IterationBitVector;
//  friend struct IBVCmp;
//
//  struct IterationBitVector {
//	T2_Iteration iteration;
//	T2_BitVector bitvector;
//	IterationBitVector(const IterationBitVector& v):
//			iteration(v.iteration), bitvector(v.bitvector) {};
//	IterationBitVector(T2_Iteration i): iteration(i), bitvector(0) {};
//  };
//  struct IBVCmp {
//	bool equal(const IterationBitVector& v1, const IterationBitVector& v2)
//		const	{ return (v1.iteration==v2.iteration)? true : false; };
//	bool lessThan(const IterationBitVector& v1,const IterationBitVector& v2)
//		const	{ return (v1.iteration<v2.iteration)? true : false; };
//  };
//
//  static IBVCmp cmp;
//
//  T2_VArray<T2_ClusterRequest*> owners;	// set of owning cluster requests
//  T2_SortedList<IterationBitVector, IBVCmp> iterations;
//					// data structure to find out which
//					// owners are active in a given
//					// iteration; the iteration data member
//					// stores the iteration and the bit
//					// map shows which in owners[] are
//					// active; as a consequence, the
//					// position of the pointers in
//					// owners[] cannot change !!
//
//  // methods for accessing the pointers in owners[]
//  int findOwner(T2_ClusterRequest* p);	// return index if found; -1 otherwise

private:
  static T2_BlockCache* bcache;		// cache for blocks

  // allocating/release buffer space
  size_t allocBuffer() { return buf.alloc(nbytes); };
  void releaseBuffer() { buf.release(); };

};


inline u_int
T2_PhyBlockRequest::getNumberStaticOwners() const
{
  return ownerinfo.getNumberStaticOwners();
}

inline T2_PhyBlockRequestStatus
T2_PhyBlockRequest::release()
{
  t2_Assert (status == t2_bAvailable);
  t2_Assert (refc > 0);
  if (--refc == 0) {			// block not needed any more
     t2_Assert (bcache != NULL);
     if (bcache->insert(this) == true)	// put into cache
        return (status = t2_bCached);	// status become cached
     else {				// bcache killed it
        t2_Assert (status == t2_bIdle);	// status should become idle
        return t2_bIdle;		// status become cached
     }
  }
  return t2_bAvailable;			// still needed, remain available
}

inline T2_PhyBlockRequestStatus
T2_PhyBlockRequest::becomePending()
{
  t2_Assert (status == t2_bActivated);
  t2_Assert (refc == 0);
  return (status = t2_bPending);
}

inline T2_PhyBlockRequestStatus
T2_PhyBlockRequest::becomeAvailable()	// pending -> available
{
  t2_Assert (status == t2_bPending);
  t2_Assert (refc == 0);
  refc = ownerinfo.getNumberCurrentOwners();
  t2_Assert (refc > 0);
  return (status = t2_bAvailable);
}

inline T2_PhyBlockRequestStatus
T2_PhyBlockRequest::kill()		// (cached || available) -> idle
{
  t2_Assert (status == t2_bCached || status == t2_bAvailable);
  t2_Assert (refc == 0);
  releaseBuffer();
  ownerinfo.resetCurrentOwnerSet();
  return (status = t2_bIdle);
}

inline void
T2_PhyBlockRequest::setBlockCache(T2_BlockCache* c)
{
  t2_Assert (bcache == NULL);		// only set once
  t2_Assert (c != NULL);
  bcache = c;
}



////////////////////////////////////////////////////////////////////////
//
// T2_ClusterRequest
//
////////////////////////////////////////////////////////////////////////
class T2_ClusterRequest {
public:
  T2_ClusterRequest(u_int l=0);
  T2_ClusterRequest(const T2_ClusterRequest& r): blocks(r.blocks) {};
  ~T2_ClusterRequest() {};

  // access methods
  T2_ClusterRequest* getAddress() { return this; };
  u_int getNumberBlockRequests() const { return blocks.getNumberElements(); };
  u_int setNumberBlockRequests(u_int n) { return blocks.clear(n); };
  const T2_PhyBlockRequest* getBlockRequest(u_int i) const { return blocks[i]; };
  T2_PhyBlockRequest* getBlockRequest(u_int i) { return blocks[i]; };
  const T2_Box& getMBR() const { return mbr; };
  void setMBR(const T2_Box& b);
  T2_DatasetEnv* getDatasetEnv() { return datasetenv; };

  // set req[i] to the given T2_PhyBlockRequest pointer
  void setBlockRequest(u_int i, T2_PhyBlockRequest* r);
  void setDatasetEnv(T2_DatasetEnv* d) { datasetenv = d; };

  // access, initialize, and decrement reference count
  u_int getRefCount() const { return refc; };
  u_int initRefCount() { refc = getNumberBlockRequests(); return refc; };
  u_int decRefCount() { t2_Assert (refc > 0); return --refc; };

  // allocate space for metadata and access metadata
  size_t allocMetaDataBuffer(size_t s);
  const T2_UsrArg& getMetaData() const { return metadata; };
  T2_UsrArg& getMetaData() { return metadata; };
  
protected:
  u_int refc;				// # of blocks not yet available
  T2_Box mbr;				// bounding box
  T2_DatasetEnv *datasetenv;		// dataset env in qenv
  T2_Array<T2_PhyBlockRequest*> blocks;

  T2_UsrArg metadata;			// meta data to be specified by index

private:
};


inline void
T2_ClusterRequest::setMBR(const T2_Box& b)
{
  t2_Assert (mbr.getNumberDimensions() == 0);
  mbr = b;
}

inline void
T2_ClusterRequest::setBlockRequest(u_int i, T2_PhyBlockRequest* r)
{
  t2_Assert (i < blocks.getNumberElements());
  t2_Assert (r != NULL);
  blocks[i] = r;
}

inline size_t
T2_ClusterRequest::allocMetaDataBuffer(size_t s)
{
# ifdef T2_ASSERTION
  {
    const T2_UsrArg& mt = metadata;
    t2_Assert (mt.getBufferSize() == 0);
  }
# endif
  return metadata.allocBuffer(s);
}



////////////////////////////////////////////////////////////////////////
//
// T2_PhyBlockRequestList
//
////////////////////////////////////////////////////////////////////////
class T2_PhyBlockRequestList {
public:
  T2_PhyBlockRequestList(u_int l=500): reqs(l) {};
  T2_PhyBlockRequestList(const T2_PhyBlockRequestList& b);

  // operator =
  T2_PhyBlockRequestList& operator =(const T2_PhyBlockRequestList& b);

  // query methods
  u_int getNumberBlockRequests() const { return reqs.getNumberElements(); };

  // access methods
  const T2_PhyBlockRequest& operator [] (u_int i) const { return reqs[i]; };
  T2_PhyBlockRequest& operator [] (u_int i) { return reqs[i]; };

  // search for a block request; return -1 if not found
  int search(int fdesc, off_t offset, size_t nbytes);

  // insert a block request into list
//  T2_PhyBlockRequest& touch(const T2_PhyBlockRequest& r);
  T2_PhyBlockRequest& touch(int fdesc, off_t offset, size_t nbytes);

  // initialize a block request in the list
  void initBlockRequests();

protected:
  T2_VArray<T2_PhyBlockRequest> reqs;	// a list of block requests

private:
};


inline
T2_PhyBlockRequestList::T2_PhyBlockRequestList(const T2_PhyBlockRequestList& b):reqs(500)
{
  // only use it for empty list
  t2_Assert (b.getNumberBlockRequests() == 0);
}

inline T2_PhyBlockRequestList&
T2_PhyBlockRequestList::operator =(const T2_PhyBlockRequestList& b)
{
  // only use it for empty list
  t2_Assert (b.getNumberBlockRequests() == 0);
  return *this;
}



////////////////////////////////////////////////////////////////////////
//
// T2_ClusterRequestList
//
////////////////////////////////////////////////////////////////////////
class T2_ClusterRequestList {
public:
  T2_ClusterRequestList(u_int l=100): reqs(l) {};

  // query methods
  u_int getNumberClusterRequests() const { return reqs.getNumberElements(); };

  // access methods
  const T2_ClusterRequest& operator [] (u_int i) const { return reqs[i]; };
  T2_ClusterRequest& operator [] (u_int i) { return reqs[i]; };

  // set the size of the list (this should only be called once)
//  void setNumberClusterRequests(u_int n);

  // insert an empty cluster request into list
  T2_ClusterRequest& insert();

protected:
  T2_VArray<T2_ClusterRequest> reqs;

private:
  // do not use the copy constructor nor operator =
  T2_ClusterRequestList(const T2_ClusterRequestList& c);
  T2_ClusterRequestList& operator =(const T2_ClusterRequestList& c);
};


inline
T2_ClusterRequestList::T2_ClusterRequestList(const T2_ClusterRequestList& c):
						reqs(100)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_ClusterRequestList copy constructor");
}

inline T2_ClusterRequestList&
T2_ClusterRequestList::operator =(const T2_ClusterRequestList& c)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_ClusterRequestList operator =");
  return *this;
}

inline T2_ClusterRequest&
T2_ClusterRequestList::insert()
{
  reqs << T2_ClusterRequest();
  return reqs[reqs.getNumberElements()-1];
}



#endif	/* _T2_CLUSTERREQ_H_ */
