
/* $Id: t2_qplan.h,v 1.14 2000/03/16 00:46:24 chialin Exp $ */

/* t2_qplan.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/8/98
 * query plan
 */


#ifndef _T2_QPLAN_H_
#define _T2_QPLAN_H_


#include	"t2_header.h"
#include	"t2_map.h"
#include	"t2_tarray.h"
#include	"t2_clusterreq.h"
#include	"t2_prj.h"
#include	"t2_timer.h"
#include	"t2_queryinfo.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_BlockRequestListMap
//
// A map of <T2_DSID, T2_PhyBlockRequestList> entries, which allows
// block request lists to be accessed by their corresponding dataset ids.
// That is, each dataset accessed by the current batch of queries has
// a T2_PhyBlockRequestList in the map.
//
/////////////////////////////////////////////////////////////////////////
class T2_BlockRequestListMap {
public:
  T2_BlockRequestListMap(u_int l=50): map(l) {};

  // query methods
  u_int getNumberBlockRequestLists() const { return map.getNumberElements(); };

  // return the list for dataset i; if not found, return the static emptyList
  // and set the bool flag "found" to false
  const T2_PhyBlockRequestList& getBlockRequestList(T2_DSID i, bool& found);

  // return the list for dataset i if found, otherwise insert a new empty list
  T2_PhyBlockRequestList& touchBlockRequestList(T2_DSID i);

  // initialize each block request
  void initBlockRequests();

protected:
  T2_PMap<T2_DSID, T2_PhyBlockRequestList> map;

private:
  // do not use copy constructor and operator =
  T2_BlockRequestListMap(const T2_BlockRequestListMap& m);
  T2_BlockRequestListMap& operator =(const T2_BlockRequestListMap& m);
};


inline void
T2_BlockRequestListMap::initBlockRequests()
{
  const u_int n = getNumberBlockRequestLists();
  for (u_int i=0; i<n; i++) {
     T2_PhyBlockRequestList& brlist = map.getValueByPosition(i);
     brlist.initBlockRequests();
  }
}

inline
T2_BlockRequestListMap::
T2_BlockRequestListMap(const T2_BlockRequestListMap& m): map(50)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_BlockRequestListMap copy constructor");
}

inline T2_BlockRequestListMap&
T2_BlockRequestListMap::operator =(const T2_BlockRequestListMap& m)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_BlockRequestListMap operator =");
  return *this;
}



/////////////////////////////////////////////////////////////////////////
//
// T2_QSubplan
//
// Query plan for one strip-mined iteration of a query.
//
/////////////////////////////////////////////////////////////////////////
class T2_QSubplan: public T2_QueryIterTimer {
public:
  T2_QSubplan(u_int l=100);
  T2_QSubplan(const T2_QSubplan&);
  ~T2_QSubplan() {};

  // operator =
  T2_QSubplan& operator =(const T2_QSubplan&);

  // query methods
  u_int getNumberClusterRequests() const {return clusters.getNumberElements();};

  // access methods
  const T2_ClusterRequest* getClusterRequestPtr(u_int i) const;
  T2_ClusterRequest* getClusterRequestPtr(u_int i) { return clusters[i]; };

  // set all the cluster requests
  void insertClusterRequest(T2_ClusterRequest* p);

  // sort the clusters
  void sortClusterRequests();

protected:
  T2_VArray<T2_ClusterRequest*> clusters;

private:
};


inline void
T2_QSubplan::insertClusterRequest(T2_ClusterRequest* p)
{
  t2_Assert (p != NULL);
  clusters << p;
}

inline const T2_ClusterRequest*
T2_QSubplan::getClusterRequestPtr(u_int i) const
{
  return clusters[i];
}



/////////////////////////////////////////////////////////////////////////
//
// T2_QQPlan
//
// An object that stores per-query information for a query in a T2_QPlan
// object defined below. In particular, a T2_QQPlan object stores for
// a given query, for each of its strip-mined iteration, a list of
// pointers to the cluster requests generated by the query. The list
// is represented as a T2_QSubplan object.
//
// The constructor takes an u_int, which specifies the segment size
// of the varray for the subplans.
//
/////////////////////////////////////////////////////////////////////////
class T2_QQPlan {
public:
  T2_QQPlan(u_int l=20);
  T2_QQPlan(const T2_QQPlan& qqplan);
  ~T2_QQPlan() {};

  // operator =
  T2_QQPlan& operator =(const T2_QQPlan& qqplan);

  // get address
  T2_QQPlan* getAddress() { return this; };

  // access methods to sub-plans
  T2_Iteration getNumberSubplans() const {return subplans.getNumberElements();};
  const T2_QSubplan& getSubplan(u_int i) const { return subplans[i]; };
  T2_QSubplan& getSubplan(u_int i) { return subplans[i]; };
  T2_QSubplan& getCurrentSubplan();

  // insert cluster requests into subplans according to qwindows
  void insertClusterRequests(T2_ClusterRequestList& clusters,
			     T2_VArray<T2_Region>& qwindows,
			     T2_ProjectFuncObj& prj, const T2_System& system);

  // access to cur_iter
  T2_Iteration getCurrentIteration() const { return cur_iter; };
//  void resetCurrentIteration() { cur_iter = 0; };
//  T2_Iteration incCurrentIteration();

  // access to cur_cluster
  T2_ClusterRequest* getNextClusterRequest();
  u_int getCurrentCluster() const { return cur_cluster; };
//  void resetCurrentCluster() { cur_cluster = 0; };

  // increment cur_iter and reset cur_cluster to zero
  T2_Iteration advanceCurrentIteration();

  // are all the cluster requests issued?
  bool allIterationsDone() const;
  // force all iterations done right now
  void forceIterationsDone();

  // sort the clusters
  void sortClusterRequests();

protected:
  T2_VArray<T2_QSubplan> subplans;		// a list of subplans
  T2_Iteration cur_iter;			// current iteration
  u_int cur_cluster;				// the 1st unissued cluster

private:
};


inline
T2_QQPlan::T2_QQPlan(u_int l): subplans(l), cur_iter(0), cur_cluster(0)
{
}

inline
T2_QQPlan::T2_QQPlan(const T2_QQPlan& q):
		subplans(q.subplans.segmentSize()), cur_iter(0), cur_cluster(0)
{
  t2_Assert (q.getNumberSubplans() == 0);
  t2_Assert (q.cur_iter == 0);
  t2_Assert (q.cur_cluster == 0);
}

inline T2_QQPlan&
T2_QQPlan::operator =(const T2_QQPlan& q)
{
  t2_Assert (q.getNumberSubplans() == 0);
  t2_Assert (q.cur_iter == 0);
  t2_Assert (q.cur_cluster == 0);
  t2_Assert (getNumberSubplans() == 0);
  t2_Assert (cur_iter == 0);
  t2_Assert (cur_cluster == 0);

  return *this;
}

inline T2_QSubplan&
T2_QQPlan::getCurrentSubplan()
{
  t2_Assert (cur_iter < getNumberSubplans());
  return subplans[cur_iter];
}

inline T2_Iteration
T2_QQPlan::advanceCurrentIteration()
{
  t2_Assert (cur_iter < getNumberSubplans());
  t2_Assert (cur_cluster == getSubplan(cur_iter).getNumberClusterRequests());
  cur_cluster = 0;
  return ++cur_iter;
}

inline bool
T2_QQPlan::allIterationsDone() const
{
  return (cur_iter == getNumberSubplans())? true : false;
}

inline void
T2_QQPlan::forceIterationsDone()
{
  cur_iter = getNumberSubplans();
  cur_cluster = 0;
}

inline T2_ClusterRequest*
T2_QQPlan::getNextClusterRequest()
{
  t2_Assert (cur_iter <= getNumberSubplans());
  if (allIterationsDone() == true)
     return NULL;

  T2_QSubplan& qsplan = subplans[cur_iter];
  t2_Assert (cur_cluster <= qsplan.getNumberClusterRequests());
  if (cur_cluster == qsplan.getNumberClusterRequests())
     return NULL;

  return qsplan.getClusterRequestPtr(cur_cluster++);
}

//inline T2_Iteration
//T2_QQPlan::incCurrentIteration()
//{
//  t2_Assert(cur_iter < getNumberSubplans());
//  t2_Assert(cur_cluster == getSubplan(cur_iter).getNumberClusterRequests());
//  resetCurrentCluster();
//  return ++cur_iter;
//}

inline void
T2_QQPlan::sortClusterRequests()
{
  for (u_int i=0; i<subplans.getNumberElements(); i++)
     subplans[i].sortClusterRequests();
}



/////////////////////////////////////////////////////////////////////////
//
// T2_QPlan
//
// A query plan object. It keeps for each query a T2_QQPlan object.
// It also keeps for each active dataset (ie a dataset that is required
// by some query in the given query batch) a list of the block requests.
//
// The constructor takes an u_int argument that specifies the number
// of T2_QQPlan objects in this T2_QPlan. This number should be the
// same as the number of queries in the given query batch.
//
/////////////////////////////////////////////////////////////////////////
class T2_QPlan {
public:
  T2_QPlan(u_int l=0);
  T2_QPlan(const T2_QPlan& qp);
  ~T2_QPlan();

  // operator =
  T2_QPlan& operator = (const T2_QPlan& qp);

  // access methods to qqplan
  u_int getNumberQQPlans() const { return qqplan.getNumberElements(); };
  u_int setNumberQQPlans(u_int n);
  const T2_QQPlan& getQQPlan(u_int i) const { return qqplan[i]; };
  T2_QQPlan& getQQPlan(u_int i) { return qqplan[i]; };

  // access methods to blockreq
  u_int getNumberBlockRequestLists() const;
  const T2_PhyBlockRequestList& getBlockRequestList(T2_DSID i, bool& found);
  const T2_BlockRequestListMap& getBlockRequestListMap() const;
  T2_BlockRequestListMap& getBlockRequestListMap();
  
  // initialize each block request before executing the query plan
  void initBlockRequests() { blockreq.initBlockRequests(); };
  void initBlockRequestStaticOwners();

  // insert cluster requests into subplans according to qwindows
//  void insertClusterRequests(T2_ClusterRequestList& clusters,
//			     const T2_VArray<T2_Region>& qwindows,
//			     T2_ProjectFuncObj& prj);

  // select a cluster request from all T2_QQPlans to issue
  T2_ClusterRequest* getOneClusterRequest();

  // initialize qplan at the beginning of query execution
  void initBeforeExecution();

  // are all qqplans done issueing their requests?
//  bool allQQPlanDoneIssueingReads() const;
//  u_int getNumberQQPlanNotCompleted() const { return nqqplanNotCompleted; };
//  u_int decNumberQQPlanNotCompleted();

  // sort the clusters
//  void sortClusterRequests();

protected:
  T2_BlockRequestListMap blockreq;		// all the block requests
  T2_Array<T2_QQPlan> qqplan;			// a list of T2_QQPlan

  u_int cur_qqplan;		// which qqplan is allowed to issue read

//  u_int nqqplanNotCompleted;	// # of qqplans not yet issued all its reads

private:
};


inline u_int
T2_QPlan::setNumberQQPlans(u_int n)
{
  t2_Assert (getNumberQQPlans() == 0);
  return qqplan.clear(n);
}

inline u_int
T2_QPlan::getNumberBlockRequestLists() const
{
  return blockreq.getNumberBlockRequestLists();
}

inline const T2_BlockRequestListMap&
T2_QPlan::getBlockRequestListMap() const
{
  return blockreq;
}

inline T2_BlockRequestListMap&
T2_QPlan::getBlockRequestListMap()
{
  return blockreq;
}

inline
T2_QPlan::T2_QPlan(u_int l):
		blockreq(), qqplan(l), cur_qqplan(0)
//		nqqplanNotCompleted(0)
{
}

inline
T2_QPlan::T2_QPlan(const T2_QPlan& qp):
		blockreq(), qqplan(qp.qqplan), cur_qqplan(0)
//		nqqplanNotCompleted(0)
{
  t2_Assert (qp.getNumberBlockRequestLists() == 0);
  t2_Assert (qp.getNumberQQPlans() == 0);
  t2_Assert (qp.cur_qqplan == 0);
}

inline
T2_QPlan::~T2_QPlan()
{
}

inline T2_QPlan&
T2_QPlan::operator = (const T2_QPlan& qp)
{
  t2_Assert (qp.getNumberBlockRequestLists() == 0);
  t2_Assert (qp.getNumberQQPlans() == 0);
  t2_Assert (qp.cur_qqplan == 0);
//  t2_Assert (qp.nqqplanNotCompleted == 0);

  t2_Assert (getNumberBlockRequestLists() == 0);
  t2_Assert (getNumberQQPlans() == 0);
  t2_Assert (cur_qqplan == 0);
//  t2_Assert (nqqplanNotCompleted == 0);

  return *this;
}


//inline bool
//T2_QPlan::allQQPlanDoneIssueingReads() const
//{
//  return (nqqplanNotCompleted == 0)? true : false;
//}
//
//inline u_int
//T2_QPlan::decNumberQQPlanNotCompleted()
//{
//  t2_Assert (nqqplanNotCompleted > 0);
//  return --nqqplanNotCompleted;
//}

//inline void
//T2_QPlan::sortClusterRequests()
//{
//  for (u_int i=0; i<qqplan.getNumberElements(); i++)
//    qqplan[i].sortClusterRequests();
//}



#endif	/* _T2_QPLAN_H_ */
