
/* t2_qstatus.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/12/98
 * a data structure to keep track of qenv status
 */


#ifndef _T2_QSTATUS_H_
#define _T2_QSTATUS_H_


#include	"t2_header.h"
#include	"t2_qplan.h"


//////////////////////////////////////////////////////////////////////////
//
// T2_QState
//
//////////////////////////////////////////////////////////////////////////
enum T2_QState { 			// the current state of a query
	t2_qIdle = 0, 			// doing nothing
	t2_qLocal, 			// in the local reduction phase
	t2_qGlobal, 			// in the global-combine phase
	t2_qPostGlobal, 		// done processing all global-combine
					// msg, but the T2_OutputHandleWork
					// work entry is still in the queue
					// and has not been processed, so
					// the query has not officially
					// entered the output-handling phase
	t2_qOutput,			// in the output-handling phase
	t2_qEndIteration,		// end of an iteration
	t2_qDone			// query completed done
};


const char* toString(T2_QState s);


//////////////////////////////////////////////////////////////////////////
//
// T2_QGlobalStatus
//
//////////////////////////////////////////////////////////////////////////
class T2_QGlobalStatus {
public:
  T2_QGlobalStatus(u_int i=0);
  ~T2_QGlobalStatus() {};

  // accessing # of queries still not done
  u_int getNumberQueriesNotDone() const { return nqry_notdone; };
  u_int setNumberQueriesNotDone(u_int i);
  u_int decNumberQueriesNotDone();

protected:
  u_int nqry_notdone;			// # of queries still not done

private:
};


inline 
T2_QGlobalStatus::T2_QGlobalStatus(u_int i) : nqry_notdone(i)
{
}

inline u_int
T2_QGlobalStatus::setNumberQueriesNotDone(u_int i)
{
  t2_Assert (nqry_notdone == 0);
  return (nqry_notdone = i);
}

inline u_int
T2_QGlobalStatus::decNumberQueriesNotDone()
{
  t2_Assert (nqry_notdone > 0);
  return --nqry_notdone;
}



//////////////////////////////////////////////////////////////////////////
//
// T2_QStatus
//
// Check out back-end/doc/qstatus.fig for how the counters are used.
//
//////////////////////////////////////////////////////////////////////////
class T2_QStatus {
public:
  T2_QStatus();
  ~T2_QStatus() {};

  // accessing query state
  T2_QState getState() const { return qstate; };
  T2_QState setState(T2_QState s);

  // accessing # clusters yet to process in the current iteration
  u_int getNumberClustersYetProcess() const { return nclusters_yetprocess; };
  u_int setNumberClustersYetProcess(u_int i);
  u_int decNumberClustersYetProcess();
  bool decNumberClustersYetToProcessAndCheckEndLocalReduction(
						const T2_QQPlan& qplan);

  // accessing number of acc msgs yet to process in the current iteration
  u_int getNumberAccMsgsYetProcess() const { return naccmsgs_yetprocess; };
  u_int setNumberAccMsgsYetProcess(u_int i);
  u_int decNumberAccMsgsYetProcess();

  // accessing number of acc msgs yet to recv in the current iteration
  u_int getNumberAccMsgsYetRecv() const { return naccmsgs_yetrecv; };
  u_int setNumberAccMsgsYetRecv(u_int i);
  u_int decNumberAccMsgsYetRecv();

  // accessing number of acc msg sends yet to complete
  u_int getNumberAccSendsYetComplete() const { return naccsends_yetcomplete; };
  u_int setNumberAccSendsYetComplete(u_int i);
  u_int decNumberAccSendsYetComplete();
  T2_QState decNumberAccSendsYetCompleteAndCheckEndOfStatus(bool blocking);

  // accessing number of output writes yet to complete
  u_int getNumberOutputWritesYetComplete() const;
  u_int setNumberOutputWritesYetComplete(u_int i);
  u_int decNumberOutputWritesYetComplete();

  // initializing the counters for various phases
  void initLocalReduction(const T2_QQPlan& qqplan);
  void initGlobalCombine(u_int naccmsgs_toprocess, u_int naccmsgs_torecv,
			 u_int nsends);
  void initOutputHandle(u_int nwrites);
  void initQueryDone();

  // checking for the end of various phases
  bool endLocalReduction(const T2_QQPlan& qplan) const;
  bool endGlobalCombine(bool blocking) const ;
  bool endOutputHandle() const;
  bool endQuery(const T2_QQPlan& qplan) const;

  // get global status
  static T2_QGlobalStatus& getGlobalStatus() { return global; };

protected:
  T2_QState qstate;			// query state

  // counters used during local-reduction phase
  u_int nclusters_yetprocess;		// # clusters yet to process

  // counters used during global-combine phase:
  // naccmsgs_yetprocess keeps track of the number of accmsgs yet to
  //	process; it is initialized by initGlobalCombine() to the number of
  //	processors and decremented by T2_AccMsgWork::handler() every time
  //	an acc msg for this query is processed;
  // naccmsgs_yetrecv keeps track of the number of acc msgs yet to
  //	be received by the communication manager; when an acc msg is 
  //	received and the query is in the global combine phase (ie msg is
  //	not put into recv_buf), the comm manager decrements its counter on
  //	how many more msgs the comm manager expects to recv, decrements
  //	naccmsgs_yetrecv for this query, then puts the acc msg in the 
  //	work queue; eventually this acc msg gets to the head of the queue 
  //	and is therefore processed by the query, at which time the counter
  //	nclusters_yetprocess gets decremented; therefore we always have
  //	nclusters_yetprocess >= naccmsgs_yetrecv;
  //	the reason for having naccmsgs_yetrecv is that when a query gets
  //	killed during the global combine phase, it asks comm manager to
  //	decrement the number of msgs that the comm manager still expects to
  //	receive, and the number of msgs to decrement should be
  //	naccmsgs_yetrecv, not naccmsgs_yetprocess;
  // naccsends_yetcomplete keeps track of the number of pending sends
  //	posted for an acc msg for a query during the global combine phase;
  //	a query can proceed to the output-handling phase without waiting for
  //	all acc msg sends to complete, but it cannot go beyond the end of
  //	current strip-mined iteration (ie tile) into the next iteration
  //	until naccsends_yetcomplete reaches 0; it is decremented by
  //	T2_CommManager::actionISendCompletes()
  //
  u_int naccmsgs_yetprocess;		// # acc msgs yet to process
  u_int naccmsgs_yetrecv;		// # acc msgs yet received
  u_int naccsends_yetcomplete;		// # acc msgs sends yet to complete

  // counters used during output-handling phase
  u_int noutputwrites_yetcomplete;	// # output writes yet to complete

  static T2_QGlobalStatus global;

private:
};


inline
T2_QStatus::T2_QStatus(): qstate(t2_qIdle), nclusters_yetprocess(0),
			naccmsgs_yetprocess(0), naccmsgs_yetrecv(0),
			naccsends_yetcomplete(0), noutputwrites_yetcomplete(0)
{
}

inline T2_QState
T2_QStatus::setState(T2_QState s)
{
  return (qstate = s);
}

inline u_int
T2_QStatus::setNumberClustersYetProcess(u_int i)
{
  return (nclusters_yetprocess = i);
}

inline u_int
T2_QStatus::decNumberClustersYetProcess()
{
  t2_Assert (nclusters_yetprocess > 0);
  return --nclusters_yetprocess;
}

inline u_int
T2_QStatus::setNumberAccMsgsYetProcess(u_int i)
{
  return (naccmsgs_yetprocess = i);
}

inline u_int
T2_QStatus::decNumberAccMsgsYetRecv()
{
  t2_Assert (naccmsgs_yetrecv > 0);
  return --naccmsgs_yetrecv;
}

inline u_int
T2_QStatus::setNumberAccMsgsYetRecv(u_int i)
{
  return (naccmsgs_yetrecv = i);
}

inline u_int
T2_QStatus::decNumberAccMsgsYetProcess()
{
  t2_Assert (naccmsgs_yetprocess > 0);
  return --naccmsgs_yetprocess;
}

inline u_int
T2_QStatus::setNumberAccSendsYetComplete(u_int i)
{
  return (naccsends_yetcomplete = i);
}

inline u_int
T2_QStatus::decNumberAccSendsYetComplete()
{
  t2_Assert (naccsends_yetcomplete > 0);
  return --naccsends_yetcomplete;
}

inline u_int
T2_QStatus::getNumberOutputWritesYetComplete() const
{
  return noutputwrites_yetcomplete;
}

inline u_int
T2_QStatus::setNumberOutputWritesYetComplete(u_int i)
{
  t2_Assert (noutputwrites_yetcomplete == 0);
  return (noutputwrites_yetcomplete = i);
}

inline u_int
T2_QStatus::decNumberOutputWritesYetComplete()
{
  t2_Assert (noutputwrites_yetcomplete > 0);
  return --noutputwrites_yetcomplete;
}

inline void
T2_QStatus::initLocalReduction(const T2_QQPlan& qqplan)
{
  t2_Assert (qstate == t2_qIdle);
  t2_Assert (nclusters_yetprocess == 0);
  t2_Assert (qqplan.getCurrentCluster() == 0);

  qstate = t2_qLocal;

  const T2_Iteration i = qqplan.getCurrentIteration();
  const T2_QSubplan& qsubplan = qqplan.getSubplan(i);
  nclusters_yetprocess = qsubplan.getNumberClusterRequests();
}

inline void
T2_QStatus::initGlobalCombine(u_int naccmsgs_toprocess, u_int naccmsgs_torecv,
			      u_int nsends)
{
  t2_Assert (qstate == t2_qLocal); 
  t2_Assert (naccmsgs_yetprocess == 0); 
  t2_Assert (naccmsgs_yetrecv == 0);
  t2_Assert (naccsends_yetcomplete == 0);
  t2_Assert (nclusters_yetprocess == 0);

  qstate = t2_qGlobal;
  naccmsgs_yetprocess = naccmsgs_toprocess;
  naccmsgs_yetrecv = naccmsgs_torecv;
  naccsends_yetcomplete = nsends;
}

inline void
T2_QStatus::initOutputHandle(u_int nwrites)
{
  t2_Assert (qstate == t2_qPostGlobal || qstate == t2_qLocal);
  t2_Assert (noutputwrites_yetcomplete == 0);

  qstate = t2_qOutput;
  noutputwrites_yetcomplete = nwrites;
}

inline void
T2_QStatus::initQueryDone()
{
  t2_Assert (qstate == t2_qEndIteration);
  qstate = t2_qDone;
}

inline bool
T2_QStatus::endLocalReduction(const T2_QQPlan& qqplan) const
{
  t2_Assert (qstate == t2_qLocal);
# ifdef T2_ASSERTION
     const T2_Iteration i = qqplan.getCurrentIteration();
     const T2_QSubplan& qsubplan = qqplan.getSubplan(i);
# endif
  t2_Assert (i < qqplan.getNumberSubplans());
  t2_Assert (qqplan.getCurrentCluster() <= qsubplan.getNumberClusterRequests());

  if (nclusters_yetprocess == 0) {
     t2_Assert(qqplan.getCurrentCluster() ==
	       qsubplan.getNumberClusterRequests());
     return true;
  }

  return false;
}

inline bool
T2_QStatus::endGlobalCombine(bool blocking) const
{
  t2_Assert (qstate == t2_qGlobal);

  if (blocking == true) {		// output-handling is blocking
     if (naccmsgs_yetprocess == 0 && naccsends_yetcomplete == 0)
	return true;
     return false;
  }
  else {				// output-handling is non-blocking
     if (naccmsgs_yetprocess == 0)
	return true;
     return false;
  }
}

inline bool
T2_QStatus::endOutputHandle() const
{
  if (qstate == t2_qOutput &&
      noutputwrites_yetcomplete == 0 && naccsends_yetcomplete == 0)
     return true;

  return false;
}

inline bool
T2_QStatus::endQuery(const T2_QQPlan& qplan) const
{
  t2_Assert (qstate == t2_qEndIteration);
  t2_Assert (qplan.getCurrentIteration() <= qplan.getNumberSubplans());

  if (qplan.getCurrentIteration() == qplan.getNumberSubplans())
     return true;

  return false;
}

inline T2_QState
T2_QStatus::decNumberAccSendsYetCompleteAndCheckEndOfStatus(bool blocking)
{
  if (decNumberAccSendsYetComplete() == 0) {	// all acc isends complete
     if (blocking == true) {			// output-handling is blocking
	t2_Assert (getState() == t2_qPostGlobal);
	if (endGlobalCombine(blocking) == true)	// global-combine phase done
	   return t2_qGlobal;
	return t2_qIdle;			// global-combine phase not done
     }
     else {
	if (endOutputHandle() == true) 		// output-handling phase done
	   return t2_qOutput;
	return t2_qIdle;			//output-handling phase not done
     }
  }
  else
     return t2_qIdle;				// no phase is done
}

inline bool
T2_QStatus::decNumberClustersYetToProcessAndCheckEndLocalReduction(
					const T2_QQPlan& qqplan)
{
  decNumberClustersYetProcess();
  return endLocalReduction(qqplan);
}



#endif	/* ! _T2_QSTATUS_H_ */
