
/* $Id: t2_rts.h,v 1.8 2000/03/16 00:46:24 chialin Exp $ */

/* t2_rts.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 2/9/98
 * T2 run-time system
  */


#ifndef _T2_RTS_H_
#define _T2_RTS_H_


#include	<fcntl.h>
#include	<unistd.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_system.h"
#include	"t2_acc.h"
#include	"t2_aggr.h"
#include	"t2_prj.h"
#include	"t2_dataset.h"
#include	"t2_index.h"
#include	"t2_dscat.h"
#include	"t2_idxcat.h"
#include	"t2_dsmap.h"


#ifdef T2_METACHAOS
#include	"t2mc_usercons.H"
#endif


////////////////////////////////////////////////////////////
//
// T2_RTSDSEntry
//
// An entry in the T2_RTSDSList defined below.
// A T2_RTSDSEntry contains a dsmap to mark the data files
// owned by the local processor and their corresponding file
// descriptors, and a pointer to the dataset object.
//
////////////////////////////////////////////////////////////
class T2_RTSDSEntry {
public:
  T2_RTSDSEntry(u_int l=100);
  T2_RTSDSEntry(const T2_RTSDSEntry&);
  ~T2_RTSDSEntry();

  // operator =
  T2_RTSDSEntry& operator =(const T2_RTSDSEntry&);

  // const access methods
  T2_DSID getDatasetID() const { return id; };
  const T2_Dataset* getDatasetPtr() const { return dataset; };
  const T2_Dataset& getDataset() const { return *dataset; };
  const T2_DSFileMap& getDSFileMap() const { return map; };
  const T2_BitVector& getDSFileIDBitVector() const { return my_datafile_id; };
  bool isActivated() const { return activated; };

  // non-const access methods
  T2_DSID& getDatasetID() { return id; };
  T2_Dataset*& getDatasetPtr() { return dataset; };
  T2_Dataset& getDataset() { return *dataset; };
  T2_DSFileMap& getDSFileMap() { return map; };
  T2_BitVector& getDSFileIDBitVector() { return my_datafile_id; };

  // reset all fields
  void kill();
  void closeDataFiles();

  // mark/unmark as activated
  void activate() { activated = true; };
  void deactivate() { activated = false; };

  // create dataset object
  void createDataset();

protected:
  T2_DSID id;
  T2_Dataset *dataset;
  T2_DSFileMap map;
  T2_BitVector my_datafile_id;	// bit vector to tell local data files

  bool activated;		// is the current entry activated?

private:
};


inline void
T2_RTSDSEntry::closeDataFiles()
{
  for (u_int i=0; i<map.getNumberElements(); i++)
     close(map.getFileDescriptorByPosition(i));
}

inline
T2_RTSDSEntry::T2_RTSDSEntry(u_int l): dataset(NULL), map(l), 
				       my_datafile_id(0), activated(false)
{
  id = 0;
}

inline
T2_RTSDSEntry::~T2_RTSDSEntry()
{
  delete dataset;
  closeDataFiles();
}




////////////////////////////////////////////////////////////
//
// T2_RTSDSList
//
// An associative map with dataset ids as keys and
// T2_RTSDSEntry as values.
// A RTSDSEntry is either active (ie the corresponding dataset
// is being used by one of the current queries) or inactive
// (ie the corresponding dataset was used before and is kept
// around for caching). A new RTSDSEntry is added to the
// list if the corresponding dataset is needed by one of the
// current queries but the RTSDSEntry is not present in the
// list.
//
////////////////////////////////////////////////////////////
class T2_RTSDSList {
public:
  T2_RTSDSList(u_int l=100);

  // query methods
  u_int getNumberEntries() const { return list.getNumberElements(); };

  // activate the dataset with id i
  T2_Dataset* activateDataset(T2_DSID i, T2_DSCatalog& dscatalog,
			      const T2_System& system);

  // deactivate all dataset entries in the list
  void deactivateAll();

  // kill all dataset entries in the list; this method is added so that one 
  // can force to close fd's for all datasets
  void killAll();

  // search for an entry
//  bool search(T2_DSID i, T2_RTSDSEntry& e);
  T2_RTSDSEntry& search(T2_DSID i, bool& b);

protected:
  T2_TSet<T2_RTSDSEntry> list;

  static T2_RTSDSEntry emptyEntry;

  // search for an entry with a given id
  int search(T2_DSID i) const;

  // search for an inactivated entry
  int searchInactivated() const;

private:
};


inline void
T2_RTSDSList::deactivateAll()
{
  const u_int n = getNumberEntries();
  for (u_int i=0; i<n; i++)
     list[i].deactivate();
}

inline void
T2_RTSDSList::killAll()
{
  const u_int n = getNumberEntries();
  for (u_int i=0; i<n; i++)
     list[i].kill();
  list.emptySet();
}

//inline bool
//T2_RTSDSList::search(T2_DSID i, T2_RTSDSEntry& e)
inline T2_RTSDSEntry&
T2_RTSDSList::search(T2_DSID i, bool& b)
{
  int k = search(i);
  if (k < 0) {
     b = false;
     return emptyEntry;
  }
  b= true;
  return list[k];
}


////////////////////////////////////////////////////////////
//
// T2_RTS
//
////////////////////////////////////////////////////////////
class T2_RTS {
public:
  T2_RTS();
  ~T2_RTS();

  // operator =
  T2_RTS& operator = (const T2_RTS& r);

  // initializing RTS
  // there are 3 different scenarios: (see pbe.C for more details)
  // 1) no disk assignment file is given (ie diskassign_file == NULL)
  //    and no '-r' was set (ie config_byrank == false)
  //    --> follow the disk assignment in the header of the dataset catalog
  //        and assign to local processors all logical disk ids with a
  //        matching hostname
  bool initHostName(const char* datacatalog, const char* idxcatalog,
		    const char* hostname, const char* mc_pname=NULL);
  // 2) a disk assignment file is given (ie diskassign_file != NULL),
  //    in which case all previous '-r' were ignored (ie config_byrank == false)
  //    --> use the disk assignmentment file along with the header of the
  //        dataset catalog to assign logical disk ids
  bool initDiskAssignFile(const char* datacatalog, const char* idxcatalog,
			  const char* diskassign_file, const char* hostname,
			  const char* mc_pname=NULL);
  // 3) no disk assignment file is given (ie diskassign_file == NULL),
  //    and '-r' was set (ie config_byrank == false)
  //    --> simply assign logical processor ids to processors based on ranks,
  //        ie (logical_processor_id) % numnodes == mynode
  bool initRank(const char* datacatalog, const char* idxcatalog,
		const char* mc_pname=NULL);

  // initializing dataset rts entries from a query dataset spec
  T2_Dataset* activateDataset(T2_DSID id);

  // deactivate all dataset rts entries, which should be done at the end
  // of each query execution
  void deactivateAllDatasets() { dslist.deactivateAll(); };

  // initialization succeeded?
  operator const void*() const { return (initflag)? this : 0; };

  // access methods
  T2_DSCatalog& getDatasetCatalog() { return dscatalog; };
  T2_RTSDSList& getRTSDatasetInfo() { return dslist; };
  T2_IndexCatalog& getIndexCatalog() { return indexcatalog; };

  // access the various constructors
  static T2_AccMetaConstructorPtr getAccMetaConstructorPtr(u_int i);
  static T2_AggregateFuncConstructorPtr getAggrFuncConstructorPtr(u_int i);
  static T2_ProjectFuncConstructorPtr getPrjFuncConstructorPtr(u_int i);
  static T2_DatasetConstructorPtr getDatasetConstructorPtr(u_int i);
  static T2_IndexConstructorPtr getIndexConstructorPtr(u_int i);
#ifdef T2_METACHAOS
  static T2_MetaChaosUserConstructorPtr getMetaChaosUserConstructorPtr(u_int i);
#endif

  // access the processor information
  const T2_System& getSystem() const { return system; };
  T2_System& getSystem() { return system; };
  u_int getNumberNodes() const { return system.getNumberNodes(); };
  T2_ProcID getMyNode() const { return system.getMyNode(); };
  void setProcessorInfo(u_int nn, T2_ProcID me);

  // init static arrays
  static void initStaticArrays();

  // write to ostream for debugging purpose
  ostream& write(ostream& os);

protected:
  T2_DSCatalog dscatalog;	// dataset catalog
  T2_RTSDSList dslist;		// a list of dataset rts entries

  T2_IndexCatalog indexcatalog;	// index catalog

  bool initflag;		// did initialization succeed?

  T2_System system;

  static T2_Array<T2_AccMetaConstructorPtr> acc_cons;
  static T2_Array<T2_AggregateFuncConstructorPtr> aggr_cons;
  static T2_Array<T2_ProjectFuncConstructorPtr> prj_cons;
  static T2_Array<T2_DatasetConstructorPtr> ds_cons;
  static T2_Array<T2_IndexConstructorPtr> index_cons;

#ifdef T2_METACHAOS
  static T2_Array<T2_MetaChaosUserConstructorPtr> mcu_cons;
#endif

private:
};

inline T2_AccMetaConstructorPtr
T2_RTS::getAccMetaConstructorPtr(u_int i)
{
  return (i < acc_cons.getNumberElements())? acc_cons[i] : NULL;
}

inline T2_AggregateFuncConstructorPtr
T2_RTS::getAggrFuncConstructorPtr(u_int i)
{
  return (i < aggr_cons.getNumberElements())? aggr_cons[i] : NULL;
}

inline T2_ProjectFuncConstructorPtr
T2_RTS::getPrjFuncConstructorPtr(u_int i)
{
  return (i < prj_cons.getNumberElements())? prj_cons[i] : NULL;
}

inline T2_DatasetConstructorPtr
T2_RTS::getDatasetConstructorPtr(u_int i)
{
  return (i < ds_cons.getNumberElements())? ds_cons[i] : NULL;
}

inline T2_IndexConstructorPtr
T2_RTS::getIndexConstructorPtr(u_int i)
{
  return (i < index_cons.getNumberElements())? index_cons[i] : NULL;
}


#ifdef T2_METACHAOS
inline T2_MetaChaosUserConstructorPtr
T2_RTS::getMetaChaosUserConstructorPtr(u_int i)
{
  return (i < mcu_cons.getNumberElements())? mcu_cons[i] : NULL;
}
#endif

inline T2_Dataset*
T2_RTS::activateDataset(T2_DSID id)
{
  return dslist.activateDataset(id, dscatalog, system);
}

inline void
T2_RTS::setProcessorInfo(u_int nn, T2_ProcID me)
{
  system.setProcessorInfo(nn, me);
}

inline ostream&
operator << (ostream& os, T2_RTS& rts)
{
  return rts.write(os);
}



#endif	/* _T2_RTS_H_ */
