
/* t2_timer.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 9/29/98
 * Classes to measure execution times of various sorts.
 */


#ifndef _T2_TIMER_H_
#define _T2_TIMER_H_

#include	"t2_header.h"

extern int _g_mynode;


///////////////////////////////////////////////////////////////////////
//
// T2_BE_Timer
//
///////////////////////////////////////////////////////////////////////
class T2_BE_Timer {
public:
  T2_BE_Timer();
  ~T2_BE_Timer() {};

  // reset timer
  void startTimer() { start = t2_dclock(); };

  // 
protected:
  double start;

private:
};



///////////////////////////////////////////////////////////////////////
//
// T2_QBatchTimer
//
///////////////////////////////////////////////////////////////////////
class T2_QBatchTimer {
public:
  T2_QBatchTimer(): start(0.), qplan_time(0.), qexec_time(0.) {};
  ~T2_QBatchTimer() {};

  // start timer
  void startTimer() { start = t2_dclock(); };

  // set end of query planning time and query execution time
  void setEndQBatchPlanning() { qplan_time = t2_dclock(); };
  void setEndQBatchExecution() { qexec_time = t2_dclock(); };

  // access methods
  double getQBatchPlanningTime() const;
  double getQBatchExecutionTime() const;

  // output
  ostream& write(ostream& os) const;

protected:
  double start,				// beginning of query processing
         qplan_time,			// epoch at end of query planning
         qexec_time;			// epoch at end of query execution

private:
};



inline double
T2_QBatchTimer::getQBatchPlanningTime() const
{
  t2_Assert (start > 0.);
  t2_Assert (qplan_time > 0.);
  return qplan_time - start;
}

inline double
T2_QBatchTimer::getQBatchExecutionTime() const
{
  t2_Assert (qplan_time > 0.);
  t2_Assert (qexec_time > 0.);
  return qexec_time - qplan_time;
}


inline ostream&
T2_QBatchTimer::write(ostream& os) const
{
  os << "QBatch planning time = " << getQBatchPlanningTime()
     << " sec, execution time = " << getQBatchExecutionTime()
     << " sec";
  return os;
}

inline ostream&
operator << (ostream& os, const T2_QBatchTimer& t)
{
  return t.write(os);
}




///////////////////////////////////////////////////////////////////////
//
// T2_ExecuteTimer
//
// The class to record timing info for one execution instance of
// the execution service.
//
///////////////////////////////////////////////////////////////////////
class T2_ExecuteTimer {
public:
  T2_ExecuteTimer() { resetTimer(); };
  ~T2_ExecuteTimer() {};

  // reset the timer
  void resetTimer() {
	start_t = endqexecsetup_t = endqplaninit_t = 
	endqenvinit_t = endqexec_t =
	issueread_t = commpoll_t = diskpoll_t = outsockpoll_t = dowork_t = 0.;
#if (T2_TIME_EXECUTE > 1)
	liolistio_t = 0.;
#endif
  };

  // start the timer
  void startTimer() { resetTimer(); start_t = t2_dclock(); };

  // record time for the end of each phase
  void setEndQExecSetup() { endqexecsetup_t = t2_dclock(); };
  void setEndQPlanInit() { endqplaninit_t = t2_dclock(); };
  void setEndQEnvInit() { endqenvinit_t = t2_dclock(); };
  void setEndQExec() { endqexec_t = t2_dclock(); };

  // access the timers
  double getQExecSetupTime() const { return endqexecsetup_t - start_t; };
  double getQPlanInitTime() const { return endqplaninit_t - endqexecsetup_t; };
  double getQEnvInitTime() const { return endqenvinit_t - endqplaninit_t; };
  double getQExecTime() const { return endqexec_t - endqenvinit_t; };
  double getTotalTime() const { return endqexec_t - start_t; }

  double getIssueReadTime() const { return issueread_t; };
  double getPollCommTime() const { return commpoll_t; };
  double getPollDiskReadTime() const { return diskpoll_t; };
  double getPollSockWriteTime() const { return outsockpoll_t; };
  double getDoWorkTime() const { return dowork_t; };

  // add time to the timers
  void incIssueReadTime(double t) { issueread_t += t; }
  void incPollCommTime(double t) { commpoll_t += t; };
  void incPollDiskReadTime(double t) { diskpoll_t += t; };
  void incPollSockWriteTime(double t) { outsockpoll_t += t; };
  void incDoWorkTime(double t) { dowork_t += t; };

#if (T2_TIME_EXECUTE > 1)
  // add time to the timers
  void incLiolistioTime(double t) { liolistio_t += t; };

  // access the timers
  double getLiolistioTime() const { return liolistio_t; };
#endif

  // write timers
  ostream& write(ostream& os) const; 

protected:
  double start_t,			// starting time of query execution
	 endqexecsetup_t,		// end of execution setup phase
	 endqplaninit_t,		// end of qplan initialization phase
	 endqenvinit_t,			// end of qenv initialization phase
	 endqexec_t;			// end of query execution

  double issueread_t,			// time to issue disk reads
         commpoll_t,			// time to poll communcation
         diskpoll_t,			// time to poll disk reads
         outsockpoll_t,			// time to poll for socket writes
         dowork_t;			// time to do real work


#if (T2_TIME_EXECUTE > 1)
  // some more detailed timing information
  double liolistio_t;			// time to issue lio_listio
#endif

private:
};



inline ostream&
T2_ExecuteTimer::write(ostream& os) const
{
  os << _g_mynode
     << ": exec total-elapsed-time = " << getTotalTime() << " sec" << endl;
  os << _g_mynode
     << ":      exec setup-time = " << getQExecSetupTime() << " sec" << endl;
  os << _g_mynode
     << ":      exec qplan-init-time = " << getQPlanInitTime() << " sec" << endl;
  os << _g_mynode
     << ":      exec qenv-init-time = " << getQEnvInitTime() << " sec" << endl;
  os << _g_mynode
     << ":      exec loop-exec-time = " << getQExecTime() << " sec" << endl;
  os << _g_mynode
     << ":           exec issue-read-time = " << getIssueReadTime() 
		<< " sec"<<endl;
#if (T2_TIME_EXECUTE > 1)
  os << _g_mynode
     << ":                exec liolistio-time = " << getLiolistioTime()
		<< " sec" << endl;
#endif
  os << _g_mynode
     << ":           exec poll-comm-time = " << getPollCommTime() 
		<< " sec" << endl;
  os << _g_mynode
     << ":           exec poll-disk-read-time = " << getPollDiskReadTime() 
		<< " sec" << endl;
  os << _g_mynode
     << ":           exec poll-sock-write-time = " << getPollSockWriteTime() 
		<< " sec" << endl;
  os << _g_mynode
     << ":           exec do-work-time = " << getDoWorkTime() << " sec";

  return os;
}

inline ostream&
operator << (ostream& os, const T2_ExecuteTimer& t)
{
  return t.write(os);
}



///////////////////////////////////////////////////////////////////////
//
// T2_QueryIterTimer
//
// The class to record timing info for the execution of an iteration
// during query execution.
//
///////////////////////////////////////////////////////////////////////
#ifdef T2_TIME_QUERY

class T2_QueryIterTimer {
public:
  T2_QueryIterTimer() { resetTimer(); };
  ~T2_QueryIterTimer() {};

  // reset timers
  void resetTimer() {
	init_iter_start_t = init_iter_end_t =
	reduce_start_t = combine_start_t = ohandle_start_t =
	enditer_start_t = 0.;
#if (T2_TIME_QUERY > 1)
	dafacc_t = aifacc_t = fillaccmsgbuf_t = processaccmsg_t =
	finalize_t = 0.;
#endif
  }

  // set timers
  void setStartInitIteration() { init_iter_start_t = t2_dclock(); };
  void setEndInitIteration() { init_iter_end_t = t2_dclock(); };
  void setStartReduction() { reduce_start_t = t2_dclock(); };
  void setStartCombine() { combine_start_t = t2_dclock(); };
  void setStartOHandle() { ohandle_start_t = t2_dclock(); };
  void setStartEndIteration() { enditer_start_t = t2_dclock(); };
  void setEndEndIteration() { enditer_end_t = t2_dclock(); };

  // get timers
  double getInitIterationTime() const
			{ return init_iter_end_t - init_iter_start_t; };
  double getStartReduction() const { return reduce_start_t; };
  double getStartCombine() const { return combine_start_t; };
  double getStartOHandle() const { return ohandle_start_t; };
  double getReductionTime() const { return combine_start_t - reduce_start_t; };
  double getCombineTime() const { return ohandle_start_t - combine_start_t; };
  double getOHandleTime() const { return enditer_start_t - ohandle_start_t; };
  double gettEndIterationTime() const { return enditer_end_t-enditer_start_t; };

#if (T2_TIME_QUERY > 1)

  // set more detailed timers
  void incDafAccTime(double t) { dafacc_t += t; };
  void incAifAccTime(double t) { aifacc_t += t; };
  void incFillAccMsgBufferTime(double t) { fillaccmsgbuf_t += t; };
  void incProcessAccMsgTime(double t) { processaccmsg_t += t; };
  void incFinalizeTime(double t) { finalize_t += t; };

  // get more detailed timers
  double getDafAccTime() const { return dafacc_t; };
  double getAifAccTime() const { return aifacc_t; };
  double getFillAccMsgBufferTime() const { return fillaccmsgbuf_t; };
  double getProcessAccMsgTime() const { return processaccmsg_t; };
  double getFinalizeTime() const { return finalize_t; };

#endif	/* (T2_TIME_QUERY > 1) */

  // write timers
  ostream& write(ostream& os) const; 

protected:
  double init_iter_start_t,		// start time for QEnv::initIteration
	 init_iter_end_t,		// end time for QEnv::initIteration
	 reduce_start_t,		// start time for reduction phase
         combine_start_t,		// start time for combine phase
         ohandle_start_t,		// start time for output handle phase
         enditer_start_t,		// start time for end iteration
         enditer_end_t;			// end time for end iteration

#if (T2_TIME_QUERY > 1)

  // some more detailed timing information
  double dafacc_t,			// time spent in Aggr::dafAcc
         aifacc_t,			// time spent in Aggr::aifAcc
         fillaccmsgbuf_t,		// time spent in Aggr::fillAccMsgBuffer
         processaccmsg_t,		// time spent in Aggr::processAccMsg
         finalize_t;			// time spent in Aggr::finalize

#endif	/* (T2_TIME_QUERY > 1) */

private:
};


inline ostream&
T2_QueryIterTimer::write(ostream& os) const
{
  os << _g_mynode
     << ": query init-iteration-time = " << getInitIterationTime()
		<< " sec" << endl;
  os << _g_mynode
     << ": query reduction-time = " << getReductionTime() << " sec" << endl;
  os << _g_mynode
     << ": query combine-time = " << getCombineTime() << " sec" << endl;
  os << _g_mynode
     << ": query ohandle-time = " << getOHandleTime() << " sec" << endl;
  os << _g_mynode
     << ": query end-iteration-time = " << gettEndIterationTime()
		<< " sec";


#if (T2_TIME_QUERY > 1)

  // print some more detailed timers
  os << endl;
  os << _g_mynode
     << ": query daf-acc-time = " << getDafAccTime() << " sec" << endl
     << _g_mynode
     << ": query aif-acc-time = " << getAifAccTime() << " sec" << endl
     << _g_mynode
     << ": query fill-acc-msg-buffer-time = " << getFillAccMsgBufferTime()
		<< " sec" << endl
     << _g_mynode
     << ": query process-acc-msg-time = " << getProcessAccMsgTime()
		<< " sec" << endl
     << _g_mynode
     << ": query finalize-time = " << getFinalizeTime() << " sec";

#endif	/* (T2_TIME_QUERY > 1) */

  return os;
}

inline ostream&
operator << (ostream& os, const T2_QueryIterTimer& t)
{
  return t.write(os);
}


#else	/* T2_TIME_QUERY */


class T2_QueryIterTimer {
public:
  T2_QueryIterTimer() {};
  ~T2_QueryIterTimer() {};
};


#endif	/* T2_TIME_QUERY */



#endif 	/* !_T2_TIMER_H_ */
