
/* $Id: bits.C,v 1.4 2000/11/21 23:04:23 chialin Exp $ */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/23/97
 * Implementation for bit vectors. Code adapted from the part
 * of Titan code implemented by Carter T. Shock.
 */

#include	<string.h>
#include	"t2_bits.h"


/////////////////////////////////////////////////////
//
// T2_Bits
//
/////////////////////////////////////////////////////

// rather than spend all of that time shifting bits, perhaps a set of masks?
// note that the bits are indexed as 0, 1, 2, 3, ..., 31
static u_long masks [32] = {
  0x00000001UL, 0x00000002UL, 0x00000004UL, 
  0x00000008UL, 0x00000010UL, 0x00000020UL, 0x00000040UL, 
  0x00000080UL, 0x00000100UL, 0x00000200UL, 0x00000400UL, 
  0x00000800UL, 0x00001000UL, 0x00002000UL, 0x00004000UL, 
  0x00008000UL, 0x00010000UL, 0x00020000UL, 0x00040000UL, 
  0x00080000UL, 0x00100000UL, 0x00200000UL, 0x00400000UL, 
  0x00800000UL, 0x01000000UL, 0x02000000UL, 0x04000000UL, 
  0x08000000UL, 0x10000000UL, 0x20000000UL, 0x40000000UL, 
  0x80000000UL
  };

/*
 * The old version indexes the bits as 1, 2, 3, ..., 32 (ie no bit zero)
 *
static u_long masks [33] = {
  0x00000000UL, 0x00000001UL, 0x00000002UL, 0x00000004UL, 
  0x00000008UL, 0x00000010UL, 0x00000020UL, 0x00000040UL, 
  0x00000080UL, 0x00000100UL, 0x00000200UL, 0x00000400UL, 
  0x00000800UL, 0x00001000UL, 0x00002000UL, 0x00004000UL, 
  0x00008000UL, 0x00010000UL, 0x00020000UL, 0x00040000UL, 
  0x00080000UL, 0x00100000UL, 0x00200000UL, 0x00400000UL, 
  0x00800000UL, 0x01000000UL, 0x02000000UL, 0x04000000UL, 
  0x08000000UL, 0x10000000UL, 0x20000000UL, 0x40000000UL, 
  0x80000000UL
  };

 */


T2_Bits & T2_Bits::set(u_int i)
{ v |= masks[i]; return *this; }

T2_Bits & T2_Bits::clr(u_int i)
{ v &= (~(masks[i])); return *this; }

T2_Bits & T2_Bits::tog(u_int i)
{ v ^= masks[i]; return *this; }

bool T2_Bits::isSet(u_int i) const
{ return (v & masks[i]) ? true : false; }

bool T2_Bits::isClr(u_int i) const
{ return (v & masks[i]) ? false : true; }


char *
T2_Bits::sprintb(char *s) const
{
  for(int i = 0; i < 32; i++)
    s[i] = (isSet(i)) ? '1' : '0';
  s[32] = '\0';
  return s;
}

char *
T2_Bits::sprintb(char *s, const char *const *const names, char term) const
{
  *s = '\0';
  for(int i = 0; i < 32; i++) {
    if(names[i][0] == term)
      break;
    if(isSet(i))
      strcat(s,names[i]);
  }
  return s;
}


T2_OBStream&
T2_Bits::bwrite(T2_OBStream& obf) const
{
  obf << v;
  return obf;
}

T2_IBStream&
T2_Bits::bread(T2_IBStream& ibf)
{
  ibf >> v;
  return ibf;
}



/////////////////////////////////////////////////////
//
// T2_BitVector
//
/////////////////////////////////////////////////////
T2_BitVector::T2_BitVector(u_int n): nbits(n)
{
  const u_int rem = n % T2_Bits::getNumberBits();
  veclen = n / T2_Bits::getNumberBits() + ((rem)? 1:0);
  t2_Alloc(bits, veclen, T2_Bits);

  // set excess_bits_mask:
  // in the case where n cannot be divided by T2_Bits::getNumberBits()
  // (that is, when more bits than actually needed are allocated),
  // the last T2_Bits in bits[] has excess bits that are not supposed to be
  // used; in some cases, such as counting the number of bits set to one,
  // these excess bits should be ignored;
  // excess_bits_mask is a bit vector (though stored as a u_long)
  // used to set all excess bits in the last T2_Bits in bits[] to zero;
  // excess_bits_mask therefore should have all excess bits set to 0 and 
  // others (ie real bits) to 1; in the case where no excess bits were
  // allocated, we have rem = 0, and excess_bits_mask = 0
  excess_bits_mask = 0;
  for (u_int i=0; i<rem; i++)
     excess_bits_mask |= masks[i];
}

T2_BitVector::T2_BitVector(const T2_BitVector& bv)
{
  nbits = bv.nbits;
  veclen = bv.veclen;
  t2_Alloc(bits, veclen, T2_Bits);
  for (u_int i=0; i<veclen; i++)
    bits[i] = bv.bits[i];
  excess_bits_mask = bv.excess_bits_mask;
}

void
T2_BitVector::setNumberBits(u_int n)
{
  if (nbits == n)
     return;

  delete [] bits;
  nbits = n;
  const u_int rem = n % T2_Bits::getNumberBits();
  veclen = n / T2_Bits::getNumberBits() + ((rem)? 1:0);
  t2_Alloc(bits, veclen, T2_Bits);

  // see the comment in T2_BitVector::T2_BitVector(u_int) for descriptions
  // of excess_bits_mask
  excess_bits_mask = 0;
  for (u_int i=0; i<rem; i++)
     excess_bits_mask |= masks[i];
}


char *
T2_BitVector::sprintb(char *s) const
{
  char *t = s;
  for(u_int i = 0; i < veclen; i++) {
    bits[i].sprintb(t);
    t += T2_Bits::getNumberBits();
  }
  return s;
}


T2_OBStream&
T2_BitVector::bwrite(T2_OBStream& obf) const
{
  obf << veclen << nbits;
  for (u_int i=0; i<veclen; i++)
     obf << bits[i];
  return obf;
}

T2_IBStream&
T2_BitVector::bread(T2_IBStream& ibf)
{
  delete [] bits;
  ibf >> veclen >> nbits;

  t2_Alloc(bits, veclen, T2_Bits);
  for (u_int i=0; i<veclen; i++)
     ibf >> bits[i];

  // see the comment in T2_BitVector::T2_BitVector(u_int) for descriptions
  // of excess_bits_mask
  const u_int rem = nbits % T2_Bits::getNumberBits();
  excess_bits_mask = 0;
  for (u_int j=0; j<rem; j++)
     excess_bits_mask |= masks[j];

  return ibf;
}

