
/* $Id: diskassign.C,v 1.1 2000/01/25 20:54:43 chialin Exp $ */

/* diskassign.C */

#include	<strings.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_diskassign.h"

/////////////////////////////////////////////////////////////////////////////
//
// T2_DiskAssignmentFile methods
//
/////////////////////////////////////////////////////////////////////////////
void
T2_DiskAssignmentFile::open(const char* hostname, u_int rank, const char* fname)
{
  close();
  if (fname == NULL)
     return;

  T2_TxtFile file('#', 2048);
  file.open(fname, true);
  if (!file) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open disk assignment file '%s'.", fname);
     return;
  }

  // get assignment type
  const char* string = file.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot find assignment type in file '%s'.", fname);
     file.close();
     close();
     return;
  }
  if (string[0] == '\0' || string[1] == '\0') {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid assignment type '%s' at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), fname);
     file.close();
     close();
     return;
  }

  bool r = false,
       invalid_type = false;	// set to true when the type is invalid
				// (only makes sense when r is set to false)
  switch(string[0]) {
     case 'H':
     case 'h':	// hostname
		if (hostname == NULL) {
		   t2_err_ret(__FILE__, __LINE__,
			"ERROR: a hostname is needed to process the disk assignment file '%s'.",
			fname);
		}
		else {
		  switch (string[1]) {
		    case 'D':
		    case 'd':	// logical disk id
				r = readHostname2Diskid(file, hostname);
				break;
		    case 'P':
		    case 'p':	// logical proc id
				r = readHostname2Procid(file, hostname);
				break;
		    default:	// invalid char
				invalid_type = true;
				break;
		  }
		}
		break;
     case 'R':
     case 'r':	// rank
		switch (string[1]) {
		   case 'D':
		   case 'd':	// logical disk id
				r = readRank2Diskid(file, rank);
				break;
		   case 'P':
		   case 'p':	// logical proc id
				r = readRank2Procid(file, rank);
				break;
		   default:	// invalid char
				invalid_type = true;
				break;
		}
		break;
  }				// switch (string[0])
  if (r == false) {
     if (invalid_type == true)
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid assignment type '%s' at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), fname);
     close();
  }
  file.close();
}


bool
T2_DiskAssignmentFile::
read(T2_TxtFile& file, const char* hostname, const u_int rank,
     const char* id_name, T2_BitVector& bitmap)
{
  if (!file)
     return false;
  t2_Assert (file.getFileName() != NULL);
  t2_Assert (id_name != NULL);

  // Note that this method assumes that the logical ids given in the
  // file are either logical disk ids (in which case the caller should
  // have id_name points to the string "disk" and make bitmap a reference 
  // to the data member diskmap) or logical proc ids (in which case the 
  // caller should have id_name points to the string "proc" and make 
  // bitmap a reference to the data member procmap).
  // The string id_name is only used when printing out error messages.
  //
  // The disk assignment file has the following format:
  //	char[2] type			-- assignment type
  //	u_int number_logical_ids	-- that is, all valid logical ids
  //					   fall in the range between 0 and
  //					   number_logical_ids - 1
  //	<record_1>
  //	<record_2>
  //	   :
  //
  // where type can be one of the followig:
  //	hd				-- map hostnames to logical disk ids
  //	hp				-- map hostnames to logical proc ids
  //	rd				-- map MPI ranks to logical disk ids
  //	rp				-- map MPI ranks to logical proc ids
  //
  // and each record has the following format:
  //	<identifier>
  //	u_int logical_proc_id_1
  //	u_int logical_proc_id_2
  //		:
  //	-1
  //
  // where each identifier is one of the following, depending on whether 
  // hostname is NULL or nor.
  //	char[] host_name		-- when hostname is not NULL
  //					   this gives the hostname of the local
  //					   processor
  // or
  //	u_int processor_rank		-- when hostname is NULL,
  //					   this gives the MPI rank of the local
  //					   processor
  //
  // It is assumed that the caller has already read the field 'type',
  // so this method does not process it at all. This is done by method open().
  //
  // Note that each logical id (either logical processor id or logical disk
  // id) can only appear once in the disk assignment file. However,
  // the hostname or the processor rank can appear multiple times.
  // That is, the identifiers can appear multiple times, but all logical ids
  // can only appear once.

  const char* string = file.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: cannot find number of logical %s ids in file '%s'.",
	   id_name, file.getFileName());
     return false;
  }
  int atoi_err;
  int nids = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || nids < 0) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid number of logical %s ids '%s' at line %d in file '%s'.",
		id_name, string, file.getCurrentLineNumber(),
		file.getFileName());
     return false;
  }
  bitmap.setNumberBits(nids);
  bitmap.clrAll();

  // keep track of logical ids assigned to other processors to avoid error,
  // since each logical id should only appear once in the disk assignment file
  T2_BitVector others_bitmap(nids);

  // keep track of the line number each logical id first appeared so that
  // more informed error message can be generated
  T2_Array<u_int> id_lineno(nids);

  while ((string = file.getNextUncommentString()) != NULL) {
     bool my_record = false;
     u_int lineno = file.getCurrentLineNumber();

     if (hostname != NULL) {			// compare hostname
	if (strcmp(string, hostname) == 0) 	// a match
	   my_record = true;
     }
     else {					// compare rank
	int atoi_err;
	int target_rank = t2_atoi(string, &atoi_err);
	if (atoi_err != 0 || target_rank < 0) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid processor rank '%s' at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), file.getFileName());
	   return false;
	}
	if ((u_int) target_rank == rank)	// a match
	   my_record = true;
     }

     // go through the list of logical ids
     while ((string = file.getNextUncommentString()) != NULL) {
	T2_IDRange range;
	if (range.stringToRange((char*) string) == false)
	   return false;
	if (range.id_low < 0)			// end of record
	   break;
	if (range.id_high >= nids) {		// id out of range
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid logical %s id '%s' at line %d in file '%s', as a valid one must be less than %d.",
		id_name, string, file.getCurrentLineNumber(),
		file.getFileName(), nids);
	   return false;
	}

	for (int i=range.id_low; i<=range.id_high; i++) {
	   // first check if this logical id has ever appeared before
	   if (bitmap.isSet(i) == true || others_bitmap.isSet(i) == true) {
	      t2_err_ret(__FILE__, __LINE__,
		   "ERROR: logical %s id %d at line %d in file '%s' has already appeared at line %d. Each logical %s id can only appear once in the disk assignment file.",
		   id_name, i, file.getCurrentLineNumber(), file.getFileName(),
		   id_lineno[i], id_name);
	      return false;
	   }

	   if (my_record == true)		// record as my logical ids
	      bitmap.set(i);
	   else
	      others_bitmap.set(i);

	   id_lineno[i] = file.getCurrentLineNumber();	// remember lineno

	}	// for (int i=range.id_low; i<=range.id_high; i++)

     }

     if (string == NULL) {			// see eof without end of record
	t2_err_ret(__FILE__, __LINE__,
		"WARNING: encounting eof before seeing -1 to end the list of logical %s ids assigned to the hostname given at line %d in file '%s'. Assume -1 exists right before eof.",
		id_name, lineno, file.getFileName());
	break;					// eof
     }
  }		// while ((string = file.getNextUncommentString()) != NULL)

  // check if all logical ids have been assigned
  for (int i=0; i<nids; i++) {
     if (bitmap.isSet(i) == false && others_bitmap.isSet(i) == false)
	t2_err_ret(__FILE__, __LINE__,
		"WARNING: logical %s id %d did not appear in the disk assignment file '%s'.",
		id_name, i, file.getFileName());
  }

  return true;
}


ostream&
T2_DiskAssignmentFile::write(ostream& os) const
{
  const T2_BitVector* bitmap = NULL;

  switch (type) {
    case assign_error:
		os << "assignment-type = error" << endl;
		break;

    case assign_disk_ids:
		os << "assignment-type = assign_disk_ids" << endl
		   << "my logical disk ids = ";
		bitmap = &diskmap;
		break;

    case assign_proc_ids:
		os << "assignment-type = assign_proc_ids" << endl
		   << "my logical proc ids = ";
		bitmap = &procmap;
		break;

    default:
		t2_err_quit(__FILE__, __LINE__,
			"ERROR: invalid type %d", (int) type);
  }

  if (bitmap != NULL) {		// print bit map
     for (u_int i=0; i<bitmap->getNumberBits(); i++) {
	if (bitmap->isSet(i) == true)
	   os << " " << i;
     }
     os << endl;
  }

  return os;
}
