
/* $Id: qplan.C,v 1.14 2000/11/14 16:08:44 chialin Exp $ */

/* qplan.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/14/98
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<iostream.h>
#include	"t2_qplan.h"


extern int _g_mynode;

//////////////////////////////////////////////////////////////////////
//
// T2_BlockRequestListMap methods
//
//////////////////////////////////////////////////////////////////////
const T2_PhyBlockRequestList&
T2_BlockRequestListMap::getBlockRequestList(T2_DSID i, bool& found)
{
  return map.getValue(i, found);
}


T2_PhyBlockRequestList&
T2_BlockRequestListMap::touchBlockRequestList(T2_DSID i)
{
  return map.touch(i);
}



//////////////////////////////////////////////////////////////////////
//
// T2_QSublan methods
//
//////////////////////////////////////////////////////////////////////
T2_QSubplan::T2_QSubplan(u_int l): T2_QueryIterTimer(), clusters(l)
{
}


T2_QSubplan::T2_QSubplan(const T2_QSubplan& p):
			clusters(p.clusters.segmentSize())
{
  t2_Assert (p.clusters.getNumberElements() == 0);
}


T2_QSubplan&
T2_QSubplan::operator =(const T2_QSubplan& p)
{
  t2_Assert (p.clusters.getNumberElements() == 0);
  t2_Assert (getNumberClusterRequests() == 0);
  return *this;
}


int
cmp_cluster_func(const void* p1, const void* p2)
{
  t2_Assert (p1 != NULL);
  t2_Assert (p2 != NULL);

  {
     const T2_ClusterRequest **tmp;
     tmp = (const T2_ClusterRequest **) p1;
     t2_Assert (*tmp != NULL);
     tmp = (const T2_ClusterRequest **) p2;
     t2_Assert (*tmp != NULL);
  }

  const T2_ClusterRequest *cluster1 = *((const T2_ClusterRequest **) p1),
                          *cluster2 = *((const T2_ClusterRequest **) p2);

  // let's just compare the first block request of each cluster;
  // what we'll do for now is to compare the file descriptors first
  // and then the offsets

  t2_Assert (cluster1->getNumberBlockRequests() > 0);
  t2_Assert (cluster2->getNumberBlockRequests() > 0);
  const T2_PhyBlockRequest *block1 = cluster1->getBlockRequest(0),
			   *block2 = cluster2->getBlockRequest(0);
  t2_Assert (block1 != NULL);
  t2_Assert (block2 != NULL);

  if (block1->getFileDescriptor() < block2->getFileDescriptor()) {
     return -1;
  }
  if (block1->getFileDescriptor() > block2->getFileDescriptor()) {
     return 1;
  }
  if (block1->getOffset() < block2->getOffset()) {
     return -1;
  }
  if (block1->getOffset() > block2->getOffset()) {
     return 1;
  }

  // equality should not happen !!
  t2_Assert (block1->getOffset() != block2->getOffset());

  return 0;
}


void
T2_QSubplan::sortClusterRequests()
{
  // I think the easiest implementation is to copy all request pointers
  // in clusters into a temporary array, call qsort on the temporary array,
  // and then store all the pointers back into clusters. Since interleaving
  // the requests toward different disks seem to reduce the amount of time
  // it takes lio_listio to return, I'll interleave the requests when I
  // copy the pointers from tmp back to clusters. This ain't too hard since
  // the sorting method would sort the cluster requests by file descriptors
  // then by offsets. I'll just scan through tmp once to find out the starting
  // and ending position of each fd, and then use another pass to copy the
  // pointers from tmp to clusters.

  const size_t nclusters = getNumberClusterRequests();
  T2_ClusterRequest **tmp;
  t2_Alloc (tmp, nclusters, T2_ClusterRequest*);
  for (u_int i=0; i<nclusters; i++)
     tmp[i] = clusters[i];

  qsort(tmp, nclusters, sizeof(T2_ClusterRequest*), cmp_cluster_func);

  T2_VArray<size_t> start(20);		// the starting index of each fd
  T2_VArray<size_t> end(20);		// one after the ending index of each fd
  start << 0;
  for (u_int k=1; k<nclusters; k++) {
     if (tmp[k]->getBlockRequest(0)->getFileDescriptor() !=
         tmp[k-1]->getBlockRequest(0)->getFileDescriptor())
     {
	end << k;
	start << k;
     }
  }
  end << nclusters;

  // start.getNumberElements() now gives the # of different fd's in clusters;
  // in the following code, j points to the current position in clusters to
  // fill and m points to the file descriptor that we'll get from tmp next.
  for (size_t j=0, m=0; j<nclusters; j++) {
     t2_Assert (start[m] < end[m]);
     clusters[j] = tmp[start[m]];
     start[m]++;

     size_t old_m = m;
     do 		// move to next fd still with pointers in tmp[]
     {
	m++;
	t2_Assert (m <= start.getNumberElements());
	if (m == start.getNumberElements())
	   m = 0;
     } while (start[m] == end[m] && m != old_m);
  }

  delete [] tmp;
}



//////////////////////////////////////////////////////////////////////
//
// T2_QQPlan methods
//
//////////////////////////////////////////////////////////////////////
void
T2_QQPlan::insertClusterRequests(T2_ClusterRequestList& clusters,
				 T2_VArray<T2_Region>& qwindows,
				 T2_ProjectFuncObj& prj,
				 const T2_System& system)
{
  const u_int niter = qwindows.getNumberElements(),
              nsubplans = getNumberSubplans(),
              nclusters = clusters.getNumberClusterRequests();

  // make sure we have enough subplans first
  for (u_int n=nsubplans; n<niter; n++)
     subplans << T2_QSubplan();



  //************************************************************************
  // a quick fix to force the back-end node with no data to do one iteration

  if (getNumberSubplans() == 0) {		// no local data
     subplans << T2_QSubplan();
     t2_Assert (qwindows.getNumberElements() == 0);
     qwindows << T2_Region();			// insert an empty region
  }
 
  //************************************************************************

  t2_say2(cout, "assigning " << nclusters 
		<< " clusters returned by index to " << niter
		<< " strip-mined tiles" << endl);

  for (u_int i=0; i<nclusters; i++) {
     T2_ClusterRequest& cluster = clusters[i];

     t2_say4(cout, "checking cluster " << i << " with mbr "
			<< cluster.getMBR() << endl);

     // since the MBR of the cluster is given in input attribute space and
     // the query windows are given in the output attribute space,
     // projection is needed before intersection test can be made
     bool intersects = false;
     T2_Box mbr;
     if (prj.projectBox(system, cluster.getMBR(), intersects, mbr)
	    != T2_UDFRet_OK)
     {
	   t2_err_ret(__FILE__, __LINE__, "project.projectBox returns ERROR");

	   // not sure what to do now .... assume no intersection for now
	   continue;
     }


#    if (T2_VERBOSE > 3)
	u_int intersects_cnt = 0;
	if (intersects == false)
	   t2_say4(cout, "T2_ProjectFuncObj::projectBox sets argument 3 'succeed' to false" << endl);
	else
	   t2_say4(cout, "T2_ProjectFuncObj::projectBox sets argument 3 'succeed' to true, projected region as follows." << endl << mbr << endl);
#    endif

     if (intersects == false)		// no intersection
	continue;

     for (u_int j=0; j<niter; j++) {
	if (!(mbr ^ qwindows[j])) 	// no intersection
	   continue;

#	if (T2_VERBOSE > 3)
	   intersects_cnt++;
	   t2_say4(cout, "cluster " << i << " intersects with tile " << j
			<< endl);
#	endif

	// cluster needs to be retrieved for iteration j, so it must be
	// inserted into subplan j
	T2_QSubplan& sp = getSubplan(j);
	sp.insertClusterRequest(cluster.getAddress());
     }

#    if (T2_VERBOSE > 3)
	if (intersects_cnt == 0)
	   t2_say4(cout, "cluster " << i << " does not intersect with any tile"
			<< endl);
#    endif

  }


# if T2_VERBOSE > 1
  {
     u_long tot=0;
     cout << system.getMyNode() << ": # clusters per iter =";
     for (u_int i=0; i<niter; i++) {
	cout << " " << getSubplan(i).getNumberClusterRequests();
	tot += getSubplan(i).getNumberClusterRequests();
     }
     cout << "  total=" << tot << endl;
  }
# endif


}



//////////////////////////////////////////////////////////////////////
//
// T2_QPlan methods
//
//////////////////////////////////////////////////////////////////////
void
T2_QPlan::initBlockRequestStaticOwners()
{
  // Each block request carries a count of how many times the block would
  // be processed by the local processor. This count is used to initialize
  // the set of consuming cluster requests that is claiming the block
  // request during execution time.
}


T2_ClusterRequest*
T2_QPlan::getOneClusterRequest()
{
  const u_int nqqplans = getNumberQQPlans();
  const u_int start = cur_qqplan;

  do {
    t2_Assert (cur_qqplan < nqqplans);

    T2_ClusterRequest* r = NULL;
    if (!qqplan[cur_qqplan].allIterationsDone()) {
	r = qqplan[cur_qqplan].getNextClusterRequest();

//    t2_cond_say(r > 0,
//		cout, "qqplan::getOneClusterRequest " << cur_qqplan
//		      << " returns cluster " << (int) r << endl);

    }

    if (++cur_qqplan == nqqplans)
       cur_qqplan = 0;
   
    if (r != NULL)
	return r;

  } while (cur_qqplan != start);

  return NULL;				// nothing to issue now
}


void
T2_QPlan::initBeforeExecution()
{
  const u_int nqry = getNumberQQPlans();

  // initialize nqqplanNotCompleted to the number of valid queries
  for (u_int i=0; i<nqry; i++) {
    if (getQQPlan(i).getNumberSubplans() > 0) {
	t2_Assert (getQQPlan(i).getCurrentIteration() == 0);
	t2_Assert (getQQPlan(i).getCurrentCluster() == 0);
    }
  }
}

