
/* $Id: ssbe.C,v 1.11 2000/03/16 17:56:34 chialin Exp $ */

/* ssbe.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/5/98
 * Main program for the full sequential T2 back-end server with socket interface
 * talking to a front-end process.
 */


#include	<iostream.h>
#include	<stdlib.h>
#include	<new.h>
#include	"t2_header.h"
#include	"t2_bfile.h"
#include	"t2_qspec.h"
#include	"t2_rts.h"
#include	"t2_bcache.h"
#include	"t2_packets.h"


enum usageError { eUNKNOWN, eBADARG };
static void printusage(const char*);
static void badusage(usageError, const char*, char);
void query_processing(T2_Array<T2_RTS>& rts, T2_QSpec& qspec, size_t mem,
                      const char* ofname, bool verbose = false);
void be_qbatch_file(const char* qbatch_file, T2_Array<T2_RTS>& rts,
		    size_t usr_mem, const char* ofname, bool verbose);
void be_fe_socket(const char* fe_hname, int fe_port, T2_Array<T2_RTS>& rts,
             	    size_t usr_mem, const char* ofname, bool verbose);
void
process_query_batch(T2_QBatch& qbatch, T2_Array<T2_RTS>& rts,
                    size_t usr_mem, const char* ofname, bool verbose);


//////////////////////////////////////////////////////////////
//
// new handler
//
//////////////////////////////////////////////////////////////
static void t2_newHandler(void)
{
# if defined(T2_ASSERTION)
     static int wCount = 0;
     if (errno == EINTR && ++wCount < 25) {
       t2_say3(cerr, "EINTR malloc, retry ..." << endl);
       return;
     }
# else /* T2_ASSERTION */
     if (errno == EINTR) return;
# endif /* T2_ASSERTION */

  t2_say_label(cerr);
  cerr << "out of free store." << endl;
  abort();
}


//////////////////////////////////////////////////////////////
//
// some default values
//
//////////////////////////////////////////////////////////////
const size_t default_usrmem = 1048576;		// usr memory in bytes
char default_dscatfile[] = "dataset-catalog";	// dataset catalog
char default_idxcatfile[] = "index-catalog";	// index catalog
char default_qbatchfile[] = "qbatch-file";	// query batch
char *default_diskassignfile = NULL;		// disk assignment file
//char default_outputfile[] = "output-file";	// output file
char *default_outputfile = NULL;		// output file
const u_int default_numnodes = 1;		// # of processors

const char helpmsg[] = "use command-line argument -h to see usage";


//////////////////////////////////////////////////////////////
//
// main program
//
//////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  // set new handler
  set_new_handler(t2_newHandler);

  // variables
  size_t usr_mem = default_usrmem;	// memory for accumulator (in bytes)
  char *dscat_file = default_dscatfile,		// dataset catalog
       *idxcat_file = default_idxcatfile,	// index catalog
       *qbatch_file = default_qbatchfile,	// query batch
       *diskassign_file = default_diskassignfile,  // disk assignment file
       *output_file = default_outputfile;	// output file
  u_int numnodes = default_numnodes;		// # of processors
  bool  verbose = false;
  char *fe_hname = NULL;			// front-end hostname
  int  fe_port = 0;				// front-end port


  // deal with the command-line arguments
  int i;
  for (i=0; i<argc; i++) {
     if (argv[i][0] == '-') {
	switch(argv[i][1]) {
	case 'h':
		printusage(argv[0]); return 0; break;
	case 'd':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		dscat_file = argv[i];
		break;
	case 'i':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		idxcat_file = argv[i];
		break;
	case 'q':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		qbatch_file = argv[i];
		fe_hname = NULL;
		break;
	case 'a':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		diskassign_file = argv[i];
		break;
	case 'o':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		output_file = argv[i];
		break;
	case 'm':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		usr_mem = atoi(argv[i]);
		break;
	case 'p':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		numnodes = atoi(argv[i]);
		break;
	case 'v':
		verbose = true;
		break;
	case 'f':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		fe_hname = argv[i];
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-2][1]); return -1; }
		fe_port = atoi(argv[i]);
		qbatch_file = NULL;
		break;
	default:
		badusage(eUNKNOWN, argv[0], argv[i][1]); return -1; break;
	}
     }
  }

  t2_Assert (fe_hname == NULL || qbatch_file == NULL);

  // reporting arguments
  cout << "dataset catalog prefix = " << dscat_file << endl;
  cout << "index catalog prefix = " << idxcat_file << endl;
  cout << "disk assignment file = " << t2_stringEmptyWrap(diskassign_file)
       << endl;
  if (fe_hname == NULL)
     cout << "query batch input file = " << qbatch_file << endl;
  else
     cout << "front-end hostname = " << fe_hname << ", port = " << fe_port
	  << endl;
  if (output_file != NULL)
     cout << "forcing to write query output into file = "
	  << output_file << endl;
  else
     cout << "sending output by however specified by the query specification"
	  << endl;
  cout << "user memory space = " << usr_mem << " bytes" << endl;
  cout << "number of processors = " << numnodes << endl;
  cout << endl;

  // initialize T2 rts
  T2_Array<T2_RTS> rts(numnodes);
  for (u_int p=0; p<numnodes; p++) {
     rts[p].getSystem().setProcessorInfo(numnodes, p);
     rts[p].initRank(dscat_file, idxcat_file);
     if (!rts[p]) {
	t2_err_quit(__FILE__, __LINE__,
		   "T2_RTS initialization failed for backend node %d of %d\n%s",
		   p, numnodes, helpmsg);
	return -1;
     }

     rts[p].setProcessorInfo(numnodes, p);
  }
  T2_RTS::initStaticArrays();

  // initialize an empty block cache
  T2_BlockCache bcache(0);
  T2_PhyBlockRequest::setBlockCache(&bcache);


  if (fe_hname == NULL)
     be_qbatch_file(qbatch_file, rts, usr_mem, output_file, verbose);
  else
     be_fe_socket(fe_hname, fe_port, rts, usr_mem, output_file, verbose);


  // cleaning up


  return 0;
}



//////////////////////////////////////////////////////////////
//
// query processing functions
//
//////////////////////////////////////////////////////////////
void
be_qbatch_file(const char* qbatch_file, T2_Array<T2_RTS>& rts, size_t usr_mem,
	       const char* ofname, bool verbose)
{
  // reading query batch
  T2_IBFile qibf(qbatch_file);
  if (!qibf) {
     t2_err_quit(__FILE__, __LINE__,
		"Cannot open query batch file %s\n%s", qbatch_file, helpmsg);
     return;
  }

  t2_say(cout, "reading query batch from '" << qbatch_file << "'" << endl);
  T2_QBatch qbatch;
  qibf >> qbatch;
  qibf.close();
  if (qbatch.isValid() == false) {	// qbatch read failed
     t2_err_quit(__FILE__, __LINE__,
		"Failed to read query batch from %s", qbatch_file);
     return;
  }
  if (qbatch.getNumberQueries() == 0)	// got an empty query batch
     return;

  process_query_batch(qbatch, rts, usr_mem, ofname, verbose);
}


void
be_fe_socket(const char* fe_hname, int fe_port, T2_Array<T2_RTS>& rts,
             size_t usr_mem, const char* ofname, bool verbose)
{
  const int fake_numnodes = 1,
            fake_mynode = 0;
  const u_char endian = rts[0].getSystem().getBackEndEndian();

  // connecting to the front-end
  T2_FEConnection fecon;
  if (fecon.connectFrontEnd(fe_hname, fe_port, fake_numnodes, fake_mynode,
			    endian)
     == false)
  {
     t2_err_ret(__FILE__, __LINE__,
                "T2_FEConnection cannot connect to front-end %s port %d\n",
                fe_hname, fe_port);
     return;
  }

  t2_say(cout, "connected to front-end " << fe_hname << " at port "
		<< fe_port << " successfully" << endl);


  // recv packets from front-end
  int tag;
  u_int cnt = 0;
  while ((tag = fecon.recvPacketTag()) != T2_ErrorFEBEPacketTag) {
     switch (tag) {
	case T2_ConnectPacketTag:
	case T2_AcceptPacketTag:
	case T2_ReadyPacketTag:
			errno = 0;
			t2_err_ret(__FILE__, __LINE__,
				 "unexpected packet tag %d from the front-end",
				  tag);
			break;

	case T2_QueryPacketTag: {
			   T2_QBatch qbatch;
			   fecon.recvQBatch(qbatch);
			   cout << endl
				<< "back-end "
				<< " gets qbatch # " << cnt++ << " with "
				<< qbatch.getNumberQueries()
				<< " queries"
				<< endl;
			   
			   process_query_batch(qbatch, rts, usr_mem,
					       ofname, verbose);
			}
			break;

	case T2_ShutdownPacketTag:
			cout << endl
			     << "back-end "
			     << " recvs shutdown msg tag, shutting down"
			     << endl << endl;
			break;

	default:
			errno = 0;
			t2_err_ret(__FILE__, __LINE__,
				"unrecognized packet tag %d from the front-end",
				  tag);
     }

     if (tag == T2_ShutdownPacketTag)		// to shutdown
	break;

  }				// while (tag != T2_ErrorFEBEPacketTag)


  if (tag == T2_ErrorFEBEPacketTag) {
     t2_err_ret(__FILE__, __LINE__, "recvPacketTag error, shutting down");
  }
}


void
process_query_batch(T2_QBatch& qbatch, T2_Array<T2_RTS>& rts,
             size_t usr_mem, const char* oprefix, bool verbose)
{
  static int qid = 0;

  for (u_int i=0; i<qbatch.getNumberQueries(); i++, qid++) {
      // process each query in the batch one-by-one
      t2_say(cout, endl << "processing query # " << qid 
		   << " (" << i << " out of " << qbatch.getNumberQueries()
		   << " in the qbatch)" << endl);

      char fname[256];
      char *ofname = fname;
      if (oprefix == NULL)
	 ofname = NULL;
      else
         sprintf(ofname, "%s-%d", oprefix, qid);
      query_processing(rts, qbatch.getQuerySpec(i), usr_mem, ofname, verbose);

      t2_say(cout, "done processing query # " << qid  << endl);
  }
}



//////////////////////////////////////////////////////////////
//
// some help functions
//
//////////////////////////////////////////////////////////////
void printusage(const char *name)
{
  cout << "Usage: " << name
    << " [-h] | [-d <dataset catalog file>] [-i <index catalog file>] " << endl
    << " [-q <query batch file>]"
    << " [-p <number of processors>]"
    << endl
    << " [-o <output file prefix>] [-m <memory size (bytes) for accumulator>]"
    << endl
    << " [-f <front-end hostname> <port number>]"
    << " [-v]"
    << endl;

  cout << "Where:" << endl;
  cout << "-h prints this message" << endl;
  cout << "-d specifies the dataset catalog file prefix"
       << " (default = " << default_dscatfile << ")" << endl;
  cout << "-i specifies the index catalog file prefix"
       << " (default = " << default_idxcatfile << ")" << endl;
  cout << "-a specifies the disk assignment file"
       << " (default = " << t2_stringNullWrap(default_diskassignfile)
       << ")" << endl;
  cout << "-q specifies the query batch file"
       << " (default = " << default_qbatchfile << ")" << endl;
  cout << "-p specifies the number of processors involved"
       << " (default = " << default_numnodes << ")" << endl;
  cout << "-o specifies the prefix for the output file; the complete output"
       << endl
       << "   file name will be in the form of:"
       << endl
       << "   <output file prefix>-<query serial no.>-<iteration no.>.<proc no.>"
       << endl
       << "   where <query serial no.> is some assigned query id" << endl
       << "         <iteration no.> is a strip-mined iteration counter" << endl
       << "         <proc no.> is the processor number" << endl
       << "   (default output file prefix = NULL)"
       << endl;
  cout << "-m specifies the memory size (bytes) for accumulators"
       << " (default = " << default_usrmem << " bytes)" << endl;
  cout << "-f specifies hostname and port number for the front-end process"
       << endl
       << "   (by default, program reads a query batch from a file specified by -q)"
       << endl;
  cout << "-v enables verbose mode" << endl;
}


void badusage(usageError e, const char* p, char a)
{
  switch (e) {
    case eUNKNOWN:
	cerr << "Unknown parameter -" << a << " in command line" << endl;
	break;
    case eBADARG:
	cerr << "Missing argument for parameter -" << a << " in command line"
	     << endl;
	break;
    default:
	cerr << "Some kind of error in the parameter -" << a
	     << " in command line" << endl;
  }
  printusage(p);
}


