
/* $Id: custom.C,v 1.12 2000/03/14 22:55:12 chialin Exp $ */

/* custom.C */


#include	<iostream.h>
#include	<fstream.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<strings.h>
#include	<ctype.h>
#include	<time.h>
#include	"t2_custom.h"


/////////////////////////////////////////////////////////////////////////////
//
// T2_UdfConsRegistrationFile methods
//
/////////////////////////////////////////////////////////////////////////////
T2_UdfConsRegistrationFile::T2_UdfConsRegistrationFile(const char* fname)
				: file('#', 2048), comment('#')
{
  if (fname != NULL)
     open(fname);
}


T2_UdfConsRegistrationEntry*
T2_UdfConsRegistrationFile::getNextEntry(int& eflag)
{
  eflag = 1;

  if (file.eof() == true) {
     eflag = 0;
     return NULL;
  }
  if (!file)				// some other error
     return NULL;

  t2_Assert (file.getFileName() != NULL);

  // Format of the constructor registration file:
  // Each entry has the following fields:
  //    type: which could be dataset, index, accumulator, projection,
  //	      aggregation, metachaos (only the first two chars are examined)
  //    a unique name: for human's use (one single word)
  //    the name of the constructor function (a single word)
  //    description: which could be a blank line if no description is given
  //		     (could consists of multiple words)
  // The field 'description' has its own line, which must come immediately
  // after the line that contains the constructor function name.
  // A blank line with no '#' at the beginning of the line indicates no
  // description.
  // For dataset constructor function entries (that is, entries with types
  // equal to dataset), after the description field, there is information
  // for the iterators the corresponding dataset object provides given in the
  // following format:
  //	<number of dataset iterators>
  //	<name for iterator 1>
  //	<name for iterator 2>
  //		:
  //
  // Remarks:
  // 1) lines start with '#' are considered comments and are ignored
  // 2) max line length is 2048

  const char *p;
  size_t length;

  // read type
  const char* type_field = file.getNextUncommentString();
  if (type_field == NULL) {
     t2_Assert (file.eof() == true);
     eflag = 0;
     return NULL;
  }
  const u_int lineno = file.getCurrentLineNumber();
  T2_UDFType type = t2_shortstring2UDFType(type_field);
  if (type == T2_UDF_Unknown) {
     t2_err_ret(NULL, 0,
		"ERROR: invalid user-defined constructor type '%s' at line %d in file '%s'",
		type_field, lineno, file.getFileName());
     return NULL;
  }

  // read name
  file.getNextLine();
  const char* name_field = file.getNextUncommentString();
  if (name_field == NULL) {
     t2_Assert (file.eof() == true);
     t2_err_ret(NULL, 0,
	   "ERROR: no name for the record starting at line %d in file '%s'.",
	   lineno, file.getFileName());
     return NULL;
  }
  length = strlen(name_field);
  // look for comment char, if any
  p = name_field;
  while (*p != '\0' && *p != comment)
      p++;
  if (*p == comment)
     length = p - name_field;
  char *name;
  t2_Alloc (name, length+1, char);
  strncpy(name, name_field, length);
  name[length] = '\0';

  // read constructor function name
  const char* consf_field = file.getNextUncommentString();
  if (consf_field == NULL) {
     t2_Assert (file.eof() == true);
     t2_err_ret(NULL, 0,
		"ERROR: no constructor funation name is given for the record starting at line %d in file '%s'.",
		lineno, file.getFileName());
     delete name;
     return NULL;
  }
  length = strlen(consf_field);
  // look for comment char, if any
  p = consf_field;
  while (*p != '\0' && *p != comment)
      p++;
  if (*p == comment)
     length = p - consf_field;
  char *consf;
  t2_Alloc (consf, length+1, char);
  strncpy(consf, consf_field, length);
  consf[length] = '\0';

  if (check_function_name(consf) == false) {
     cout << "WARNING: constructor function name '" << consf << "'" << endl
	  << "         at line " << lineno << " in the construtor registration file does not"
	  << endl
	  << "         seem like a valid C++ function name."
	  << endl;
  }

  // read description, if any
  const char *descrp_field = NULL;
  char *descrp = NULL;
  file.getNextLine();
  if (file.eof() == false) {
     if ((descrp_field = file.getStartOfNextUncommentString(file.getLine()))
	 !=NULL)
     {
	length = strlen(descrp_field);

	// look for comment char, if any
	p = descrp_field;
	while (*p != '\0' && *p != comment)
	   p++;
	if (*p == comment)
	   length = p - name_field;

	t2_Alloc(descrp, length+1, char);
	strncpy(descrp, descrp_field, length);
	descrp[length] = '\0';
     }
     file.getNextLine();	// skip current line
  }				// if (file.eof() == false) {

  T2_UdfConsRegistrationEntry *entry = new T2_UdfConsRegistrationEntry;
  entry->type = type;
  entry->name = name;
  entry->function = consf;
  entry->description = descrp;

  if (type == T2_UDF_Dataset) {	// look for iterators
     const char* num_iters_string = file.getNextUncommentString();
     if (num_iters_string == NULL) {
	t2_err_ret(NULL, 0,
		"WARNING: no iterators was specified for dataset constructor function '%s'.",
		entry->name);
     }
     int num_iters, atoi_error;
     num_iters = t2_atoi(num_iters_string, &atoi_error);
     if (atoi_error != 0 || num_iters < 0) {
	t2_err_ret(NULL, 0,
		"ERROR: invalid number of iterators '%s' at line %d in file '%s'.",
		num_iters_string, file.getCurrentLineNumber(),
		file.getFileName());
	delete entry;
	return NULL;
     }
     const char* iter_name;
     T2_VArray<T2_CharPtr>& e_iterators = entry->iterators;
     for (u_int ni=0; ni<(u_int) num_iters; ni++) {
	if ((iter_name = file.getNextUncommentString()) == NULL) {
	   t2_err_ret(NULL, 0,
		"ERROR: only %d out of %d of the iterators for dataset constructor function '%s' are given before reaching eof.",
		ni, num_iters, entry->name);
	   delete entry;
	   return NULL;
	}
	char*& iter_p = e_iterators.appendElement();
	size_t len = strlen(iter_name) + 1;
	t2_Alloc (iter_p, len, char);
	memcpy(iter_p, iter_name, len);
     }			// for (u_int ni=0; ni<(u_int) num_iters; ni++)
  }

  eflag = 0;
  return entry;
}


bool
T2_UdfConsRegistrationFile::check_function_name(const char* funcname)
{
  while (*funcname != '\0') {
     if (isalnum((int) *funcname) || *funcname == '-' || *funcname == '_') {
	funcname++;
	continue;
     }
     return false;
  }
  return true;
}



/////////////////////////////////////////////////////////////////////////////
//
// T2_UdfConstructorFile methods
//
/////////////////////////////////////////////////////////////////////////////
T2_UdfConstructorFile::T2_UdfConstructorFile(const char* fname)
				: valid(false), counts(T2_NUMBER_UDFTypes)

{
  t2_Alloc(files, T2_NUMBER_UDFTypes, fstream);
  if (fname != NULL)
     open(fname);
}


void
T2_UdfConstructorFile::open(const char* fname)
{
  close();

  if (fname == NULL)
     return;

  // check if file already exists
  ifstream ifs(fname, ios::in);
  if (ifs) {				// file already exists
     ifs.close();
     sprintf(tmpfname[0], "%s.bak", fname);
     t2_say(cout, "WARNING: Rename " << fname << " to " << tmpfname[0] << endl);
     if (rename(fname, tmpfname[0]) != 0) {	// try to rename existing file
	t2_err_ret(__FILE__, __LINE__, "ERROR: rename %s to %s fails",
		  fname, tmpfname[0]);
	return;
     }
  }

  // files[0] is open with filename fname, others open with fname-tmpi
  files[0].open(fname, ios::in | ios::out);
  if (!files[0]) {			// open fails
     t2_err_ret(__FILE__, __LINE__, "ERROR: cannot open file '%s'", fname);
     return;
  }

  u_int i;
  for (i=1; i<T2_NUMBER_UDFTypes; i++) {
     sprintf(tmpfname[i], "%s-tmp%d__xyz_", fname, i);
     ifstream ff(tmpfname[i], ios::in);
     if (ff) {
	ff.close();
	char ffname[256];
	sprintf(ffname, "%s.bak", tmpfname[i]);
	if (rename(tmpfname[i], ffname) != 0) {	
	   t2_err_ret(__FILE__, __LINE__, "ERROR: rename %s to %s fails",
		tmpfname[i], ffname);
	   for (u_int j=0; j<i; j++) {
	      files[j].close();
	      unlink(tmpfname[j]);
	   }
	   return;
	}
	t2_say(cout, 
		"WARNING: Rename " << tmpfname[i] << " to " << ffname << endl);
     }
     files[i].open(tmpfname[i], ios::in | ios::out);
     if (!files[i]) {
	t2_err_ret(__FILE__, __LINE__,
		  "ERROR: cannot open temporary file '%s'", tmpfname[i]);
	for (u_int j=0; j<i; j++) {
	  files[j].close();
	  unlink(tmpfname[j]);
	}
	return;
     }
     counts[i] = 0;
  }

  printPrologue(files[0], 0);
  valid = true;
}


void
T2_UdfConstructorFile::close()
{
  if (valid == true) {
     // append files[1..T2_NUMBER_UDFTypes-1] to files[0]
     for (u_int i=1; i<T2_NUMBER_UDFTypes; i++) {
	printPrologue(files[0], i);
	appendFile(files[0], files[i], counts[i]);
	printEpilogue(files[0], i);
	files[i].close();
	unlink(tmpfname[i]);
     }
     files[0].close();
     valid = false;
  }
}


void
T2_UdfConstructorFile::printPrologue(fstream& f, u_int i)
{
  switch (i) {
    case 0:	// dataset
		{
		time_t t = time(NULL);
		char *curtime = ctime(&t);
		if (curtime[24] == '\n')
		   curtime[24] = '\0';

		f << endl
		  << "/*" << endl
		  << " * constructors.C" << endl
		  << " * generated by class T2_UdfConstructorFile at "
		  << curtime << " " << tzname[0] << endl
		  << " */" << endl << endl
		  << "#include	\"t2_header.h\"" << endl
		  << "#include	\"t2_rts.h\"" << endl
		  << endl;
		}
		return;

    case 1:	// dataset
		f << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl << "//" << endl
		  << "// to include constructors for dataset objects"
		  << endl << "//" << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl
		  << "void" << endl
		  << "initDatasetConstructors(T2_Array<T2_DatasetConstructorPtr>& cons)"
		  << endl
		  << "{" << endl
		  << "  // specify number of dataset constructors expected"
		  << endl;
		return;

    case 2:	// index
		f << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl << "//" << endl
		  << "// to include constructors for index objects"
		  << endl << "//" << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl
		  << "void" << endl
		  << "initIndexConstructors(T2_Array<T2_IndexConstructorPtr>& cons)"
		  << endl
		  << "{" << endl
		  << "  // specify number of index constructors expected"
		  << endl;
		return;

    case 3:	// acc meta
		f << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl << "//" << endl
		  << "// to include constructors for accumulator meta objects"
		  << endl << "//" << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl
		  << "void" << endl
		  << "initAccMetaConstructors(T2_Array<T2_AccMetaConstructorPtr>& cons)"
		  << endl
		  << "{" << endl
		  << "  // specify number of acc meta constructors expected"
		  << endl;
		return;

    case 4:	// aggregation function
		f << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl << "//" << endl
		  << "// to include constructors for aggregate function objs"
		  << endl << "//" << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl
		  << "void" << endl
		  << "initAggregateFuncConstructors(T2_Array<T2_AggregateFuncConstructorPtr>& cons)"
		  << endl
		  << "{" << endl
		  << "  // specify number of aggregate function constructors expected"
		  << endl;
		return;

    case 5:	// projection
		f << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl << "//" << endl
		  << "// to include constructors for projection function objs"
		  << endl << "//" << endl
		  << "/////////////////////////////////////////////////////////"
		  << endl
		  << "void" << endl
		  << "initProjectFuncConstructors(T2_Array<T2_ProjectFuncConstructorPtr>& cons)"
		  << endl
		  << "{" << endl
		  << "  // specify number of projection function constructors expected"
		  << endl;
		return;

    case 6:	// meta chaos
		f << endl
		  << "#ifdef T2_METACHAOS" << endl
		  << "///////////////////////////////////////////////////////"
		  << endl << "//" << endl
		  << "// to include constructors for T2_MetaChaosUser objects"
		  << endl << "//" << endl
		  << "///////////////////////////////////////////////////////"
		  << endl
		  << "void" << endl
		  << "initMetaChaosUserConstructors(T2_Array<T2_MetaChaosUserConstructorPtr>& cons)"
		  << endl
		  << "{" << endl
		  << "  // specify number of meta chaos constructors expected"
		  << endl;
		return;
  }

  t2_err_quit(__FILE__, __LINE__, "ERROR: unknown type %d", i);
}


void
T2_UdfConstructorFile::printEpilogue(fstream& f, u_int i)
{
  switch (i) {
     case 6:	// meta chaos
		f << "#endif	/* ! T2_METACHAOS */" << endl;
  }
}


bool
T2_UdfConstructorFile::appendFile(fstream& dest, fstream& src, u_int n)
{
  t2_Assert (dest);
  t2_Assert (src);
//  t2_Assert (n > 0);
  dest << "  cons.clear(" << n << ");" << endl << endl
       << "  // now add the constructors one by one" << endl;

  const size_t max_line_length = 2048;
  char line[max_line_length];
  src.clear();
  src.seekg(0, ios::beg);
  while (src.getline(line, max_line_length) || src.gcount() > 0)
     dest << line << endl;

  dest << "}" << endl << endl;
  return true;
}


bool
T2_UdfConstructorFile::addEntry(T2_UDFType t, const char* funcname, u_int& id)
{
  u_int i = (u_int) t;
  if (t == T2_UDF_Unknown || funcname == NULL || !files[0])
     return false;

  t2_Assert (i < T2_NUMBER_UDFTypes);
  id = counts[i]++;

  switch (t) {
    case T2_UDF_Dataset:
		files[1] << "  T2_DatasetConstructor "
			 << funcname << ";" << endl
			 << "  cons[" << id << "] = " << funcname << ";"
			 << endl;
		break;
    case T2_UDF_Index:
		files[2] << "  T2_IndexConstructor "
			 << funcname << ";" << endl
			 << "  cons[" << id << "] = " << funcname << ";"
			 << endl;
		break;
    case T2_UDF_AccMeta:
		files[3] << "  T2_AccMetaConstructor "
			 << funcname << ";" << endl
			 << "  cons[" << id << "] = " << funcname << ";"
			 << endl;
		break;
    case T2_UDF_Aggregation:
		files[4] << "  T2_AggregateFuncConstructor "
			 << funcname << ";" << endl
			 << "  cons[" << id << "] = " << funcname << ";"
			 << endl;
		break;
    case T2_UDF_Projection:
		files[5] << "  T2_ProjectFuncConstructor "
			 << funcname << ";" << endl
			 << "  cons[" << id << "] = " << funcname << ";"
			 << endl;
		break;
    case T2_UDF_MetaChaos:
		files[6] << "  T2_MetaChaosUserConstructorPtr "
			 << funcname << ";" << endl
			 << "  cons[" << id << "] = " << funcname << ";"
			 << endl;
		break;
    default:	// unrecognized type
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: unrecognized type %d", i);
		return false;
  }
  return true;
}
